/*
 *  TEX Device Driver
 *
 *  bitblt.c:   tpic specials (version 2.2) support routines
 *              Based on the specifications included in transfig package
 *              by Oh-Yeah?     14 November 1992
 *              Note.  bitblt_ap() is almost equivalent to w_font() in the 
 *              original package.  bitblt_ow() is an overwriting version
 *              of bitblt_ap(), i.e., destination bitmap underneath is lost.
 *
 *  modified for Minix-386 1992 by youchan
 */

#include "config.h"
#include "dd.h"
#include <math.h>

#ifndef NOTPIC
#ifndef NOTPIC_EXTENSION
/* bitmap.c */
typedef struct {PIXEL x, y;} pixelpoint;
pixelpoint turn_off_trimming(void);
void turn_on_trimming(void);
#endif
#endif

void bitblt_ap(BUFFER*, int, int, int, BUFFER*, int, int, int);
void (*bitblt)(BUFFER*, int, int, int, BUFFER*, int, int, int) = bitblt_ap;
/* ptr to the actual function */

#ifndef NOTPIC
void bitblt_ow(BUFFER*, int, int, int, BUFFER*, int, int, int);

/* called by init_tile() and fin_tile() in bitmap.c */
void alter_bitblt_ow(BOOL);

#ifndef NOTPIC_EXTENSION
/* called by tpic_rt(), freeze_tpic_status(), melt_tpic_status() in tpic.c */
void alter_bitblt_rt(BOOL);
void set_rot_matrix(PIXEL, PIXEL, double);

void bitblt_rt_ap(BUFFER*, int, int, int, BUFFER*, int, int, int);
void bitblt_rt_ow(BUFFER*, int, int, int, BUFFER*, int, int, int);
#else
#define bitblt_rt_ow bitblt_ow  /* just for alter_bitblt_ow() */
#define bitblt_rt_ap bitblt_ap
#endif
#endif

static union {
    uint  x;
    uchar _h[4];
#define hi _h[3]
#define m2 _h[2]
#define m1 _h[1]
#define lo _h[0]
} u,v;

void xchg()
{
    uchar a;

    a = u._h[0];
    u._h[0] = u._h[3];
    u._h[3] = a;
    a = u._h[1];
    u._h[1] = u._h[2];
    u._h[2] = a;
}

void xchg2()
{
    uchar a;

    a = v._h[0];
    v._h[0] = v._h[3];
    v._h[3] = a;
    a = v._h[1];
    v._h[1] = v._h[2];
    v._h[2] = a;
}


/* bitmap block transfer from source to dest (bits are appended)
 *
 * dest: byte addr = dest_byte + d_offset_y_bit * d_linewidth_byte 
 *  + d_offset_x_bit / 8); bit offset = d_offset_x_bit & 0x07
 * if d_offset_x_bit or d_offset_y_bit < 0, trim away left/top bits
 * caution: no trimming is carried out for right/bottom bits
 */

void bitblt_ap(BUFFER *source_byte, int s_linewidth_byte, int s_width_bit, 
  int s_height_bit, BUFFER *dest_byte, int d_linewidth_byte, 
  int d_offset_x_bit, int d_offset_y_bit)
{
    static BUFFER *d_byte_prev = NULL;
    static BUFFER *d_addr_prev;
    static int d_y_prev = -1;           /* for dest y cache */
    static int d_lwidth_prev = -1;
    uchar  r_mask, leftest_mask;  /* byte mask */
    uint   rightest_mask;         /* word mask */
    int    word_count, byte;
    uchar  keep_left_bit, k0;

    if (d_offset_y_bit < 0) {           /* trim top */
        s_height_bit += d_offset_y_bit;
        if (s_height_bit <= 0) return;  /* nothing inside */
        source_byte += (-d_offset_y_bit) * s_linewidth_byte; /* trim source */
        d_offset_y_bit = 0;
    }

    if (d_offset_y_bit == d_y_prev && dest_byte == d_byte_prev
      && d_lwidth_prev == d_linewidth_byte) {
        dest_byte = d_addr_prev;
    }
    else {
        d_y_prev = d_offset_y_bit;
        d_byte_prev = dest_byte; 
        d_lwidth_prev = d_linewidth_byte;
        dest_byte += d_offset_y_bit * d_linewidth_byte;
        d_addr_prev = dest_byte;
    }

    if (d_offset_x_bit >= 0) {
        dest_byte += d_offset_x_bit >> 3;
        leftest_mask = 0xff; /* write all */
        d_offset_x_bit &= 0x07; /* positive mod (0-7) */
    } else {                 /* trim left bits */
        if (s_width_bit + d_offset_x_bit <= 0) return; /* nothing inside */
        byte = (-d_offset_x_bit) >> 3;
        s_width_bit -= byte << 3;
        source_byte += byte; /* trim source bytes */
        d_offset_x_bit &= 0x07;
        /* positive mod (0-7) = (8 * (byte + 1) + d_offset_x_bit) % 8 */
        if (d_offset_x_bit == 0) {
            leftest_mask = 0xff; /* write all */
        } else { /* residual bits, trim */
            dest_byte--; /* might be wrong but Okay, won't be written anyway */
            leftest_mask = 0x00; /* trim away leftest bytes */
        }
    }

    r_mask = (uchar)0xff >> d_offset_x_bit;
        /* AND mask for getting right bits */
    leftest_mask = r_mask;        /* leftest byte: get right bits and trim */

    s_width_bit += d_offset_x_bit; /* total bits */
    word_count = s_width_bit >> 5;
    s_width_bit &= 0x1f; /* mod bits (0 - 31) */
    rightest_mask = ~((uint)0xffffffff >> s_width_bit);
        /* rightest word: get left bits */

    while (s_height_bit-- > 0) {
        uint *s = (uint*)source_byte;
        uint *d = (uint*)dest_byte;
        int   x = word_count;
        uchar c;

        if (x-- > 0) {   /* leftest word */
            u.x = *s++;
            keep_left_bit = u.hi << (8 - d_offset_x_bit);
            xchg();
            u.x = u.x >> d_offset_x_bit;
            xchg();
            u.lo &= leftest_mask;
            *d++ |= u.x;

            while (x-- > 0) {   /* following words */
                u.x = *s++;
                k0 = u.hi << (8 - d_offset_x_bit);
                xchg();
                u.x = u.x >> d_offset_x_bit;
                xchg();
                u.lo |= keep_left_bit;
                *d++ |= u.x;
                keep_left_bit = k0;
            }
        }
        else {
            keep_left_bit = 0;  /* initial keep for previous left bits */
        }

        if (s_width_bit > 0) {   /* rightest word */
            u.x = *s++;
            xchg();
            u.x = (u.x >> d_offset_x_bit) & rightest_mask;
            xchg();
            u.lo |= keep_left_bit;
            *d++ |= u.x;
        }
        dest_byte += d_linewidth_byte;
        source_byte += s_linewidth_byte;
    }
}

#ifndef NOTPIC

/* this is an OVERWRITE version of bilblt_ap(). Search OVERWRITE for changes
 * bitmap block transfer from source to dest
 * dest: byte addr = dest_byte + d_offset_y_bit * d_linewidth_byte 
 *  + d_offset_x_bit / 8); bit offset = d_offset_x_bit & 0x07
 * if d_offset_x_bit or d_offset_y_bit < 0, trim away left/top bits
 * caution: no trimming is carried out for right/bottom bits
 */

void bitblt_ow(BUFFER *source_byte, int s_linewidth_byte, int s_width_bit, 
  int s_height_bit, BUFFER *dest_byte, int d_linewidth_byte, 
  int d_offset_x_bit, int d_offset_y_bit)
{
    static BUFFER *d_byte_prev = NULL;
    static BUFFER *d_addr_prev;
    static int d_y_prev = -1;           /* for dest y cache */
    static int d_lwidth_prev = -1;      /* for dest linewidth cache */
    BUFFER *s;
    BUFFER *d;                    /* this is actually a huge ptr */
    uchar  r_mask, leftest_mask;  /* byte mask */
    uint   rightest_mask;         /* word mask */
    int    word_count, byte;
    uchar  keep_left_bit, k0;

    if (d_offset_y_bit < 0) {           /* trim top */
        s_height_bit += d_offset_y_bit;
        if (s_height_bit <= 0) return;  /* nothing inside */
        source_byte += (-d_offset_y_bit) * s_linewidth_byte; /* trim source */
        d_offset_y_bit = 0;
    }

    if (d_offset_y_bit == d_y_prev && dest_byte == d_byte_prev
      && d_lwidth_prev == d_linewidth_byte) { /* cache */
        dest_byte = d_addr_prev;
    } else {
        d_y_prev = d_offset_y_bit;
        d_byte_prev = dest_byte;
        d_lwidth_prev = d_linewidth_byte;
        dest_byte += d_offset_y_bit * d_linewidth_byte;
        d_addr_prev = dest_byte;
    }

    if (d_offset_x_bit >= 0) {
        dest_byte += d_offset_x_bit >> 3;
        leftest_mask = 0xff; /* write all */
        d_offset_x_bit &= 0x07; /* positive mod (0-7) */
    } else {                 /* trim left bits */
        if (s_width_bit + d_offset_x_bit <= 0) return; /* nothing inside */
        byte = (-d_offset_x_bit) >> 3;
        s_width_bit -= byte << 3;
        source_byte += byte; /* trim source bytes */
        d_offset_x_bit &= 0x07;
        /* positive mod (0-7) = (8 * (byte + 1) + d_offset_x_bit) % 8 */
        if (d_offset_x_bit == 0) {
            leftest_mask = 0xff; /* write all */
        } else { /* residual bits, trim */
            dest_byte--; /* might be wrong but Okay, won't be written anyway */
            leftest_mask = 0x00; /* trim away leftest bytes */
        }
    }

    r_mask = (uchar)0xff >> d_offset_x_bit;
        /* AND mask for getting right bits */
    leftest_mask = r_mask;        /* leftest byte: get right bits and trim */

    s_width_bit += d_offset_x_bit; /* total bits */
    word_count = s_width_bit >> 5;
    s_width_bit &= 0x1f; /* mod bits (0 - 31) */
    rightest_mask = ~((uint)0xffffffff >> s_width_bit);
        /* rightest word: get left bits */

    while (s_height_bit-- > 0) {
        uint *s = (uint*)source_byte;
        uint *d = (uint*)dest_byte;
        int   x = word_count;
        uchar c;

        if (x-- > 0) {   /* leftest word */
            u.x = *s++;
            keep_left_bit = u.hi << (8 - d_offset_x_bit);
            xchg();
            u.x = u.x >> d_offset_x_bit;
            xchg();
            u.lo &= leftest_mask;
            v.x = *d;
            v.lo &= (~leftest_mask);
            v.m1 = v.m2 = v.hi = 0;
            *d++ = v.x | u.x;

            while (x-- > 0) {   /* following words */
                u.x = *s++;
                k0 = u.hi << (8 - d_offset_x_bit);
                xchg();
                u.x = u.x >> d_offset_x_bit;
                xchg();
                u.lo |= keep_left_bit;
                *d++ = u.x;
                keep_left_bit = k0;
            }
        }
        else {
            keep_left_bit = 0;  /* initial keep for previous left bits */
        }

        if (s_width_bit > 0) {   /* rightest word */
            u.x = *s++;
            xchg();
            u.x = (u.x >> d_offset_x_bit) & rightest_mask;
            xchg();
            u.lo |= keep_left_bit;
            v.x = *d;
            xchg2();
            v.x &= (~rightest_mask);
            xchg2();
            *d++ = v.x | u.x;
        }
        dest_byte += d_linewidth_byte;
        source_byte += s_linewidth_byte;
    }
}

static BOOL overwrite = FALSE, rotate = FALSE;

void alter_bitblt_ow(BOOL flag)
{
    if (flag)  bitblt = (rotate) ? bitblt_rt_ow : bitblt_ow;
    else       bitblt = (rotate) ? bitblt_rt_ap : bitblt_ap;
    overwrite = flag;
    return;
}

#ifndef NOTPIC_EXTENSION

/* Here are rotation routines */

static PIXEL hmax, vmax;

void alter_bitblt_rt(BOOL flag)
{
    pixelpoint trimmax;

    if (flag) {
        bitblt = (overwrite) ? bitblt_rt_ow : bitblt_rt_ap;
        trimmax = turn_off_trimming(); /* do not trim in bitmap.c */
        hmax = trimmax.x; /* bitmap max. use 'em for trimming here */
        vmax = trimmax.y;
    } else {
        bitblt = (overwrite) ? bitblt_ow : bitblt_ap;
        turn_on_trimming();
    }
    rotate = flag;
    return;
}

#define ROT_COEFF long  /* rotation matrix coefficients */
#define ROT_TIMES 10000 /* constant for double-to-int conversion */

static ROT_COEFF rot_mat[2][3];

void set_rot_matrix(PIXEL origin_x, PIXEL origin_y, double angle)
    /* angle in radians, clockwise */
{
    double c, s;

    c = cos(angle);
    s = sin(angle);
    rot_mat[0][0] = (ROT_COEFF)(c * ROT_TIMES);
    rot_mat[0][1] = -(ROT_COEFF)(s * ROT_TIMES);
    rot_mat[0][2] = (ROT_COEFF)((origin_x * (1.0 - c) + origin_y * s) * ROT_TIMES);
    rot_mat[1][0] = (ROT_COEFF)(s * ROT_TIMES);
    rot_mat[1][1] = (ROT_COEFF)(c * ROT_TIMES);
    rot_mat[1][2] = (ROT_COEFF)((origin_y * (1.0 - c) - origin_x * s) * ROT_TIMES);
    return;
}

/* bitmap block transfer from source to dest (bits are appended) */
/* dest (in TeX): byte addr = dest_byte + d_offset_y_bit * d_linewidth_byte 
    + d_offset_x_bit / 8); bit offset = d_offset_x_bit & 0x07 */
/* dest (real)  : every bit is rotated by
                d_x = rot_mat[0][0] * d_offset_x_bit 
                + rot_mat[0][1] * d_offset_y_bit + rot_mat[0][2];
                d_y = rot_mat[1][0] * d_offset_x_bit 
                + rot_mat[1][1] * d_offset_y_bit + rot_mat[1][2];
    byte addr  = dest_byte + d_y * d_linewidth_byte + d_x / 8); 
    bit offset = d_x & 0x07 */
/* negative d_offset_x_bit or d_offset_y_bit is just accepted, without any 
trimming.  d_x and d_y are checked and trimmed if necessary */
/* Note: this routine is slow, not optimized at all (^^); */

void bitblt_rt_ap(BUFFER *source_byte, int s_linewidth_byte, int s_width_bit, 
  int s_height_bit, BUFFER *dest_byte, int d_linewidth_byte, 
  int d_offset_x_bit, int d_offset_y_bit)
{
    int s_x, s_y, sbit, d_x0, d_x, d_y, dbit;
    unsigned char sbit_value;
    BUFFER *s_byte;
    BUFFER *d_byte;

    for (s_y = 0; s_y < s_height_bit; s_y++, source_byte += s_linewidth_byte, 
        d_offset_y_bit++) {
        for (s_x = 0, s_byte = source_byte, sbit = 7, d_x0 = d_offset_x_bit; 
            s_x < s_width_bit; s_x++, sbit--, d_x0++) {
            if (sbit < 0) { /* next byte */
                s_byte++;
                sbit = 7;
            }
            sbit_value = (*s_byte >> sbit) & 0x01;
            if (sbit_value) { /* source bit is on */
                d_x = (int)((rot_mat[0][0] * d_x0 
                      + rot_mat[0][1] * d_offset_y_bit + rot_mat[0][2]) / ROT_TIMES);
                d_y = (int)((rot_mat[1][0] * d_x0 
                      + rot_mat[1][1] * d_offset_y_bit + rot_mat[1][2]) / ROT_TIMES);
                if (d_x < 0 || d_x >= hmax) continue; /* trim */
                if (d_y < 0 || d_y >= vmax) continue;
                d_byte = dest_byte + (ulong) d_y * d_linewidth_byte;
                d_byte += (d_x >> 3);
                dbit = 7 - (d_x & 0x07);
                *d_byte |= (sbit_value << dbit);
            }
        }
    }
}


/* this is an OVERWRITE version of bilblt_rt_ap(). see OVERWRITE for changes */
/* bitmap block transfer from source to dest */
/* dest (in TeX): byte addr = dest_byte + d_offset_y_bit * d_linewidth_byte 
    + d_offset_x_bit / 8); bit offset = d_offset_x_bit & 0x07 */
/* dest (real)  : every bit is rotated by
                d_x = rot_mat[0][0] * d_offset_x_bit 
                + rot_mat[0][1] * d_offset_y_bit + rot_mat[0][2];
                d_y = rot_mat[1][0] * d_offset_x_bit 
                + rot_mat[1][1] * d_offset_y_bit + rot_mat[1][2];
    byte addr  = dest_byte + d_y * d_linewidth_byte + d_x / 8); 
    bit offset = d_x & 0x07 */
/* negative d_offset_x_bit or d_offset_y_bit is just accepted, without any 
trimming.  d_x and d_y are checked and trimmed if necessary */
/* Note: this routine is slow, not optimized at all (^^); */

void bitblt_rt_ow(BUFFER *source_byte, int s_linewidth_byte, int s_width_bit, 
  int s_height_bit, BUFFER *dest_byte, int d_linewidth_byte, 
  int d_offset_x_bit, int d_offset_y_bit)
{
    int s_x, s_y, sbit, d_x0, d_x, d_y, dbit;
    unsigned char sbit_value;
    BUFFER *s_byte;
    BUFFER *d_byte;

    for (s_y = 0; s_y < s_height_bit; s_y++, source_byte += s_linewidth_byte, 
        d_offset_y_bit++) {
        for (s_x = 0, s_byte = source_byte, sbit = 7, d_x0 = d_offset_x_bit; 
            s_x < s_width_bit; s_x++, sbit--, d_x0++) {
            if (sbit < 0) { /* next byte */
                s_byte++;
                sbit = 7;
            }
            sbit_value = (*s_byte >> sbit) & 0x01;
            d_x = (int)((rot_mat[0][0] * d_x0 
                  + rot_mat[0][1] * d_offset_y_bit + rot_mat[0][2]) / ROT_TIMES);
            d_y = (int)((rot_mat[1][0] * d_x0 
                  + rot_mat[1][1] * d_offset_y_bit + rot_mat[1][2]) / ROT_TIMES);
            if (d_x < 0 || d_x >= hmax) continue; /* trim */
            if (d_y < 0 || d_y >= vmax) continue;
            d_byte = dest_byte + (ulong) d_y * d_linewidth_byte;
            d_byte += (d_x >> 3);
            dbit = 7 - (d_x & 0x07);
            *d_byte |= (sbit_value << dbit);
            *d_byte &= ~(0x01 << dbit);
        }
    }
}

#endif /* end of ifndef NOTPIC_EXTENSION */
#endif /* end of ifndef NOTPIC */

/* end of bitblt.c */
