/*
 *  TEX Device Driver  ver 2.02-2.22
 *  (c)copyright 1988, 1989 by TSG, 1990 by SHIMA
 *
 *  bitmap.c : to write characters and rules on bitmap buffer
 *             2nd edition
 *             14 Nove 1992 : modified for tpic by Oh-Yeah?
 *             22 July 1992 : modified for pTeX by Naochan!.
 *
 *  modified for Minix-386 1992 by obuk, youchan
 */

#include "config.h"
#include "dd.h"
#include "err.h"

/* bitblt.c */
extern void (*bitblt)(BUFFER *, int, int, int, BUFFER *, int, int, int);
#ifndef NOTPIC
void alter_bitblt_ow(BOOL);
#endif
  
/* size.c, prtsize.c */
PIXEL sptopixel(SCALED_PT);
PIXEL vtopixel(SCALED_PT);

static BUFFER *map_top_ptr;
static PIXEL vmax, hmax, hmaxb, vmin, hmin;
BOOL    draw_baseline = 0;
BOOL    draw_boxpic = 0;
extern char ptex_d;

void bitmap_init(OUTPUT_INFO *out);
void boxpic(PIXEL h, PIXEL v, PIXEL w, PIXEL d);
void write_font(PIXEL h, PIXEL v, PREAMBLE *pxl);
void write_rule(PIXEL h, PIXEL v, PIXEL a, PIXEL b);
#ifndef NOTPIC
void init_tile(double grayscale) /* initialize overriding for tiles */;
void fin_tile(void) /* finish tiling. recover original data */;
void write_tile(PIXEL h, PIXEL v, PIXEL a, PIXEL b);
void set_tile(double grayscale);
void adjust_tile(PIXEL h, PIXEL v) /* adjust pattern by (h, v) position */;
#ifdef VFTPIC
/* called by draw_character() in vfont.c */
void init_alter_bitmap(BUFFER *dest, PIXEL d_linewidth_byte);
void fin_alter_bitmap(void);
#endif
#ifndef NOTPIC_EXTENSION
/* called by alter_bitblt_rt() in bitblt.c */
typedef struct {PIXEL x, y;} pixelpoint;
pixelpoint turn_off_trimming(void);
void turn_on_trimming(void);
#endif
#endif

BUFFER* rotate_raster(PREAMBLE* preamble);
uchar leftside_right(uchar orig);

void bitmap_init(OUTPUT_INFO *out)
{
    map_top_ptr = (BUFFER *)out->bitmap_ptr;
    hmax = out->width;
    vmax = out->height;
    vmin = hmin = 0;
    hmaxb = out->byte_width;
}

#define  WIDTH_OF_LINE  26214L  /* 26214(scaled point) = 0.4pt */

void boxpic(PIXEL h, PIXEL v, PIXEL w, PIXEL d)
{
    PIXEL   w_h, w_v;
    w_h = vtopixel(WIDTH_OF_LINE);
    w_v = sptopixel(WIDTH_OF_LINE);
  
    write_rule(     h     , v + (w_h - 1), w_h,  w );
    write_rule(     h     ,   v + d - 1  , w_h,  w );
    write_rule(     h     ,   v + d - 1  ,  d , w_v);
    write_rule(h + w - w_v,   v + d - 1  ,  d , w_v);
}
  
void write_font(PIXEL h, PIXEL v, PREAMBLE *pxl)
{
    PIXEL he, ve; /* end points */
    PIXEL haba, takasa, feed;
    int   byte_width;
    BUFFER* raster;
  
    haba = pxl->width;
    takasa = pxl->height;
    raster = pxl->raster;

    if (ptex_d == 1) {
        h -= (pxl->height - pxl->y_offset);
        v -= pxl->x_offset;

        h += pxl->shift_right_ptex;
        v -= pxl->shift_up_ptex;

        if (pxl->rotate_ptex & 1) {
            haba = pxl->height;
            takasa = pxl->width;
            raster = rotate_raster(pxl);
        }
    } else {
        h -= pxl->x_offset;
        v -= pxl->y_offset;
    }
    byte_width = (haba + 7) / 8;
    if (v < vmax && h < hmax
         && (he = h + haba) > hmin && (ve = v + takasa) > vmin) {
        if (he > hmax) he = hmax; /* right trimmed */
        if (ve > vmax) ve = vmax; /* bottom trimmed */
        if (draw_boxpic != 0) {
            v += pxl->shift_up_ptex;
            if ((he = h + haba - pxl->shift_right_ptex) > hmax) he = hmax;
            boxpic(h, v, he - h, ve - v);
        } else
            bitblt(raster, byte_width, he - h, ve - v, 
                map_top_ptr, hmaxb, h, v); /* left and top will be trimmed */

    }
}
  
#define LINE_LEN 4
#define ONEBYTE  8

static uchar box_data[LINE_LEN] = {0xff,0xff,0xff,0xff}; /* black */
static BUFFER *tile_ptr = (BUFFER *)box_data;
static int tile_width_bit = LINE_LEN * ONEBYTE; /* pattern width */
static int tile_inc_byte = 0; /* increment = 0; use same pattern repeatedly */
static int tile_height_bit = 10000; /* actually infinite */
    /* used for overriding write_rule() by write_tile() for tpic */

void write_rule(PIXEL h, PIXEL v, PIXEL a, PIXEL b)
{
    PIXEL he, ve; /* end points */
    PIXEL hs, hspan, hrep, hres, vspan, vrep, vres, i;

    if (a >= 0)
        v -= a - 1;
    else
        a = -a;

    if (b < 0) {
        b = -b;
        h -= b - 1; /* (h, v) = rule's top-left point */ 
    }
  
    if (v < vmax && h < hmax
        && (he = h + b) > hmin && (ve = v + a) > vmin) {
        if (he > hmax) he = hmax; /* right trimmed */
        if (ve > vmax) ve = vmax; /* bottom trimmed */
        /* left and top will be trimmed in bitblt() */

        vrep = (vspan = ve - v) / tile_height_bit;
        vres = vspan % tile_height_bit;
        hrep = (hspan = he - h) / tile_width_bit;
        hres = hspan % tile_width_bit;
        hs = h;

        while (vrep-- > 0) {
            i = hrep;
            h = hs;
            while (i-- > 0) {
                bitblt(tile_ptr, tile_inc_byte, tile_width_bit, 
                    tile_height_bit, map_top_ptr, hmaxb, h, v);
                h += tile_width_bit;
            }
            if (hres > 0)
                bitblt(tile_ptr, tile_inc_byte, hres,
                    tile_height_bit, map_top_ptr, hmaxb, h, v);
            v += tile_height_bit;
        }
        if (vres > 0) {
            i = hrep;
            h = hs;
            while (i-- > 0) {
                bitblt(tile_ptr, tile_inc_byte, tile_width_bit, 
                    vres, map_top_ptr, hmaxb, h, v);
                h += tile_width_bit;
            }
            if (hres > 0)
                bitblt(tile_ptr, tile_inc_byte, hres,
                    vres, map_top_ptr, hmaxb, h, v);
        }

    }

}

/* end of file : bitmap.c */

/* additional functions for tpic */

#ifndef NOTPIC

/* tiling functions */
  
static BUFFER *temp_tile; /* just for keeping 'em */
static int temp_width, temp_inc, temp_height;

void init_tile(double grayscale) /* initialize overriding for tiles */
{
    temp_tile = tile_ptr; /* just keep 'em */
    temp_width = tile_width_bit;
    temp_inc = tile_inc_byte;
    temp_height = tile_height_bit;
    alter_bitblt_ow((grayscale <= 0) ? TRUE : FALSE);  /* if white, ow */
    set_tile(grayscale);
}

void fin_tile(void) /* finish tiling. recover original data */
{
    tile_ptr = temp_tile; /* just recover 'em */
    tile_width_bit = temp_width;
    tile_inc_byte = temp_inc;
    tile_height_bit = temp_height;
    alter_bitblt_ow(FALSE);  /* no overwriting */
}

void write_tile(PIXEL h, PIXEL v, PIXEL a, PIXEL b)
{
    adjust_tile(h, v); /* modify pattern by position, if necessary */
    write_rule(h, v, a, b);
}

#ifndef TPIC_SCATTERED

#define PATTERN_BIT  4   /* maximum = 4 */
#define MAX_PATTERN  10  /* maximum = PATTERN_BIT^2, including full black */

static uchar pattern[MAX_PATTERN - 1][PATTERN_BIT * 2] = {
    /* 4 x 4 patterns (except for full black). lower 4 bits and last 4 bytes 
    are just repeated patterns for avoiding rotation of bits */

/* these are rational but too dark because of dots emphasized by printers */
/*      {0x00, 0x22, 0x00, 0x00, 0x00, 0x22, 0x00, 0x00},
        {0x00, 0x22, 0x00, 0x88, 0x00, 0x22, 0x00, 0x88},
        {0x00, 0xaa, 0x00, 0x88, 0x00, 0xaa, 0x00, 0x88},
        {0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa},
        {0x00, 0xaa, 0x44, 0xaa, 0x00, 0xaa, 0x44, 0xaa},
        {0x11, 0xaa, 0x44, 0xaa, 0x11, 0xaa, 0x44, 0xaa},
        {0x44, 0xaa, 0x55, 0xaa, 0x44, 0xaa, 0x55, 0xaa},
        {0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa},
        {0xbb, 0x55, 0xaa, 0x55, 0xbb, 0x55, 0xaa, 0x55},
        {0xee, 0x55, 0xbb, 0x55, 0xee, 0x55, 0xbb, 0x55},
        {0xff, 0x55, 0xbb, 0x55, 0xff, 0x55, 0xbb, 0x55},
        {0xff, 0x55, 0xff, 0x55, 0xff, 0x55, 0xff, 0x55},
        {0xff, 0x55, 0xff, 0x77, 0xff, 0x55, 0xff, 0x77},
        {0xff, 0xdd, 0xff, 0x77, 0xff, 0xdd, 0xff, 0x77},
        {0xff, 0xdd, 0xff, 0xff, 0xff, 0xdd, 0xff, 0xff} */

/* these are from trial-and-errors */
        {0x00, 0x22, 0x00, 0x00, 0x00, 0x22, 0x00, 0x00},
        {0x00, 0x22, 0x00, 0x88, 0x00, 0x22, 0x00, 0x88},
        {0x00, 0x22, 0x44, 0x00, 0x00, 0x22, 0x44, 0x00},
        {0x00, 0x66, 0x44, 0x00, 0x00, 0x66, 0x44, 0x00},
        {0x00, 0xaa, 0x00, 0x88, 0x00, 0xaa, 0x00, 0x88},
        {0x00, 0x66, 0x66, 0x00, 0x00, 0x66, 0x66, 0x00},
        {0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa, 0x00, 0xaa},
        {0x11, 0xaa, 0x44, 0xaa, 0x11, 0xaa, 0x44, 0xaa},
        {0xee, 0x55, 0xbb, 0x55, 0xee, 0x55, 0xbb, 0x55}

};

static uchar tile_data[PATTERN_BIT]; /* selected pattern for gray tiling */
static int black; /* selected pattern number */

/* set_tile() :
   set static parameters: tile_ptr, tile_width_bit, tile_inc_byte and
   tile_height_bit */
/* gray image is realized by periodic pattern */

void set_tile(double grayscale)
{
    static uchar white_data[LINE_LEN] = {0x00,0x00,0x00,0x00}; /* white */
    static double max_pat_d = MAX_PATTERN * 1.01; /* against float calc err */

    black = (int)(grayscale * max_pat_d);
    if (black >= MAX_PATTERN) {      /* black, fill by append (OR) */
        tile_ptr = (BUFFER *)box_data;
        tile_width_bit = LINE_LEN * ONEBYTE;
        tile_inc_byte = 0;
        tile_height_bit = 10000;
    } else if (black <= 0) { /* white, fill by overwrite (AND) */
        tile_ptr = (BUFFER *)white_data;
        tile_width_bit = LINE_LEN * ONEBYTE;
        tile_inc_byte = 0;
        tile_height_bit = 10000;
    } else {                 /* gray pattern (4x4), fill by append (OR) */
        tile_ptr = (BUFFER *)tile_data; /* actual pattern is not copied yet */
        tile_width_bit = PATTERN_BIT;
        tile_inc_byte = 1;
        tile_height_bit = PATTERN_BIT;
    }
}

void adjust_tile(PIXEL h, PIXEL v) /* adjust pattern by (h, v) position */
{
    int i, n, rot_x, rot_y;

    if (black > 0 && black < MAX_PATTERN) { /* gray. rotate pattern */
        rot_x = (unsigned)h % PATTERN_BIT;
        rot_y = (unsigned)v % PATTERN_BIT;
        /* mod for rotation by (h, v) position */

        n = black - 1;
        for (i = 0; i < PATTERN_BIT; i++)
        tile_data[i] = (pattern[n][i + rot_y] << rot_x);
    }
}

#else /* if TPIC_SCATTERED is defined ..... */

/* set static parameters: tile_ptr, tile_width_bit, tile_inc_byte and
   tile_height_bit */
/* gray image is realized by randomly scattered pattern */

void set_tile(double grayscale)
{
    static uchar white_data[LINE_LEN] = {0x00,0x00,0x00,0x00}; /* white */
    static uchar tile_data[LINE_LEN];
    int black, i, j;

    black = (int)(grayscale * ONEBYTE);
    if (black >= ONEBYTE) {      /* black, fill by append (OR) */
        tile_ptr = (BUFFER *)box_data;
        tile_width_bit = LINE_LEN * ONEBYTE;
        tile_inc_byte = 0;
        tile_height_bit = 10000;
    } else if (black <= 0) { /* white, fill by overwrite (AND) */
        tile_ptr = (BUFFER *)white_data;
        tile_width_bit = LINE_LEN * ONEBYTE;
        tile_inc_byte = 0;
        tile_height_bit = 10000;
    } else {                     /* gray (8x4), fill by append (OR) */
        tile_ptr = (BUFFER *)tile_data;
        tile_width_bit = LINE_LEN * ONEBYTE;
        tile_inc_byte = 0;
        tile_height_bit = 10000;

        for (i = 0; i < LINE_LEN; i++) { /* randomly scattered pattern */
            tile_data[i] = 0;
            for (j = 0; j < ONEBYTE; j++)
                if (rand() % ONEBYTE < black) tile_data[i] |= (0x01 << j);
        }
    }
}

void adjust_tile(PIXEL dummy1, PIXEL dummy2)
{
	return;
}

void adjust_tile(PIXEL dummy1, PIXEL dummy2)
{
	return;
}

#endif /* end of ifndef TPIC_SCATTERED ... else ... */

#ifdef VFTPIC

/* vector-font support functions */

static BUFFER *temp_map_top_ptr; /* keepings for overriding dest bitmap */
static PIXEL temp_hmaxb;

void init_alter_bitmap(BUFFER *dest, PIXEL d_linewidth_byte)
{
	temp_map_top_ptr = map_top_ptr; /* keep 'em */
	temp_hmaxb = hmaxb;
	map_top_ptr = dest;
	hmaxb = d_linewidth_byte;
}

void fin_alter_bitmap(void)
{
	map_top_ptr = temp_map_top_ptr; /* recover 'em */
	hmaxb = temp_hmaxb;
}

#endif /* end of ifdef VFTPIC */

#ifndef NOTPIC_EXTENSION
/* rotation routines */

#define MAX_PIXEL 32766  /* max value of type short int */
static PIXEL temp_hmax, temp_vmax;
static BOOL is_trimming_on = TRUE;

/* Stop trimming and ret hmax and vmax */
pixelpoint turn_off_trimming()
{
    pixelpoint answer;

    answer.x = temp_hmax = hmax;
    answer.y = temp_vmax = vmax;
    hmax = vmax = MAX_PIXEL;
    hmin = vmin = -MAX_PIXEL;  /* minimum is zero, not kept */
    is_trimming_on = FALSE;
    return(answer);
}

void turn_on_trimming()
{
    if (!is_trimming_on) {
        hmax = temp_hmax;
        vmax = temp_vmax;
        hmin = vmin = 0;
        is_trimming_on = TRUE;
    }
}
#endif /* end of ifndef NOTPIC_EXTENSION */
#endif /* end of ifndef NOTPIC */

BUFFER* rotate_raster(PREAMBLE* preamble)
{
    uchar work[2048];
    uchar work2[17];
      /* 2048 = 128^2 / 8, 17 = 128 / 8 + 1
       * 118dpi: 128x128 -> 78.4pt
       * 360dpi: 128x128 -> 25.7pt
       * 640dpi: 128x128 -> 14.4pt
       */
    BUFFER* new_raster;
    BUFFER* old_raster;
    int old_height, old_byte_width;
    int new_width, new_height, new_byte_width;
    int ix, iy, bit, start, mask, leftshift;
    uchar   c;
    ushort  ci;
    
    new_height = preamble->width;
    new_width = old_height = preamble->height;
    old_byte_width = preamble->byte_width;    
    new_byte_width = (new_width + 7) / 8;
    old_raster = preamble->raster;

    if (new_byte_width * new_height <= 2048) {
        for(iy = 0; iy < new_height; iy++) {
            for(ix = 0; ix < new_byte_width; ix++) {
                mask = 0x0080 >> (iy % 8);
                start = (old_height - ix*8 - 1)*old_byte_width + iy/8;
                c = 0;
                for(bit = 0; bit < 8; bit++,start -= old_byte_width)
                    c += (((old_raster[start] & mask) << (iy % 8)) >> bit);
                work[ iy * new_byte_width + ix ] = c;
            }
        }
        if (preamble->rotate_ptex > 1) {
            leftshift = (8 - (new_width % 8)) % 8;
            mask = ((leftshift == 0) ? 0 : (0x00FF >> (8 - leftshift)));
            for(iy = 0; iy < new_height; iy++) {
                for(ix = 0; ix < new_byte_width; ix++) {
                    work2[ new_byte_width - ix - 1 ]
                      = leftside_right(work[ iy * new_byte_width + ix ]);
                }
                for(ix = 0; ix < new_byte_width; ix++) {
                    ci = (ushort)work2[ix];
                    ci <<= 8;
                    ci += work2[ix+1];
                    ci <<= leftshift;
                    ci &= 0xFF00;
                    ci >>= 8;
                    work[ iy * new_byte_width + ix ] = (uchar)ci;
                }
            }
        }
        new_raster = work;
    } else {
        new_raster = old_raster;
    }
    return(new_raster);
}

unsigned char leftside_right(uchar orig)
{
    int     i;
    uchar   work, mask1, mask2;

    work = 0;
    mask1 = 1; mask2 = 0x80;

    for(i=0; i<8; i++) {
        if (orig & mask1) work |= mask2;
        mask1 <<= 1;
        mask2 >>= 1;
    }
    return  work;
}

/* end of all */
