/*
 *  TEX Device Driver  ver 2.02-2.31
 *  (c) copyright 1988, 1989 by TSG, 1990 by SHIMA
 *
 *  buffer.c :
 *  Apr. 15, 1989 : first edition
 *  Jul.  1, 1989 : 2nd edition
 *                      programed by T.IWAI
 *  modfied for non PC-9801 machines by sempa 1992
 *
 *  modified for Minix-386 1992 by obuk, youchan
 */


#include "config.h"
#include "dd.h"
#include <fcntl.h>
#include "err.h"
#include "buffer.h"
#ifdef linux
#include <sys/stat.h>
#endif

/* loadpk.c */
void flush_pk();

extern int f_get_pk;
extern int r_b_f_count;
extern double pxl_per_sp;
extern char *goth;
extern int  k_scale;
extern int  k_base;
extern int  sum_option;

int p_b_f_count;
char *fc_name;
static  int fc_id = ('f'+'c'+'_' + 'i'+'d');

static void buffer_alloc(BUF_INFO*);
static void set_buffer(BUF_INFO*);
void buffer_init(void);
void pk_buf_flush(void);
char* marea(uint);
void clear_buf(BUF_INFO*);
char* dup_string(char*);
FONT_INFO* font_info_alloc(void);
void file_write(int, BUFFER*, int);
void load_fc(void);
void save_fc(void);
int check_sum(void);
static void buf_write(int, BUF_INFO*);
static int buf_read(int, BUF_INFO*);

#define omit_divide(a,b) (((a)/(b)) * (b))

static void set_buffer(BUF_INFO* buf)
{
    buf->end = buf->start + buf->size;
    buf->current = buf->start;
}

static void buffer_alloc(BUF_INFO* buf)
{
    if ((buf->start = (BUFFER*)malloc(buf->size)) == NULL)
        error(NO_MEMORY, "Cannot get Buffer");

    set_buffer(buf);
}


void buffer_init(void)
{
    buffer_alloc(raster_buf_pointer);
    buffer_alloc(pk_buf_pointer);
    buffer_alloc(bitmap_buf_pointer);
    load_fc();
}


void load_fc(void)
{
    BUF_INFO    *bufc;
    struct FONT_INFO_REC font_info;
    FONT_INFO* font;
    int     i, fh;
    uint    size;

    struct  FC_HEAD {
        char        goth_top;
        int         kscale;
        int         kbase;
        int         option;
        int         chksum;
        long        sp_pxl;
        BUF_INFO    bufo[3];
    } fc_head;

    if (fc_name == NULL ||
#ifdef msdos
        (fh = open(fc_name, O_RDONLY|O_BINARY)) == -1) return;
#else
        (fh = open(fc_name, O_RDONLY)) == -1) return;
#endif
    read(fh, &fc_head, sizeof(struct FC_HEAD));

    if (fc_head.goth_top    != *goth
      || fc_head.sp_pxl     != (long)(1.0/pxl_per_sp)
      || fc_head.kscale     != k_scale
      || fc_head.kbase      != k_base
      || fc_head.option     != sum_option
      || fc_head.chksum     != check_sum()) goto quit;

    bufc = raster_buf_pointer;

    for(i = 0; i < 3; bufc++, i++)
        if (bufc->start != fc_head.bufo[i].start) goto quit;

    for(font = first_font_info; font; font = font->next_font) {
        read(fh, &font_info, sizeof(FONT_INFO));
        if ( font_info.font_code != font->font_code
          || font_info.next_font != font->next_font
          || font_info.c    != font->c
          || font_info.name != font->name) goto quit;
    }

    lseek(fh, (long)(sizeof(struct FC_HEAD)), SEEK_SET);

    for(font = first_font_info; font; font = font->next_font) {
        read(fh, &font_info, sizeof(FONT_INFO));
        for(i = 0; i < CHAR_ROOT ; i++) 
            font->char_info[i] = font_info.char_info[i];
        font->f_goth = font_info.f_goth;
        font->k_top  = font_info.k_top;
        font->k_next = font_info.k_next;
    }

    raster_buf_pointer->current = fc_head.bufo[0].current;
    
    if (buf_read(fh, raster_buf_pointer) != 0
     || (read(fh, &i, sizeof(int)), i != fc_id)) {
        error(WARNING, "Bad file %s\n", fc_name);
        goto quit;
    }
    fprintf(stderr, "Loading Font Cache\n");
quit:   close(fh);
}


static int buf_read(int fd, BUF_INFO* buf)
{
    long        total;
    BUFFER*     ptr;
    uint        size;
    int         cond;

    total  = (BUFFER *)(buf->current) - (BUFFER *)(ptr = buf->start);
    while(total > 0L) {
        size = (total > (unsigned)0xf000)?0xf000:total;
        if ((cond = size - file_read(fd, ptr, size)) != 0)
            return(cond);
        total -= size;
        ptr   += size;
    }
    return(0);
}


int check_sum(void)
{
    FONT_INFO*  font;
    int         i, chksum;
    uchar       *adr;

    chksum = 0;
    for(font = first_font_info; font; font = font->next_font) {
        if (*(adr = (uchar*)font->name) == '*') {
            for(i = 9; i > 0; i--) chksum += *adr++;
            do{
                chksum += *adr;
            }while(*adr++);
        }
        while(*adr) chksum += *adr++;
    }
    return(chksum);
}

void save_fc(void)
{
    int         fh, i, chksum;
    FONT_INFO*  font;
    struct  FC_HEAD0    {
        char        goth_top;
        int         kscale;
        int         kbase;
        int         option;
        int         chksum;
        long        sp_pxl;
    } fc_head0;

    if (fc_name == NULL
     || *fc_name == 0
     || f_get_pk == 0
     || r_b_f_count != 0
   ) return;
    for(i = 0; ; i++) {
#ifdef msdos
        if ((fh = open(fc_name, O_RDWR|O_TRUNC|O_CREAT|O_BINARY, S_IREAD|S_IWRITE)) > 0)
#else
        if ((fh = open(fc_name, O_RDWR|O_TRUNC|O_CREAT, S_IREAD|S_IWRITE)) > 0)
#endif
            break;
        if (i == 1) return;
        flush_pk();
    }

    fc_head0.goth_top = *goth;
    fc_head0.kscale = k_scale;
    fc_head0.kbase = k_base;
    fc_head0.option = sum_option;
    fc_head0.chksum = check_sum();
    fc_head0.sp_pxl = (long)(1.0/pxl_per_sp);

    write(fh, &fc_head0, sizeof(struct FC_HEAD0));
    write(fh, raster_buf_pointer, sizeof(BUF_INFO)*3);

    for(font = first_font_info; font; font = font->next_font)
        file_write(fh, (BUFFER *)font, sizeof(FONT_INFO));

    buf_write(fh, raster_buf_pointer);
    write(fh, &fc_id, sizeof(int));
    fprintf(stderr, "\nSaved Font Cache");

    close(fh);
}

static void buf_write(int fd, BUF_INFO* buf)
{
    long        total;
    BUFFER*     ptr;
    uint        size;

    total  = (BUFFER *)(buf->current) - (BUFFER *)(ptr = buf->start);
    while(total > 0L) {
        size = (total > (unsigned)0xf000)?0xf000:total;
        file_write(fd, (BUFFER*)ptr, size);
        total -= size;
        ptr   += size;
    }
}

void file_write(int fd, BUFFER* ptr, int size)
{
    static char bad_fc[2] = {0xff, 0xff};
    
    if (write(fd, ptr, size) != size) {
        lseek(fd, 0L, SEEK_SET);
        write(fd, bad_fc, 2);
        error(PROGRAM_STOP, "Cannot write font cache");
    }
}

void pk_buf_flush(void)
{
    FONT_INFO *font;
    
    for (font = first_font_info; font; font = font->next_font) 
        font->pk = NULL;
    pk_buf_pointer->current = pk_buf_pointer->start;
    p_b_f_count++;
}


FONT_INFO* font_info_alloc(void)
{
    FONT_INFO* font;

    if ((font = (FONT_INFO *)malloc(sizeof(FONT_INFO))) == NULL)
        error(NO_MEMORY, "Cannot get FONT Buffer");
    
    return(font);
}


char* marea(uint size)
{
    char* ptr;

    if (size == 0) size = 1;
    if ((ptr = (char*)malloc(size)) == NULL)
        error(NO_MEMORY, "Not enough heap area");
    
    return(ptr);
}


char* dup_string(char* src)
{
    char* dst;
    
    if ((dst = (char*)strdup(src)) == NULL)
        error(NO_MEMORY, "Not enough heap area");
    return(dst);
}


void clear_buf(BUF_INFO* buf)
{
    memset(buf->start, 0, buf->size);
}

/* end of file : buffer.c */
