/*
 *  TEX Device Driver  ver 2.02-2.30
 *  copyright(c) 1988,89 by TSG, 1990 by SHIMA
 *
 *  dd.h : common header file
 *
 *  need to include <stdio.h> previously
 *
 *  changed for Minix-386 1992 by obuk, youchan
 */

/* include files */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>

/* standard definitons */
typedef enum { FALSE = 0, TRUE = 1 } BOOL;
#ifdef linux
typedef unsigned char uchar;
#else
#define uchar   unsigned char
#define uint    unsigned int
#define ushort  unsigned short
#define ulong   unsigned long
#endif
typedef void (*void_func_ptr)();
typedef int  (*int_func_ptr)();

#ifdef msdos
#define MAXPATH 260
#define movedata((x),(y),(z))  memcpy((y),(x),(z))
#define getch()                fgetc(stdin)
#define strcmpi((x),(y))       strcmp((char*)strupr((x)),(char*)strupr((y)))
#define _read((x),(y),(z))     read((x),(y),(z))
#define DELIMITER   ';'
#define PATHCHAR    '\\'
#define PATH_PREFIX "\\"
long filelength(int);
typedef struct {
    int selecter;
    int paragraph;
    unsigned char *address;
} LOW_MEM;
LOW_MEM *alloc_low(int);
#else  /* unix */
#define DELIMITER   ':'
#define PATHCHAR    '/'
#define PATH_PREFIX "/"
#endif

#define TRUE    1
#define FALSE   0
#define NOTHING 0
#define FAILURE (-1)

#ifndef MAXPATH
#define MAXPATH PATH_MAX
#endif

#ifndef linux
#define S_IREAD  0444
#define S_IWRITE 0222
#endif

#ifndef SEEK_SET
    #define SEEK_SET    0
    #define SEEK_CUR    1
    #define SEEK_END    2
#endif

#if defined(LIPS3) || defined(ESCPAGE)
#define LBP
#define FDOWN
#endif

/* definitons for device driver */

#define BUFFER    uchar
    /*   Both BUFFER pointer and BUFFER pointer are 32bit-pointer.
     *   BUFFER will be used for larger buffers ( correctly speaking, 
     * ones which keep more data than 64Kbytes ).
     *   And it is noteworthy that the pointers of these types are impossible
     * to express continually ( ex. BUFFER *a, *b ), because they are not
     * "typedef" but "#define."  This foolish error is because of 8086CPU,
     * which has only 16bit-pointer, so we must add "far" or "huge" for use of
     * 32bit-pointer. However, "far" and "huge" are permitted to "typedef." 
     * How wonderful SEGMENT !!
     */

#define SIZE_PARA   double
    /*   This device driver calculate units with double.
     */

#define F_SIZE_PARA   long
    /*   This device driver calculate units with double.
     */

#define PIXEL       int
#define SCALED_PT   long
    /*   the units of pixel and scaled-point
     */

#define HEADER_DEPTH    0xe0000L
#define FOOTER_DEPTH    0xe0000L
    /*   the header depth and the footer depth ( scaled point )
     */

typedef enum {
    UNKNOWN, PK_FONT, PXL1001, PXL1002, PXL1003, PKD_FONT, JXL, ZS_FONT,
#ifdef BEZIERFONT
    JG_FONT,
#endif
    JLBP, JIS_FONT, JIS_PK
} FONT_TYPE;
    /*   FONT_TYPE shows the type of fonts */

typedef enum { HORIZONTAL, VERTICAL } PRINT_DIR;
    /*   Direction at printing, HORIZONTAL normally.
     */

typedef enum {
    PreviousPart, NextPart,
    PreviousPage, NextPage,
    NextPageDirection,
    EscapeQuit
} NEXT_ACTION;
    /*   These values are returned by device-output module, which says
     * what to do next.
     */

#define PREAMBLE struct PREAMBLE_REC
    /*   Information about Preamble of a font-file
     */
struct PREAMBLE_REC {
    PIXEL width, height, x_offset, y_offset;
        /*   the size of the character, and the offset for typesetting.
         */
    BUFFER* raster;
        /*   the pointer of Raster-data buffer
         */
    long tfm_width;
    int byte_width;
        /*   equal width/8, raised decimal
         */
    PIXEL shift_up_ptex, shift_right_ptex;
        /*   the offset when lateral-kanji used (when in pTeX tate mode).
         */
    int rotate_ptex;
        /*   the flag of rotation (when in pTeX tate mode).
         */
};


#define CHAR_INFO struct CHAR_INFO_REC
    /*   This buffer is a linked-list one, which keeps pointers of unpacked
     * fonts.
     */
struct CHAR_INFO_REC {
    int code;
        /*   character code
         */
    CHAR_INFO* next_char;
    CHAR_INFO* former_char;
        /*   needless to say? */
    struct PREAMBLE_REC pre;
        /*   preamble of an unpacked font
         */
};

#define f_HUGE  0x8000

#define BUF_INFO  struct BUF_INFO_REC
    /*   Information about buffers
     */
struct BUF_INFO_REC {
    BUFFER* start; 
    BUFFER* current;
    BUFFER* end;
        /*   Each means starting, current, and ending pointer of the buffer.
         * Especially, 'end' means 'start+size', so the pointer above ( or 
         * which equals )'end' will be out of the buffer.
         */
    long size;
        /*   'size' can have "larger" data ( above segment, hum !! ).
         */
    int type;
};


typedef struct {
    long offset;
    long number;
} PAGE_INDEX;
    /* page index buffer */

#define DVIFILE_INFO struct DVIFILE_INFO_REC
    /*   Information about Preamble of DVI-file. ( to say the truth, this is
     * the same thing as Preamble itself. So, you'd better see 
     * "TeX : The Program." )
     */
struct DVIFILE_INFO_REC {
    FILE* file_ptr;
    char* file_name;
    long post, last_bop;
    long num, den, mag;
    SCALED_PT l,u;
    int stack_depth;
    int total_page;
};

#define DIMENSION struct DIMENSION_REC
    /*  Information about device-output
     */

struct DIMENSION_REC {
    int dpi;
        /* is dot-per-inch ratio of the device.
         */
    PIXEL max_width, max_height;
        /*   These tell us the maximum size the device can output.
         * They are represented by pixel-unit.
         */
    PIXEL buf_width, buf_height;
        /*   These tell us the maximum size Bitmap-buffer can hold.
         */
    int mag;
        /*  is the magnification for output. This is the same value as 'mag' in
         * DVI-file ( so equal the one in DVI_INFO ), when no '-mag' & '-half'
         * options directed.
         */
    PIXEL text_width, text_height;
        /*   These are the maxmum size directed in DVI-file.
         */
    int split;
        /*   Our Device-driver divides the text to several parts, if that can't
         * be output at once. 'split' is the number of parts. Then, if 'split'
         * holds '1', Device-driver will put the text at a time.
         */
    PRINT_DIR print_direction;
        /*   This shows the direction for output, responsing '-V' option. 
         * So there is no meaning for PREVIEWER.
         */
    SIZE_PARA   size_para;
        /* is size-parameter for conversion pixel<-->scaled pt.
         * ( equal "num * den / mag" )
         */
    PIXEL x_offset, y_offset;
        /*   the offset of the head of text. These are directed by '-X' and
         * '-Y' options.
         */
    PAGE_INDEX* page_index;
        /* the index of file-offset of pages in dvifile
         */
    int total_page;
        /* same as 'total_page' in DVI_INFO.
         */
    int start_page, end_page;
        /* the starting and the ending page to print.
         */
    int prt_type;
        /* the type of the printer, ESC/P, PC-PR, PC-MN, LIPS3 
         * or Previewer etc
         */
};

#define OUTPUT_INFO struct OUTPUT_INFO_REC
    /*   Information about output. Interpreting-module and Device-output-module
     * receive the pointer of this.
     */
struct OUTPUT_INFO_REC {
    FILE* dvifile_ptr;
        /* DVI-file pointer
         */
    long page_start_offset;
        /* the file-offset at the position the page starts from.
         */
    PRINT_DIR print_direction;
        /* equal the one in DIMENSION.
         */
    PIXEL h_0, v_0;
        /* the original point for output.
         */
    PIXEL width, height;
        /* the size for output.; it is assigned for all the size of the text
         * to be held in the buffer.
         */
    int byte_width, byte_height;
        /*  equal width/8 and height/8.
         */
    PIXEL shift_up_ptex, shift_right_ptex;
        /*  the offset when lateral-Kanji used (when in pTeX tate mode)
         */
    int rotate_ptex;
        /* the flag of rotation (when in pTeX tate mode)
         */
    int split, page;
        /* the current part and page.
         */
    BUFFER* bitmap_ptr;
        /* Bitmap pointer 
         */
};

typedef struct {                    /* KANJI fonts */
    int     width;          /* width by dots  (without counting K_DBL) */
    int     height;         /* height by dots (without counting K_DBL) */
    int     size;           /* size by bytes = (width+7)/8*height */
    int     base;           /* number of dots under base line */
    int     fh;             /* file handle number */
    char    *name;          /* Kanji font file_name */
} KFONT;

typedef struct {
    int   char_type;
    int   id;
    long  width;  /* character width */
    long  height; /* character height(length above the baseline) */
    long  depth;  /* character depth (length under the baseline) */
    long  italic;
} JFM_DATA;
  
#define KFONT_SIZE  sizeof(KFONT)

#define FONT_INFO struct FONT_INFO_REC
    /*   Information about fonts. This is linked-list.
     */

union EXT {
        BUFFER* dir;
        /*  address of the directory for PXLFONT
         */
        KFONT*  kdir;
        int     fh;
        long    count;
};

#define CHAR_ROOT   16          /* should be a power of 2 */

struct FONT_INFO_REC {          /* 4*CHAR_ROOT + 46 + 4 = 114 bytes (FDOWN) */
    int font_code;
        /*   sent from dvi-file, when the font is defined.
         */
    FONT_TYPE font_type;
        /*   type of font.
         */
    int f_goth;
        /* flag for gothic etc.
         */
    char* name;
        /*   the file-name of the font.
         */
    BUFFER* pk;
        /*   the pointer of unpacked data of the font.
         * it is NULL, when data is unloaded or flushed.
         */
    CHAR_INFO* char_info[CHAR_ROOT];
        /*   the pointers of the first CHAR_INFO of the font.
         * Also, it is NULL, when data is unloaded or flushed.
         */
    FONT_INFO*  k_top;
    FONT_INFO*  k_next;

    F_SIZE_PARA size_para;
        /*  design size.
         */
    FONT_INFO* next_font;
        /*   equal NULL, when no next font.
         */
    FONT_INFO* before_font;
        /*   equal NULL, when no former font
         */
    FONT_INFO* after_font;
        /*   equal NULL, when no after font
         */
    int k_width;
        /*   normal allowable maximal width of Kanji character by bits
         *   counting k_scale
         */
    int last;
        /* last JIS CODE of Kanji font
         */
    long    c;
        /* checksum (id)
         */
    union EXT ext;
#ifdef FDOWN
    uchar *d_flag;
#endif
    int   dummy;
};

#ifdef LBP
#ifdef FDOWN
typedef struct {
    int assign_num;
    int maxwidth;
    int maxheight;
    int mincharcode;
    int maxcharcode;
    long total_send;
    int numoffonts;
} DL_INFO;
#endif /*FDOWN*/
typedef struct {
    void (*init)(void);
    void (*init2)(int, int);
    void (*end)(void);
    void (*set_font)(FONT_INFO*, KFONT*, int, int, int);
    void (*move_cap)(PIXEL, PIXEL);
    void (*move_homepoint)(PIXEL,PIXEL);
    void (*pre_put_char)(FONT_INFO*, uint*);
    void (*put_jis_font)(uint);
    void (*put_dl_font)(uint);
    long (*get_dlhsize)(int);
    void (*pre_download)(FONT_INFO*, DL_INFO*, int, int);
    void (*download_raster)(FONT_INFO*, DL_INFO*, uint, int);
    void (*post_download)(FONT_INFO*, DL_INFO*);
    void (*drawbox)(PIXEL, PIXEL, PIXEL, PIXEL, int);
    struct LIPS_LFONT *data;
} LBP_FUNC;

struct LIPS_LFONT {
    char *font_name;
    int no;
};
#endif  /*LBP*/

/* extern variables */

extern FONT_INFO* first_font_info;
    /*  the first position of the font-list.
     */
extern BUF_INFO* pk_buf_pointer;
    /*  the buffer for pk-font-data loaded from disk directly.
     */
extern BUF_INFO* raster_buf_pointer;
    /*  the buffer for the raster-data of unpacked fonts.
     */
extern BUF_INFO* bitmap_buf_pointer;
    /*  Virtual VRAM buffer. Interpret-module will draw the figure onto this.
     * This is casted to BUFFER when used.
     */

#define F_GOTH    1
#define K_DBL     2
#define F_LBP     4
#define F_DWN     8
#define F_FILL    0x10
#define F_SHADOW  0x20
#define F_OUTLINE 0x40
#define F_GGOTH   0x80

#define F_ATTR    (F_GOTH|F_FILL|F_SHADOW|F_OUTLINE|F_GGOTH)

/* Definition for DVI commands */

#define SET_CHAR_0      0
#define SET1            128
#define SET_RULE        132
#define PUT1            133
#define PUT_RULE        137
#define NOP             138
#define BOP             139
#define EOP             140
#define PUSH            141
#define POP             142
#define RIGHT1          143
#define W0              147
#define W1              148
#define X0              152
#define X1              153
#define DOWN1           157
#define Y0              161
#define Y1              162
#define Z0              166
#define Z1              167
#define FNT_NUM_0       171
#define FNT1            235
#define XXX1            239
#define FNT_DEF_1       243
#define PRE             247
#define POST            248
#define POST_POST       249
#define EOFNC           255  /* end of func */

/* end of file : dd.h */

