/*
 *  TEX Device Driver  ver 2.02-2.29
 *  copyright(c) 1988,89 by TSG., 1990 by SHIMA
 *
 *  decodepk.c :
 *  Oct.  6, 1988 : first edition
 *                     programed by K.Miyashita
 *  Apr. 15, 1989 : second edition
 *  Jul.  2, 1989 : third edition
 *                     programed by T.IWAI
 *
 *  modified for Minix-386 1992 by obuk, youchan
 */


#include "config.h"
#include "dd.h"
#include "err.h"
#include "buffer.h"

#define MAX_DIV_8 1000
    /*  temp-buf size */
#define IS_EXP2 0x40

static int power[8]=
{   0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80  };

static int gpower[9]=
{   0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff };

#define pixel_integer_8(n) (*raster++=(uchar)(n))

static int bit_weight;
static int dyn_f;
static int c_width, c_height, byte_width;
static int input_byte;
    /*  used in function 'get_bit' and 'get_nyb' */
static int repeat_count;

static int lastrep;

static int expf;

static BUFFER* pk_data;
static BUFFER* raster;
static uchar row[MAX_DIV_8];

void decode_pk(BUFFER*, PREAMBLE*, int, FONT_INFO*);
void decode_jxl(BUFFER*, PREAMBLE*, int, FONT_INFO*);

static int get_nyb(void);
static void raster_by_bits(void);
static BOOL normal_packed_raster(BOOL);
static BOOL jxl_packed_raster(void);
static uint get_pknum(void);
static void send_row(void);


#define get_nyb() ((bit_weight^=1)?((input_byte=get_ubyte(pk_data))>>4):(input_byte & 0x0f))

void decode_pk(BUFFER *pk, PREAMBLE *preamble, int flag_byte, FONT_INFO *font)
{
    BOOL turn_on;
    
    bit_weight = 0;
    dyn_f = flag_byte >> 4;
    turn_on = ((flag_byte & 8) != 0);
    pk_data = pk;
    if (font->f_goth & K_DBL) {
        c_width = preamble->width/2;
        c_height = preamble->height/2;
        byte_width = (preamble->byte_width+1)/2;
    }else{
        c_width = preamble->width;
        c_height = preamble->height;
        byte_width = preamble->byte_width;
    }
    raster = preamble->raster;
    
    if (dyn_f == 14)
        raster_by_bits();
    else
        if (normal_packed_raster(turn_on))
            error(UNPACK_ERROR, "More bits than required.\n"
                "%5d: %s", font->font_code, get_font_name(font));
}


void decode_jxl(BUFFER *pk, PREAMBLE *preamble, int flag_byte, FONT_INFO *font)
{
    BOOL turn_on;

    bit_weight = 0;
    dyn_f = flag_byte & 0xf;
    turn_on = ((flag_byte & 0x80) != 0);
    pk_data = pk;
    c_width = preamble->width;
    c_height = preamble->height;
    byte_width = preamble->byte_width;
    raster = preamble->raster;

    if (dyn_f == 14) raster_by_bits();
    else
        if (((flag_byte & IS_EXP2) == 0)?
            normal_packed_raster(turn_on):jxl_packed_raster())
                error(UNPACK_ERROR, "More bits than required.\n"
                    "%5d: %s", font->font_code, get_font_name(font));
}

void raster_by_bits(void)
{
    int i, j;
    int shift, mode;

    if ((mode = (c_width & 0x7)) == 0) {
        for(i = byte_width*c_height; i > 0; i--) {
            *raster++ = *pk_data++;
        }
        return;
    }
    shift = 0;
    for(j = 0; j < c_height; ++j) {
        for(i = byte_width-1; i > 0; i--) {
            *raster = *pk_data << shift;
            *raster++ |= *(++pk_data) >> (8 - shift);
        }
        *raster = *pk_data << shift;
        if (mode + shift >= 8) {
            *raster |= *(++pk_data) >> (8 - shift);
        }
        *raster++ &= 0xff << (8-mode);
        shift = ((shift + mode) & 0x7);
    }
}


static BOOL jxl_packed_raster(void)
{
    int r_p, h_bit, rows_left, byte_weight, i, j;
    uint count;

    rows_left = c_height;
    h_bit = c_width+1;
    r_p = repeat_count = 0;
    byte_weight = expf = 1;
    
    for(i = 1; i <= c_width; i++) row[i] = 0;
    for(; ;) {
        count = get_pknum();
        while(count >= h_bit) {            /* skip to next line */
            for(i = 1; i <= byte_width; i++)
                pixel_integer_8(row[i]);
            if (--rows_left <= 0)  return(0);
            count -= h_bit;
            r_p = 1;
            h_bit = c_width;
            byte_weight = 8;
        }
        while(count >= byte_weight) {      /* skip to next byte */
            count -= byte_weight;
            h_bit -= byte_weight;
            r_p++;
            byte_weight = 8;
        }                                   /* In this byte */

        byte_weight -= count;                       /* should be positive */
        h_bit -= count;                             /* should be positive */
        repeat_count++;
        while(repeat_count >= h_bit) {     /* next ptr is next row */

            if (byte_weight < h_bit) {              /* not the last byte */
                row[r_p++] ^= gpower[byte_weight];
                h_bit -= byte_weight;
                repeat_count -= byte_weight;

            }else{                                  /* last byte in the row */
                row[r_p] ^= (gpower[byte_weight]-gpower[byte_weight-h_bit]);
                for(i = 1; i <= byte_width; i++)
                    pixel_integer_8(row[i]);
                if (--rows_left <= 0) return(0);
                r_p = 1;
                repeat_count -= h_bit;
                h_bit = c_width;
            }
            byte_weight = 8;
        }
                                            /* next ptr is in this row */
        while(repeat_count) {
            if (byte_weight > repeat_count) {   /* next ptr is in this byte */
                row[r_p] ^= 
                    (gpower[byte_weight]-gpower[byte_weight-repeat_count]);
                byte_weight -= repeat_count;
                h_bit -= repeat_count;
                repeat_count = 0;
                break;
            }
            else {                      /* next ptr is not in this byte */
                row[r_p++] ^= gpower[byte_weight];
                h_bit -= byte_weight;
                repeat_count -= byte_weight;
                byte_weight = 8;
            }
        }
    }
}

static BOOL normal_packed_raster(BOOL turn_on)
{
    int r_p, h_bit, rows_left, byte, byte_weight;
    uint  count;
    
    rows_left = c_height;
    h_bit = c_width;
    byte = repeat_count = expf = 0;
    byte_weight = 8;
    r_p = 1;

    while (rows_left > 0) {
        count = get_pknum();
        while (count > 0) {
            if ((count < byte_weight) && (count < h_bit)) {
                if (turn_on)
                    byte |= (gpower[byte_weight] -
                        gpower[byte_weight-count]);
                
                h_bit -= count;
                byte_weight -= count;
                break;
            }
            else if ((count >= h_bit) &&
                (h_bit <= byte_weight)) {
                if (turn_on)
                    byte |= (gpower[byte_weight] -
                        gpower[byte_weight-h_bit]);
                
                row[r_p] = byte;
                send_row();
                rows_left = rows_left - repeat_count - 1;
                repeat_count = 0;
                r_p = 1;
                byte = 0;
                byte_weight = 8;
                count -= h_bit;
                h_bit = c_width;
            }
            else {
                if (turn_on)
                    byte |= gpower[byte_weight];
                
                row[r_p++] = byte;
                byte = 0;
                count -= byte_weight;
                h_bit -= byte_weight;
                byte_weight = 8;
            }
        }
        turn_on = (! turn_on);
    }
    return ((rows_left != 0) || (h_bit != c_width));
}



static uint get_pknum(void)
{
    uint i;
    uint j;

    i = get_nyb();
    if (i == 0) {
        do {
            j = get_nyb();
            i++;
        } while (j == 0);
        if (i > 4) return(0xfff0);
        while (i > 0) { 
            j = (j << 4) + get_nyb();
            i--;
        }
        return(j - 15 + ((uint)(13 - dyn_f) << 4) + dyn_f);
    }
    else if (i <= dyn_f)
        return(i);
    else if (i < 14)
        return(((i - dyn_f - 1) << 4) + get_nyb() + dyn_f + 1);
    else {
        if (i == 14)
            lastrep = repeat_count = get_pknum();
        else repeat_count =  (expf==0)?1:lastrep;
        return(get_pknum());
    }
}

static void send_row(void)
{
    int tmp;
    int h;
    
    for (tmp = 0; tmp <= repeat_count; tmp++)
        for (h = 1; h <= byte_width; h++)
            pixel_integer_8(row[h]);
}

/* end of file : decodepk.c */
