/*
 *  TEX Device Driver  ver 2.02-2.36
 *  copyright(c) 1988,89 by TSG, 1990 SHIMA
 *
 *  modified for Minix-386, Linux by youchan  1992
 */

#include "config.h"
#include "dd.h"
#ifdef minix
#include <lib.h>
#endif

const char* const title_comment = "";

#define ONEBYTE     8

#define ON          1
#define OFF         0
#define ESC         0x1b
#define V_SHIFT     1

extern  int  num_view;
extern  int  num_view_buf;
extern  int  unit_pages;
extern  char resume_info[16];

#define MAX_VIEW    10

struct  VIEW {
    int current;
    int total;
    int page[MAX_VIEW];
};

static struct VIEW view;

struct SCROLL {
    int  x;
    int  y;
    int  gw_act, gh_act;
    BOOL x_lock, y_lock;
    BUFFER* mptr;
    int h_spc, v_spc;
};

typedef enum {
    BackSpace, SpaceKey,    /* PreviousPage, NextPage   */
    PgUp, PgDn,             /* PreviousPage, NextPage   */
    CaridgeReturn,          /* NextPageDirection        */
    EndKey,                 /* EscapeQuit               */
    Key_Q,                  /* Same as above            */
    Key_B,                  /* Another Previous Page    */
    Key_H,                  /* HELP                     */
    Key_M,                  /* Move to a page           */
    Key_G,                  /* Same as above            */
    Key_P,                  /* PreviousPage             */
    Key_N,                  /* NextPage                 */
    Help,
    ArrowDown, ArrowUp, ArrowLeft, ArrowRight, LoopEnd
} KeyInput;

int s_max_width  = 118*8;
int s_max_height = 118*11;

static BUFFER* map_top_ptr;         /* pointer of bitmap buffer */
static PIXEL   vmax, hmax, hmaxb;   /* size to output. hmaxb is byte-size. */
static BUFFER* g_top_ptr;           /* head pointer of virtial graphic vram */
static int  f_div;
static int  f_div2;
static BUFFER* view_buf;
static BUFFER* g_top;

int x_shift;
int y_shift;

/* Proto-types */
void device_init(DIMENSION*);
void device_end(void);
void device_cont(void);
void device_pause(void);
void device_clear(OUTPUT_INFO*);
void pr_new_page(void);
NEXT_ACTION device_out(OUTPUT_INFO*);
KeyInput extra_inkey(void);

static void first_draw(struct SCROLL*, int);
static void draw_screen(struct SCROLL*, int);
static void move_screen(struct SCROLL*, KeyInput, int, int);

static void flskey(void);
static void msg_help(void);

extern int tt_flag;
extern void ttclose(void);

/* graphics driver */
#if   defined minix      /* MINIX-386 */
#ifndef PCAT
#include "minix.drv"
#else
#include "minixv.drv"
#endif
#elif defined linux      /* Linux */
#include "linux.drv"
#endif

void error_exit(char *msg)
{
    gr_end();
    if (tt_flag != 0) ttclose();
    fprintf(stderr,msg);
    exit(1);
}

KeyInput extra_inkey(void)
{
    int        num   = 0;
    int        f_num = 0;
    int        k_dat, c;
    KeyInput   key_d;

    flskey();
    for (;;) {
        k_dat = getch();
        if (k_dat >= 0x30 && (uchar)k_dat <= 0x39) {
            f_num = 1;
            num -= (num/100)*100;
            num = num*10 + k_dat - '0';
            if (num >= 1) unit_pages = num;
            continue;
        }
        else if (k_dat == 0x1b) {
            if (getch() != 0x5b) {
                flskey();
                continue;
            }
            c = getch();
            if (c == 0x35 || c == 0x36 || c == 0x31 || c== 0x32) {
                if (getch() != 0x7e) {
                    flskey();
                    continue;
                }
            }
            switch (c) {
#if defined linux
              case 0x35: key_d = PgUp;
                         break;
              case 0x36: key_d = PgDn;
                         break;
              case 0x48: key_d = Help;
                         break;
#elif defined minix
              case 0x56: key_d = PgUp;
                         break;
              case 0x55: key_d = PgDn;
                         break;
              case 0x47: key_d = Help;
                         break;
#endif
              case 0x41: key_d = ArrowUp;
                         break;
              case 0x42: key_d = ArrowDown;
                         break;
              case 0x43: key_d = ArrowRight;
                         break;
              case 0x44: key_d = ArrowLeft;
                         break;
              case 0x59: key_d = EndKey;
                         break;
              default:   key_d = LoopEnd;
           }
       }
       else {
           switch (k_dat) {
#if defined linux
             case 0x7f: key_d = BackSpace;
                        break;
             case 0x0a: key_d = CaridgeReturn;
                        break;
#elif defined minix
             case 0x08: key_d = BackSpace;
                        break;
             case 0x0d: key_d = CaridgeReturn;
                        break;
#endif
             case 0x20: key_d = SpaceKey;
                        break;
             case 0x71: key_d = EndKey;
                        break;
             case 0x62: key_d = Key_P;
                        break;
             case 0x68: key_d = Key_H;
                        break;
             case 0x6d: key_d = Key_M;
                        break;
             case 0x67: key_d = Key_G;
                        break;
             case 0x70: key_d = Key_P;
                        break;
             case 0x6e: key_d = Key_N;
                        break;
             default:   key_d = LoopEnd;
           }
       }
       if (key_d == LoopEnd)   continue;
       if ((key_d == Key_M || key_d == Key_G) && f_num == 0) continue;
       break;
    }
    flskey();
    return(key_d);
}

void pr_new_page()
{
    /* dummy */
}

void device_init(DIMENSION* dim)
{
    int g;

    g = gr_init();
    if (g < 0) {
        fprintf(stderr,"Graphics system can't initialize.\n");
        exit(1);
    }
    if ((g_top = (BUFFER*)malloc(_g_width_b * _g_height)) == NULL)
        error_exit("virtial memory exhausted.\n");
}

void device_clear(OUTPUT_INFO* out)
{
    map_top_ptr = out->bitmap_ptr;
    hmax = out->width;
    vmax = out->height;
    hmaxb = out->byte_width;
}

void device_end()
{
    gr_end();
}

void device_pause()
{
    /* dummy */
}


void device_cont()
{
    /* dummy */
}

NEXT_ACTION device_out(OUTPUT_INFO* out)
{
    struct SCROLL s_dat;
    KeyInput key;
    int i;

    s_dat.mptr = map_top_ptr;
    first_draw(&s_dat, out->page);

    for(;;) {
        key = extra_inkey();
        switch(key) {
            case Key_H:
            case Help:   msg_help();
                         s_dat.mptr = map_top_ptr;
                         first_draw(&s_dat,out->page);
                         break;

            case Key_G:
            case Key_M:  unit_pages -= out->page;
                         return(NextPage);

            case BackSpace:
            case Key_P:
            case PgUp:   return(PreviousPage);

            case SpaceKey:
            case Key_N:
            case PgDn:   return(NextPage);

            case CaridgeReturn:
                         return(NextPageDirection);

            case EndKey:
                         return(EscapeQuit);

            case ArrowDown:
            case ArrowUp:
                         move_screen(&s_dat, key, 32, out->page);
                         break;

            case ArrowLeft:
            case ArrowRight:
                         move_screen(&s_dat, key, 4, out->page);
                         break;
        }
    }
}

static void first_draw(struct SCROLL* s_dat,int page)
/* at first, set variables and  put to screen */
{
    s_dat->x_lock = s_dat->y_lock = FALSE;
    if (hmaxb < _g_width_b) {
        s_dat->gw_act = hmaxb;
        s_dat->x_lock = TRUE;
    } else {
        s_dat->gw_act = _g_width_b;
    }

    if (vmax < _g_height) {
        s_dat->gh_act = vmax;
        s_dat->y_lock = TRUE;
    } else {
        s_dat->gh_act = _g_height;
    }

    s_dat->h_spc = (_g_width_b - s_dat->gw_act)/2;
    s_dat->v_spc = (_g_height  - s_dat->gh_act)/2;

    g_top_ptr = g_top+s_dat->h_spc+s_dat->v_spc*_g_width_b;

    s_dat->x = -s_dat->h_spc;
    s_dat->y = -s_dat->v_spc;

    draw_screen(s_dat,page);
}

static void draw_screen(struct SCROLL* s_dat, int page)
{
    int     i;
    BUFFER* mptr_tmp;
    BUFFER* gptr_tmp;

    gptr_tmp = g_top_ptr;
    mptr_tmp = s_dat->mptr;
    for(i=0;i<s_dat->gh_act;i++) {
        memcpy(gptr_tmp, mptr_tmp, s_dat->gw_act);
        gptr_tmp += _g_width_b;
        mptr_tmp += (ulong)hmaxb;
    }
    gr_screen_write((char*)g_top);
    page_no_out(page);
}

static void move_screen(struct SCROLL* s_dat, KeyInput direct, int step, int page)
{
    int x_step, y_step;
    
    x_step = y_step = 0;

    switch(direct) {
        case ArrowDown :  if (s_dat->y_lock == FALSE) y_step = step;
                          break;
        case ArrowUp :    if (s_dat->y_lock == FALSE) y_step = -step;
                          break;
        case ArrowLeft :  if (s_dat->x_lock == FALSE) x_step = -step;
                          break;
        case ArrowRight : if (s_dat->x_lock == FALSE) x_step = step;
                          break;
    }
    
    if (x_step != 0) {
        if (s_dat->x + x_step <= 0)
            x_step = - s_dat->x;
        if (s_dat->x + x_step + _g_width_b >= hmaxb)
            x_step = hmaxb - (s_dat->x + _g_width_b);
        s_dat->x += x_step;
        s_dat->mptr += x_step;
    }

    if (y_step != 0) {
        if (s_dat->y + y_step <= 0)
            y_step = - s_dat->y;
        if (s_dat->y + y_step + _g_height >= vmax)
            y_step = vmax - (s_dat->y + _g_height);
        s_dat->y += y_step;
        s_dat->mptr += y_step * hmaxb;
    }

    if (x_step != 0 || y_step != 0) {
        draw_screen(s_dat,page);
    }
}

static void flskey(void)
{
    fflush(stdin);
}

static void msg_help(void)
{
    int i;

    gr_end();
    fprintf(stderr, 
        "\n"
        "                    <<< Help >>>\n\n"
        " Arrows                : Scroll in a page\n"
        " Space, N, PgDn        : Next page\n"
        " BackSpace, P, B, PgUp : Previous page\n"
        " G, M                  : Move to a page\n"
        " numbers               : pages (skipped) by next B, G, M, N, P\n"
        " H                     : Help message\n"
        " Return                : Next block or quit\n"
        " Q, END                : Quit\n\n"
        " --> Hit SPACE Key\n"
    );
    flskey();
    while (extra_inkey() != SpaceKey) ;
    gr_init();
}

/* end of device.c */




