/*
 *  Routine to use scalable/downloaded font of LIPS3 (dviprt)
 *                        30 June    1992, written by SHIMA
 *  modifed for ESC/Page by H.Tomiie, 18 August 1992
 *  slightly modifed by OkI 14 Nov. 1992
 */

#include "config.h"

#ifdef ESCPAGE
#include "dd.h"
#include "prtout.h"
#include "err.h"
#include "prtctl.h"
#include "escpage.h"

/* escpage.c local */
static void pr_attr(int flag);
static int KetaSuu(PIXEL num);

/* p_out.c function */
static void escpage_init(void);
static void escpage_init2(int, int);
static void escpage_end();
static void set_escpage_font(FONT_INFO*, KFONT*, int, int, int);
static void move_cap(PIXEL, PIXEL);
static void move_homepoint(PIXEL, PIXEL);
static void pre_put_char(FONT_INFO*, uint*);
static void put_jis_font(uint);
static void put_dl_font(uint);
static long get_dlhsize(int);
static void pre_download(FONT_INFO*, DL_INFO*, int, int);
static void download_rastor(FONT_INFO*, DL_INFO*, uint, int);
static void post_download(FONT_INFO*, DL_INFO*);

static int j_lbp_font = -1;    /* current font in G1 */
static int HomePointBase;

extern char ptex_mode;
extern char ptex_d;    /* ptex direction: 0 = yoko, 1 = tate */
extern long ptex_dpi_adj;

/* externals from prtinit.c */
extern char *youshi;

#ifndef NOTATEGAKI
extern int tategaki;
#endif

#ifdef FDOWN
extern int maxdwidth;    /* download font width limit */
extern int maxdheight;   /* download font height limit */
extern int maxdfont;     /* num of kinds of download limit*/
#endif

extern int weight_org;
extern int k_base_l;
extern int k_base_d;
extern int k_scale_l;
extern int lbp_orig_dpi;
extern char str_exp[];

/* externals from pret.c */
extern int g_font;

PREAMBLE *get_font(uint, FONT_INFO *);

/* p_out.c routine call */
void pr_putnc(char *, int);
void pr_puts(char *);
void pr_putsn(char *, int);
void pr_putc(char);
void pr_int2(uint num);
void pr_int4(ulong num);

static void escpage_drawbox(PIXEL, PIXEL, PIXEL, PIXEL, int);

static struct LIPS_LFONT escpage_data[] =
{
    {"Mincho.J83", 10},
    {"Gothic.J83", 9},
    {"User", 10}
};

LBP_FUNC escpage_func =
{
    escpage_init,  escpage_init2,   escpage_end,  set_escpage_font,
    move_cap,      move_homepoint,  pre_put_char, put_jis_font,
    put_dl_font,   get_dlhsize,     pre_download, download_rastor,
    post_download, escpage_drawbox, escpage_data
};

static void pr_attr(int flag)
{
    if (flag & F_GOTH)  pr_puts(BOLD);
    if (flag & F_GGOTH) pr_puts(BBOLD);
}

static void escpage_init(void)
{
    char buf[256];
    char *s;
    int paper, yoko, tate, houkou, kyuushi, bin;

    if (lbp_orig_dpi == 0) lbp_orig_dpi = LBP_ORIG_DPI;
    if (k_base_l >= 10000)
        k_base_l = (lbp_orig_dpi != 240) ? K_BASE_L300 : K_BASE_L240;

    if (k_scale_l == 10000) k_scale_l = K_SCALE_L;

    k_base_d = K_BASE_D;
    /* weight_org=WEIGHT; */

#ifdef FDOWN
    maxdwidth = MAXDWIDTH;      /* maximal width of download font cell */
    maxdheight = MAXDHEIGHT;    /* maximal height of download font cell */
    maxdfont = MAXDFONT;        /* maximal number of download font sets */
#endif

    sprintf(buf ,"\x1d" "rhE" "\x1d" "0;%smuE" "\x1d" "0;%u;%udrE"
        "\x1d" "1;%u;%udrE" "\x1d" "2;%u;%udrE" "\x1d" "8;0lpP"
        "\x1d" "1abP" "\x1d" "0;0;1clfP" "\x1d" "1pmP" "\x1d" "0;0pP"
        "\x1d" "511iaF" "\x1d" "511ifF" "\x1d" "1mmE"
        ,(lbp_orig_dpi == 240) ? "0.3" : "0.24"
        ,lbp_orig_dpi, lbp_orig_dpi
        ,lbp_orig_dpi, lbp_orig_dpi
        ,lbp_orig_dpi, lbp_orig_dpi
    );

    pr_puts(buf);

    /* Youshi sitei '-y' */
    s = youshi;
    paper = 0;
    houkou = -1;
    kyuushi = 0;
    bin = 0;
    while (*s) {
        switch (*(s++)) {
          case 'A':
          case 'a':
              if (*(s++) == '5') paper = 15;
              else               paper = 14;
              while (isdigit(*s))  s++;
              break;
          case 'B':
          case 'b':
              if (*(s++) == '4') paper = 24;
              else               paper = 25;
              while (isdigit(*s))  s++;
              break;
          case 'H':
          case 'h':
              paper = 38;
              while (isdigit(*s))  s++;
              break;
          case 'T':
          case 't':
              paper = 0;
              while (isdigit(*s)) {
                  paper = paper * 10 + (*s - '0');
                  s++;
              }
              break;
          case 'F':
          case 'f':
              paper = -1;
              yoko = tate = 0;
              while (isdigit(*s)) {
                  yoko = yoko * 10 + (*s - '0');
                  s++;
              }
              if (*s == ':' || *s == ';') {
                  s++;
                  while (isdigit(*s)) {
                      tate = tate * 10 + (*s - '0');
                      s++;
                  }
              }
              else tate = yoko;
              break;
          case 'L':
          case 'l':
              houkou = 1;
              kyuushi = 0;
              while (isdigit(*s)) {
                  kyuushi = kyuushi * 10 + (*s - '0');
                  s++;
              }
              break;
          case 'P':
          case 'p':
              houkou = 0;
              kyuushi = 0;
              bin = 0;
              while (isdigit(*s)) {
                  kyuushi = kyuushi * 10 + (*s - '0');
                  s++;
              }
              if (*s == ':' || *s == ';') {
                  s++;
                  while (isdigit(*s)) {
                      bin = bin * 10 + (*s - '0');
                      s++;
                  }
              }
              break;
        }
    }

    if (paper != 0) {
        if (paper != -1) sprintf(buf, "\x1d" "%upsE", paper);
        else             sprintf(buf, "\x1d" "-1;%u%upsE", yoko, tate);
        pr_puts(buf);
    }
    if (houkou != -1)
        pr_putsn("\x1d" "%upoE", houkou);

    if (bin == 0) bin = 1;
    if (kyuushi != 0 && lbp_orig_dpi != 240) {
        pr_putsn("\x1d" "%u;", kyuushi);
        pr_putsn("%uiuE", bin);
    }
    ptex_dpi_adj = (long)lbp_orig_dpi *100L;
    HomePointBase = (lbp_orig_dpi * 2) / 10;
    move_homepoint(0, 0);
}

static void escpage_end()
{
    /* dummy */
}

/* Set Font in LBP */
static void set_escpage_font(FONT_INFO *fontc, KFONT *kf, int fg, int weight, int face)
{
    if (fg & F_OUTLINE) kf->width -= 2;
    g_font = (fontc->f_goth & (F_ATTR | F_LBP));
    pr_putsn(CHR_PITCH, kf->size);
    pr_putsn(CHR_SIZE, kf->width);
    pr_putsn(FACE, face);
#ifndef NOTATEGAKI
    if (tategaki) pr_puts(TATEGAKI);
    else
#endif
        pr_puts(YOKO);
    if (fg & F_GOTH)       pr_puts(BOLD);
    else if (fg & F_GGOTH) pr_puts(BBOLD);
    else                   pr_puts(NO_BOLD);
    pr_putsn(ASSIGN_G1, fontc->font_code);
    j_lbp_font = -1;
}

static int KetaSuu(PIXEL num)
{
    int Keta;

    if (num >= 0) Keta = 0;
    else {
        Keta = 1;
        num = (-num);
    }
    if (num < 10)         return(Keta + 1);
    else if (num < 100)   return(Keta + 2);
    else if (num < 1000)  return(Keta + 3);
    else if (num < 10000) return(Keta + 4);
    return(Keta + 5);
}

static void move_cap(PIXEL v, PIXEL h)
{
    static PIXEL last_v, last_h;
    PIXEL sub;

    if (v < 0) {
        last_v = last_h = -1;
        return;
    }
    if (last_v != v) {
        sub = v - last_v;
        if (KetaSuu(v) <= KetaSuu(sub) || last_v == -1)
            pr_putsn(SET_Y, v);
        else
            pr_putsn(MOVE_BELOW, sub);
        last_v = v;
    }
    if (last_h != h) {
        sub = h - last_h;
        if (KetaSuu(h) <= KetaSuu(sub) || last_h == -1)
            pr_putsn(SET_X, h);
        else
            pr_putsn(MOVE_RIGHT, sub);
        last_h = h;
    }
}

static void move_homepoint(PIXEL v, PIXEL h)
{
    pr_putsn("\x1d" "%u;", h + HomePointBase);
    pr_putsn("%uloE", v + HomePointBase);
    move_cap(-1, -1);
}

#define PUSH_POS  pr_puts("\x1d" "1ppP");
#define POP_POS   pr_puts("\x1d" "2ppP");

void MoveWrite(uint code, PIXEL h, PIXEL v)
{
    if (h != 0) pr_putsn("\x1d" "%dH", h);
    if (v != 9) pr_putsn("\x1d" "%dV", v);
    pr_int2(code);    /* output JIS code */
}

static int ass_f;
static int knjwidth;
static int last_d = -1;

static void pre_put_char(FONT_INFO *font, uint *code)
{
    ass_f = font->font_code;
    knjwidth = font->k_width;
    return;
}

static void put_jis_font(uint code)
{
    if (ass_f != j_lbp_font) {
        /* change G1 font */
        pr_putsn(CHANGE_G1, j_lbp_font = ass_f);
        last_d = -1;
    }
    if (ptex_mode) {
        if (ptex_d != last_d) {
            pr_putsn(KAITEN, 0);
            if (ptex_d == 0) pr_puts(YOKO);
            else             pr_puts(TATE);
            last_d = ptex_d + 2;
        }
    }
    if ((g_font & (F_SHADOW | F_OUTLINE)) == (F_SHADOW | F_OUTLINE)) {
        int Kage;

        PUSH_POS;
        MoveWrite(code, 0, -1);
        MoveWrite(code, 1, -1);
        MoveWrite(code, 1, 1);
        MoveWrite(code, -1, 1);
        MoveWrite(code, 0, 1);
        for (Kage = knjwidth >> 4; Kage > 0; Kage--)
            MoveWrite(code, 1, 1);
        POP_POS;
    }
    else if (g_font & F_SHADOW) {
        int Kage;

        PUSH_POS;
        for (Kage = knjwidth >> 4; Kage > 0; Kage--)
            MoveWrite(code, 1, 1);
        POP_POS;
    }
    else if (g_font & F_OUTLINE) {
        PUSH_POS;
        MoveWrite(code, 0, -1);
        MoveWrite(code, 1, -1);
        MoveWrite(code, 1, 1);
        MoveWrite(code, -1, 1);
        POP_POS;
    }
    if (g_font & (F_SHADOW | F_OUTLINE | F_FILL)) {
        if (g_font & F_FILL)
            pr_puts("\x1d" "1tsE" "\x1d" "1;0;20spE" "\x1d" "1owE");
        else
            pr_puts("\x1d" "1tsE" "\x1d" "1;0;0spE" "\x1d" "1owE");

        if (g_font & F_OUTLINE) {
            PUSH_POS;
            MoveWrite(code, 1, -1);
            POP_POS;
        }
        else  pr_int2(code);        /* output JIS code */
        pr_puts("\x1d" "0tsE");
    }
    else {
        if (g_font & F_OUTLINE) {
            PUSH_POS;
            MoveWrite(code, 1, -1);
            POP_POS;
        }
        else  pr_int2(code);        /* output JIS code */
    }
}

#define    SHIFT_ASSIGN 0x100
#define    SHIFT_CHAR 0x20

static void put_dl_font(uint code)
{
    if (code & 0x80) ass_f += SHIFT_ASSIGN;
    if (ass_f != j_lbp_font) {    /* change G0 font */
        pr_putsn(CHANGE_G1, j_lbp_font = ass_f);
        last_d = -1;
    }
    if (last_d != ptex_d) {
        if (ptex_d == 0) pr_putsn(KAITEN, 0);
        else             pr_putsn(KAITEN, 270);
        last_d = ptex_d;
    }
    pr_putc((code & 0x7f) + SHIFT_CHAR);    /* output ASC code */
}

static long get_dlhsize(int NumOfFonts)
{
    return((long)NumOfFonts * D_CHAR + D_HEAD);
}

static void pre_download(FONT_INFO *dfont, DL_INFO *dl, int num_dfont, int flag)
{
    if (flag) dl->assign_num += SHIFT_ASSIGN;

    pr_putsn("\x1d" "134;%udh{F", dl->assign_num);
    /* assign a number to FONT */
    pr_int2(2);
    pr_int2(0x86);
    pr_int2(USER_SYMBOLBASE + dl->assign_num);
    pr_int2(1);
    pr_int4(0x1000000L);
    pr_int4((long)(dl->maxheight));
    pr_int2(USER_TYPEFACEBASE + dl->assign_num);
    pr_int2(0);
    pr_int2(0);
    pr_int2(0);
    pr_int2(dl->maxcharcode + SHIFT_CHAR);
    pr_int2(dl->mincharcode + SHIFT_CHAR);
    pr_int2(0);
    pr_int2(0);
    pr_int2(0);
    pr_int2(0);
    pr_int2(0);
    pr_int2(0);
    pr_int2(dl->maxwidth);
    pr_int2(dl->maxheight);
    pr_int2(0);
    pr_int2(dl->maxheight);
    pr_int4((long)(dl->maxwidth));
    pr_putsc(dfont->name, 54);
    pr_putsc("TexFonts", 28);
    pr_putsn("\035%udcF", dl->assign_num);    /* select download font for later use*/
}

static void download_rastor(FONT_INFO *dfont, DL_INFO *dl, uint code, int flag)
{
    PREAMBLE *preamble;
    int bytesize;
    BUFFER *pat;

    preamble = get_font(flag + code, dfont);
    bytesize = preamble->byte_width * preamble->height;
    sprintf(str_exp, "\x1d" "%udcF" "\x1d" "%u;%usc{F"
            ,dl->assign_num, 16 + bytesize, (code & 0x7f) + SHIFT_CHAR);
    pr_puts(str_exp);
    pr_putc(0x1);
    pr_putc(0x10);
    pr_int2(0);
    pr_int2(preamble->width);
    pr_int2(preamble->height);
    pr_int2(0);
    pr_int2(preamble->height);
    pr_int4((long)preamble->width);
    pat = preamble->raster;
    while (bytesize--)  pr_putc(*pat++);
}

static void post_download(FONT_INFO *dfont, DL_INFO *dl)
{

    pr_putsn("\x1d" "%u;0ssF", USER_SYMBOLBASE + dl->assign_num);
    /*symbol set number*/
    pr_puts(NO_BOLD);                     /* stroke weight */
    pr_putsn(KAITEN, 0);                  /* rotation */
    pr_putsn(CHR_PITCH, dl->maxwidth);    /* character width */
    pr_putsn(CHR_SIZE, dl->maxheight);    /* character hight */
    pr_putsn("\x1d" "%utfF", USER_TYPEFACEBASE + dl->assign_num);
    /*style set */

    pr_putsn(ASSIGN_G1, dl->assign_num);

}

static void escpage_init2(int f_rotate, int f_copy)
{
    if (f_rotate) pr_puts(ROTATE);
    if (f_copy)   pr_putsn(COPYS, f_copy);
}

static int graytab[] = {0, 20, 40, 100};

static void escpage_drawbox(PIXEL vs, PIXEL hs, PIXEL height, PIXEL width, int gray)
{
    int graylevel;

    if (height < 0) height = (-height);
    else            vs += (1 - height);

    if (width < 0) {
        hs += (1 - width);
        width = (-width);
    }
    graylevel = graytab[gray];

    if (gray) {
        move_cap(vs, hs);
        pr_putsn("\x1d" "0;0;%uspE", graylevel);
        pr_putsn("\x1d" "%u;", width);
        pr_putsn("%ubgI", height);
    }
    pr_putsn("\x1d" "%u;", hs);
    pr_putsn("%u;", vs);
    pr_putsn("%u;", hs + width - 1);
    pr_putsn("%u;0rG", vs + height - 1);
    move_cap(-1, -1);
}

#endif  /* ESCPAGE */

/* end of escpage.c */
