/*
 *  TEX Device Driver  ver 2.02-2.33
 *  copyright(c) 1988, 1989 by TSG, 1990 by SHIMA, M.Watanabe
 *  1991 changed by T.Minagawa
 *
 *  fontdef.c : module to make font information list
 *  
 *  Jul.  1, 1989 : 3rd edition
 *
 *  modified for LIPS3 by SHIMA 30 June 1992
 *  modified for Minix-386 1992 by obuk, youchan
 *  modified by Naochan! 6 August 1992
 *  slightly modified for LIPS3 and ESC/Page by OkI 8 August 1992
 *  slightly modified for JGfont(3D-Bezier) by Naochan! 12 Nov 1992
 */

#include "config.h"
#include "dd.h"
#include <fcntl.h>
#include <unistd.h>
#include "err.h"
#include "dviread.h"
#include "vfont.h"

#define READ_MODE       0x04
#define MAX_NAME_LEN    25
#define MAX_SUB_LEN     15
#define TMP_BUF_SIZE    128
#define ROM_F_SIZE      16

extern char* pk_env_name;
char *pk_search_path  = (char*)NULL;
char *pkd_search_path = (char*)NULL;
char *knj_search_path = (char*)NULL;

extern int test_flg;
extern int prt_flag;

BOOL    f_long_name;
char    *goth="goth";
char    *sub_font;
int     f_search;
static  int f_loop;
static  int f_sub_font;
int     f_use_vfont = FALSE;

#ifdef LBP
int     f_use_lfont = FALSE;
#endif

struct FONT_ATR {
    long k, c, s, d;
    int  a, l;
    char* n;
};

struct  finfo{
    char      f_atrib;
    ushort    f_time;
    ushort    f_date;
    long      f_size;
    char      f_name[14];
    long      f_pos;
};

#define FINFO_SIZE  27

static  int     dpis, mags;
/* static double dpi, true_mag; */
    /* dpi, mag */

/* standard library */
#ifdef msdos
void* memmove(void*, void*, size_t);
#else
char* getenv(const char*);
#endif
#ifdef minix
int access(char*, int);
#endif

/* buffer.c */
char* marea(unsigned int);
char *dup_string(char*);
FONT_INFO *font_info_alloc(void);

/* loadpk.c  */
int openf(char *, int);

/* vfont.c */
KFONT *get_vfont_list(FONT_INFO *, char *);

/* p_out.c(lbp.c) */
KFONT *get_lbp_font(FONT_INFO *);

char    *k_dbl = "";
extern  FONT_INFO* font_info_root;

#define store_font_name(x)  dup_string(x)

static void font_define(int, FILE*);
static void make_font(struct FONT_ATR*);
static char* font_name_check(char*, int, int);
static char* str_fnt_nm(char* , char *, long, long);
static BOOL copy_token(char**, char**, int);
static int token_last(char*);
static BOOL fill_name_and_fig(char*, char*, char*, char*);
static void strdel(char *, int);
static void strins(char *, char *);
char *get_fontname(FONT_INFO *);

char tmp_buf[ TMP_BUF_SIZE ];
static  char o_font[ MAX_SUB_LEN ];
static  char s_font[ MAX_SUB_LEN ];
static  char o_fig[ MAX_SUB_LEN ];
static  char s_fig[ MAX_SUB_LEN ];

static long fsize, fpos;

static void make_font_tree(FONT_INFO** , int, int, int);
static void font_sort(int );

char *get_font_name(FONT_INFO* font)
{
    if (*(font->name) == '*') {
            if (strchr(font->name+9, ' ') == NULL)
                font->name[9+strlen((font->name) + 9)] = ' ';
            return(font->name + 9);
    }
    return(font->name);
}

void make_font_list(DVIFILE_INFO* dvi, DIMENSION* dimension)
{
    int code, count;

    dpis = dimension->dpi;
    mags = dimension->mag;
    f_loop = 3;
    if (f_search > 0) f_loop += f_search;
    fseek(dvi->file_ptr, dvi->post + 29, SEEK_SET);
        /*  set file-ptr to first font_def */

    count = 0;
    while((code = getc(dvi->file_ptr)) != POST_POST)
        if (code >= FNT_DEF_1 && code < FNT_DEF_1 + 4) {
            font_define(code, dvi->file_ptr);
            count++;
        }
        else if (code != NOP)
            error(COMMAND_ERROR, "code:%d", code);
    font_sort(count);
}

static void font_sort(int total)
{
    int     i, j;

    FONT_INFO** font_adr;
    FONT_INFO* font;

    font_adr = (FONT_INFO**)(marea(sizeof(FONT_INFO*)*total)); 
    font = first_font_info;
    for(i = total-1; i >= 0 ; i--) {
        font_adr[i] = font;
        font = font->next_font;
        if ((i == 0 && font != NULL) || (i > 0 && font == NULL))
            error(COMMAND_ERROR, "Making font table");

    }

    for(i = 1; i < total;) {
        j = i;
        while(i-- > 0 && font_adr[i]->font_code > font_adr[i+1]->font_code) {
            font = font_adr[i];
            font_adr[i] = font_adr[i+1];
            font_adr[i+1] = font;
        }
        i = j+1;
    }

    i = (total-1)>>1;
    font_info_root = font_adr[i];
    make_font_tree(font_adr, i, 0, total-1);
    free(font_adr);
}

static void make_font_tree(FONT_INFO** font_adr,
 int now, int before, int after)
{
    int     current;

    if (before < now) {
        current = before + ((now-before)>>1);
        font_adr[now]->before_font = font_adr[current];
        make_font_tree(font_adr, current, before, now-1);
    }

    if (now < after) {
        current = now + ((after-now+1)>>1);
        font_adr[now]->after_font = font_adr[current];
        make_font_tree(font_adr, current, now+1, after);
    }
}

static void font_define(int code, FILE* dvifile)
{
    int     i, len, lenm;
    struct FONT_ATR font;
    char    *tmp, tmp_name[MAX_NAME_LEN];

    font.k = read_n(dvifile, code - FNT_DEF_1 + 1);
    font.c = read_long(dvifile);
    font.s = read_long(dvifile);
    font.d = read_long(dvifile);
    font.a = read_byte(dvifile);
    font.l = read_byte(dvifile);
    font.n = tmp_name;

    if (font.a == EOF || font.l == EOF)
        error(FILE_FAULT, "DVI file");
    
    tmp = read_str(dvifile, font.a + font.l);
    len = strlen(tmp);
    i = len;
#ifdef msdos
    lenm = 8;
    while (--i >= 0) {
        if (tmp[i] == '.') lenm = 13;
        if (tmp[i] == ':' || tmp[i] == '\\') break;
    }
#else
    lenm = 14;
    while (--i >= 0 && tmp[i] != '/');
#endif
    if (len - ++i > lenm && !f_long_name) tmp[i+lenm] = '\0';
    strncpy(tmp_name, tmp, MAX_NAME_LEN - 5);
    if (len - i > lenm && f_long_name && i + lenm < MAX_NAME_LEN) {
        strcpy(tmp_name + i + 4, tmp + i + len - lenm + 4);
    }
    free(tmp);
    make_font(&font);
}

static char* next_top(char *dest, char *org)
{
    int   i, ch;

    for(i = 0; ; i++) {
        ch = *org++;
        if (i < MAX_SUB_LEN-1) *dest++ = ch;
        if (ch == 0) return(org - 1);
        if (ch == '.' || ch == '=' || ch == ';') {
            *(dest-1) = 0;
            return(org);
        }
    }
}

static void make_font(struct FONT_ATR* atr)
{
    FONT_INFO* font;
    char    *name, *path_name, *ptr;
    int     i, fig, figl, fmags;
    ulong   atr_s, atr_d, atr_t;
    static  char    null_name[1] = {0};
    
    if (first_font_info == NULL)
        font = first_font_info = font_info_alloc();
    else {
        for(font = first_font_info; font->next_font;
            font = font->next_font)
            if (font->font_code == atr->k)
                error(PROGRAM_STOP, "Twice fnt_def");
        
        font = font->next_font = font_info_alloc();
    }
    
    font->font_code = (int)atr->k;
    atr_t = (ulong)dpis*(ulong)mags;
    if (atr->s == atr->d) {
        atr_s = atr_t;
        atr_d = 10;
        i = 1;
    }
    else{
        atr_s = atr->s;
        atr_d = atr->d;
        while(atr_d > 10000L) {
            atr_s >>= 1;
            atr_d >>= 1;
        }
        fig = (atr_s*1000 + (atr_d >> 1))/atr_d;
        atr_s *= atr_t/10L;
        i = 2;
    }
    for(fmags = mags; i > 0; i--) {
        if     (fmags == 1095) atr_s += atr_s/2460;   /* 1095.4451... */
        else if (fmags == 2074) atr_s -= atr_s/5183;   /* 2073.6       */
        else if (fmags == 2488) atr_s += atr_s/7775;   /* 2488.32      */
        fmags = fig;
    }
    fig = (atr_d*50 + atr_s)/(atr_d*100);
    figl= (atr_d*10 + atr_s)/(atr_d*20);
    name = atr->n;
again:
    if ((path_name = font_name_check(name, fig, figl)) != NULL)
        goto exist;
    if (sub_font != NULL) {
        f_sub_font++;
        ptr = sub_font;
        do{
            ptr = next_top(&(s_font[0]), ptr);
            ptr = next_top(&(s_fig[0]),  ptr);
            ptr = next_top(&(o_font[0]), ptr);
            ptr = next_top(&(o_fig[0]),  ptr);
            if ((path_name = font_name_check(name, fig, figl)) != NULL) {
                f_sub_font = 0;
                goto exist;
            }
        }while(*ptr != 0);
    }
    if (f_sub_font < 2) {
        f_sub_font = 2;
        goto again;
    }
    if (test_flg < 0) error(NO_FONT, "Check your Font file!");
    test_flg = f_sub_font = 0;
    path_name = null_name;
exist:
    font->name = path_name;
    if (test_flg == 0) {
        printf(" %3d%9s.%03d(%4d): %s\n", font->font_code, name, fig, figl, get_font_name(font));
    }
    font->size_para = atr->s;
    font->c = atr->c;
    font->font_type = UNKNOWN;
    font->f_goth = 0;
    font->next_font = font->before_font = font->after_font = NULL;
    font->pk = font->ext.dir = NULL;
    for(i = 0; i < CHAR_ROOT+2; i++) font->char_info[i] = NULL;
}

int chk_goth(FONT_INFO *font)
{
    int     offset;

    if (*goth == 0) return(0);

    if (*(font->name) == '*') {
        offset = 9;
        offset += strlen((font->name) + 9) + 1;
    }
    else{
        offset = strlen(font->name) - 13;
        if (offset < 0) offset = 0;
    }
    if (strstr(font->name + offset, goth) == NULL)
        return(0);
    return(1);
}

static char* str_fnt_nm(char* name, char *ptr_sp, long pos, long size)
{
    int len;
    char *ptr, *ptr1;

    len = strlen(name);

    ptr = ptr1 = (char *)marea(len + strlen(ptr_sp + 1) + (1+2*4+1+1));
    *ptr1= '*';
    *((long *)(ptr1+1)) = pos;
    *((long *)(ptr1+5)) = size;
    strcpy(ptr1+9, name);
    strcpy(ptr1+10+len, ptr_sp+1);
    return(ptr);
}

#define FIG_COLUMN      8

static char* font_name_check(char* name, int fig, int figl)
{
    char *path, *path0, *ptr_sp;
    BOOL token_exist, number_exist;
    int  i, j;
    char fig_str[ FIG_COLUMN ], figl_str[ FIG_COLUMN ];
    char* tmp;

    if (pk_search_path == NULL) {
        if ((path0 = getenv(pk_env_name)) == NULL)  path0 = PKDIR;
    }
    else
        path0 = pk_search_path;

    do {
        for(i = j = 0; i < f_loop  ;i++) {
            sprintf(fig_str, "%d", fig + j);
            sprintf(figl_str, "%d", figl + j);
            j = -j;
            if (j >= 0) j++;
            tmp = tmp_buf;
            path = path0;
            token_exist = copy_token(&tmp , &path, DELIMITER);
            *tmp = '\0';
            if (f_sub_font & 1) {
                if (*o_font != 0  && strcmp(o_font, name) != 0) continue;
                if (*o_fig  != 0) {
                    if (strcmp(o_fig, fig_str) != 0 &&
                        strcmp(o_fig, figl_str) != 0) continue;
                }
                number_exist = fill_name_and_fig(tmp_buf, 
                    (*s_font != 0)? s_font:name,
                    (*s_fig != 0)?  s_fig:fig_str,
                    (*s_fig != 0)?  s_fig:figl_str);
            }
            else 
                number_exist = fill_name_and_fig(tmp_buf, name, fig_str, figl_str);
            if (f_sub_font >= 2) goto loop;
            if ((ptr_sp = strchr(tmp_buf, ' ')) == NULL) {
                if (access(tmp_buf, READ_MODE) != FAILURE)
                    return(store_font_name(tmp_buf));
            } else {
                if (accgth(tmp_buf, ptr_sp) != FAILURE)
                    return(str_fnt_nm(tmp_buf, ptr_sp, fpos, fsize));
            }
loop:       if (!number_exist) break;
        }
        path0 = path;
    }while(token_exist);
    return(NULL);
}


void    open_jxl(FONT_INFO *font)
{
#ifdef msdos
        if ((font->ext.fh = openf(font->name, O_RDONLY|O_BINARY)) > 0)
#else
        if ((font->ext.fh = openf(font->name, O_RDONLY)) > 0)
#endif
            return;
        error(FILE_FAULT, "Cannot re_open JXL file %s");
}

void    open_pk(FONT_INFO *font)
{
    int     i;
    char    *dir;
    char    fname[100];
    BUFFER  *ptr;

    font->k_width = font->pk[2] + (3+16);
    ptr = font->pk + 7;
    i = 0;

    if ((dir = getenv("TEXPKD")) == NULL &&
        (dir = pkd_search_path) == NULL)  dir = PKDDIR;
    for(; i < 99; i++)
        if ((fname[i] = dir[i]) == 0) break;

    while(i < 100) {
        if ((fname[i++] = *ptr++) == 0) break;
    }
#ifdef msdos
    if ((font->ext.fh = openf(fname, O_RDONLY|O_BINARY)) > 0) return;
#else
    if ((font->ext.fh = openf(fname, O_RDONLY)) > 0) return;
#endif
    error(NO_FILE, "PK file %s", fname);
}

char *get_fontname(FONT_INFO *fontc)
{
    char *tmp;
    int  ch, width, width0, widthm;
    static char font_name[10];

    if (*(tmp = fontc->name) == '*') tmp += strlen(tmp + 9) + 10;
    width = strlen(tmp) - 4;
    font_name[0] = 0;
    if (tmp[width--] == '.') {
        while((ch = tmp[width]) >= '0' && ch <= '9' && width > 0) width--;
        for(width0 = width; width0 >= 0; width0--) {
            if ((ch=tmp[width0]) == '/' || ch == ':' || ch == 0) {
                break;
            }
        }
        width0++;
        if ((widthm = width - width0) < 9) {
            for(ch = 0; ch <= widthm; ch++) font_name[ch] = tmp[width0+ch];
            font_name[ch] = 0;
        }
    }
    return(font_name);
}

KFONT*  dup_kf(KFONT *kfo)
{
    KFONT *kf;

    kf = (KFONT*)marea(KFONT_SIZE);
    *kf = *kfo;
    return(kf);
}

KFONT*  open_kfont(FONT_INFO *fontc, int maxwidth, int dir)
{
    char *tmp, *path, *path0, *pt;
    int  ch, width, width0, widthm, fh, dif, f_dbl, loop;
    char fig_str[ FIG_COLUMN ];
    char *name;
    KFONT*  kf;
    FONT_INFO*  font;
    FONT_INFO*  fontk;
    BOOL token_exist;

    dif = loop = f_dbl = 0;
    name = get_fontname(fontc);
    width = fontc->k_width;
    if (width >= ROM_F_SIZE)
        width0 = ROM_F_SIZE-1;
    else width0 = 0;

#ifndef NOVFONT
    if (f_use_vfont) {
        kf = get_vfont_list(fontc, name);
        if (kf != NULL) return(kf);
    }
#endif
#ifdef LBP
    if (f_use_lfont && prt_flag != 0) {
        kf = get_lbp_font(fontc);
        if (kf != NULL) return(kf);
    }
#endif

rep:        /* search kanji font file */
    for(; width != width0; width += dif) {
        if (width > maxwidth) goto next;
        f_dbl = 0;
        widthm = width;
repdbl:     /* search again */
        sprintf(fig_str, "%d", widthm);
        if ((path0 = getenv("TEXKNJ")) == NULL &&
            (path0 = knj_search_path) == NULL)  path0 = KNJDIR;
        if (*path0 != 0) {
            do {
                tmp = tmp_buf;
                path = path0;
                token_exist = copy_token(&tmp , &path, DELIMITER);
                *tmp = '\0';
                fill_name_and_fig(tmp_buf, name, fig_str, "");
#ifdef msdos
                fh = openf(tmp_buf, O_RDONLY|O_BINARY);
#else
                fh = openf(tmp_buf, O_RDONLY);
#endif
                if (fh > 0)  goto exit;
                path0 = path;
            } while(token_exist);
        }

        if (f_dbl == 0) {
            if (!(width & 1) 
              && (pt = strstr(k_dbl, fig_str)) != NULL 
              && *(pt-1) == DELIMITER
              && *(pt + strlen(fig_str)) == DELIMITER) {
                f_dbl = K_DBL;
                widthm = width/2;
                goto repdbl;
            }
        }

next:   if (!loop)  dif = (dif < dir)?1-dif:-dif-1;
    }

    if (loop++ == 0 && fontc->k_width < ROM_F_SIZE) {
        width = fontc->k_width + 1;
        dif = 1;
        maxwidth = width0 = ROM_F_SIZE+1;
        goto rep;
    }

exit:                   /* define Kanji Font file */
    fontc->k_top = fontc;
    kf = (KFONT *)marea(KFONT_SIZE);
    kf->name = (char *)dup_string(tmp_buf);
    if (f_dbl == K_DBL) {
        fontc->f_goth |= K_DBL;
        width >>= 1;
    }
    if (fontc->k_top == NULL)  fontc->k_top = fontc;
#ifdef BEZIERFONT
    if (font->font_type == JG_FONT)  kf->base = width >> 2;
    else
#endif
        kf->base = 0;
    kf->width = width;
    kf->height = width + kf->base;
    kf->size = (width + 7) / 8 * kf->height;
    kf->fh = fh;
    return(kf);
}


int accgth(char *fname, char *ptr_sp)
{
    int     i, stat, f_num;
    FILE    *fp;
    long    start;
    char    gth_buf[FINFO_SIZE];
    struct  finfo finfo;
    struct FARDIR {
        char    f_name[10];
        long    f_time;
        long    f_size;
        long    f_pos;
    }fardir;
    char fardir_buf[22];

    *ptr_sp = '\0';
#ifdef msdos
    if ((fp = fopen(fname, "rb")) == NULL)
#else
    if ((fp = fopen(fname, "r")) == NULL)
#endif
         goto non_exist1;

    fread(gth_buf, FINFO_SIZE, 1, fp);
    if (strcmp(gth_buf+8, "gather") == 0) goto exist;
    fseek(fp, -8L, SEEK_END);
    f_num = (int)read_short(fp);
    start = read_long(fp);
    if (read_short(fp) != (short)('f'+'o'+'n'+'t'+'a'+'r'+'c'+'h'+'i'+'v'+'e'))
        goto non_exist;

    fseek(fp, start, SEEK_SET);
    while(f_num-- > 0) {
        fread(fardir_buf, 22, 1, fp);
        memcpy((char*)(&fardir.f_name[0]), fardir_buf, 10);
        memcpy((char*)(&fardir.f_time), fardir_buf+10, 12);
        if ((stat = strcmp(fardir.f_name, ptr_sp+1)) == 0) {
            fsize = fardir.f_size;
            fpos  = fardir.f_pos;
            goto accend;
        }
        if (stat > 0) break;
    }
    goto non_exist;

exist:
    start = *((long *)(gth_buf));
    f_num  = *((int *)(gth_buf + 4));
    fseek(fp, start, SEEK_SET);
    while(f_num-- > 0) {
        fread(gth_buf, FINFO_SIZE, 1, fp);
        memcpy(&finfo.f_atrib, gth_buf, 1);
        memcpy(&finfo.f_time, gth_buf+1, 2);
        memcpy(&finfo.f_date, gth_buf+3, 2);
        memcpy(&finfo.f_size, gth_buf+5, 4);
        memcpy(&finfo.f_name, gth_buf+9, 14);
        memcpy(&finfo.f_pos, gth_buf+23, 4);
        if ((stat = strcmp(finfo.f_name, ptr_sp+1)) == 0) {
            fsize = finfo.f_size;
            fpos  = finfo.f_pos;
accend:     fclose(fp);
            return(NOTHING);
        }
        if (stat > 0) break;
    }
non_exist:
    fclose(fp);
non_exist1:
    *ptr_sp = ' ';
    return(FAILURE);
}

static BOOL copy_token(char** dst_ptr, char** src_ptr, int delim)
{
    char *src;
    char *dst;
    
    src = *src_ptr;
    dst = *dst_ptr;
    while (*src != '\0' && *src != delim)  *dst++ = *src++;
    *src_ptr = src + 1;
    *dst_ptr = dst;
    return(*src != '\0');
}

static int token_last(char* token)
{
    if (! *token)  return('\0');
    while(*++token) ;
    return(*(token - 1));
}

static void strdel(char *line, int num)
{
    int len;

    len = strlen(line);
    if (len < num)  return;
    memmove(line, line + num, len + 1);
    /* don't forget delimiter '\0'... */
}

static void strins(char *line, char *token)
{
    int len, token_len;

    len = strlen(line);
    token_len = strlen(token);
    memmove(line + token_len, line, len + 1);
    memcpy(line, token, token_len);
}

#define FLG_LEN         2
#define FLG_HEAD        '%'
#define FLG_HEAD2       '^'
#define NAME_FLG        's'
#define FIG_FLG         'd'
#define FIGL_FLG        'l'
#define GTH_FLG         'g'

static BOOL fill_name_and_fig(char* buf, char* fname, char* fig, char* figl)
{
    char *tmp;
    int i, len, name_len, fig_len, figl_len;
    BOOL    f_num = FALSE;

    len = strlen(buf);
    name_len = strlen(fname);
    fig_len = strlen(fig);
    figl_len = strlen(figl);

    for (tmp = buf; *tmp;) {
        if (*tmp == FLG_HEAD || *tmp == FLG_HEAD2) {
            if (tmp[1] == NAME_FLG) {
                len = len - FLG_LEN + name_len;
                if (len >= TMP_BUF_SIZE) {
p_over:             error(ILLEGAL_ARGS, "Path name buffer is out of range");
                }
                strdel(tmp, FLG_LEN);
                strins(tmp, fname);
                tmp += name_len;
            }
            else if (tmp[1] == FIG_FLG) {
                f_num = TRUE;
                len = len - FLG_LEN + fig_len;
                if (len >= TMP_BUF_SIZE) goto p_over;
                strdel(tmp, FLG_LEN);
                strins(tmp, fig);
                tmp += fig_len;
            }
            else if (tmp[1] == FIGL_FLG) {
                f_num = TRUE;
                len = len - FLG_LEN + figl_len;
                if (len >= TMP_BUF_SIZE) goto p_over;
                strdel(tmp, FLG_LEN);
                strins(tmp, figl);
                tmp += figl_len;
            }
            else if (tmp[1] == GTH_FLG) {
                len--;
                strdel(tmp, FLG_LEN);
                if (strchr(buf, '.') == NULL || strchr(buf, '.') > tmp - 1) {
                    strins(tmp, ".gth ");
                    tmp += 4;
                } else {
                    strins(tmp, " ");
                }
            }
            else if (tmp[1] == tmp[0]) {
                len--;
                strdel(tmp, 1);
                tmp++;
            }
            else  error(ILLEGAL_ARGS, "%%%c in TEXPK or TEXKNJ", tmp[1]);
        }
        else  tmp++;
    }
    return(f_num);
}

/* end of file : fontdef.c */
