/*
 *  TEX Device Driver  ver 2.02-2.31
 *  copyright(c) 1988, 1989 by TSG, 1990 by SHIMA
 *  1991 changed by T.Minagawa
 *
 *  getfont.c : get unpacked font
 *
 *  Apr. 15, 1989 : first edition
 *  Jul.  2, 1989 : 2nd edition
 *  modified for non PC-9801 machines by sempa 1992
 *
 *  modified for LIPS3 by SHIMA 30 June 1992
 *  modified for pTeX (tate mode) by Naochan! 6 August 1992
 *  slightly modified for pTeX on LIPS3 by OkI 8 August 1992
 *  modified by Tomiie 20 August 1992
 *  modified for Minix-386 1992 by obuk, youchan
 *  modified for LIPS3 and ESC/Page by Oki and H.Tomiie 30 Sep 1992
 *  slightly modified for JGfont(3D-Bezier) by Naochan! 12 Nov 1992
 */

#include "config.h"
#include "dd.h"
#include "err.h"
#include "buffer.h"
#include "dviread.h"

#define PK_PRE      247
#define PK_ID       89
#define PKD_ID      ('p'+'k'+'d')
#define PK_POST     245

#define PKW_ID      ((PK_ID<<8)+PK_PRE)
#define PKDW_ID     ((PKD_ID<<8)+PK_PRE)
#define PXL1001_ID  (3+(233<<8))
#define PXL1002_ID  (3+(234<<8))
#define PXL1003_ID  (3+(235<<8))
#define JXL4_0_ID   (0x0100)
#define JXL4_1_ID   (0x258B)
#define JTFM_ID     (11<<8)
#define JTFMP_ID    (9<<8)

#define CHAR_SIZE   16
#define KCG_BUF     32

/* decodepk.c */
void decode_pk(BUFFER *, PREAMBLE *, int, FONT_INFO *);
void decode_jxl(BUFFER *, PREAMBLE *, int, FONT_INFO *);

/* loadpk.c */
BUFFER* load_pk(const char *);
int ask_pk_buf(long);
int file_read(int, BUFFER*, int);

/* raster.c */
void raster_buf_alloc(PREAMBLE**);
PREAMBLE* alloc_preamble(void);

#define expand_char_info_list()    char_info_ptr

/* misc.c  */
PIXEL sptopixel(SCALED_PT);
SCALED_PT mult20(SCALED_PT, SCALED_PT);

/* fontdef.c */
KFONT* open_kfont(FONT_INFO*, int, int);
void open_pk(FONT_INFO *);
void open_jxl(FONT_INFO *);
int chk_goth(FONT_INFO*);
char *get_font_name(FONT_INFO *);
#ifndef NOVFONT
int chk_vfont_goth(FONT_INFO*);
#endif

/* vfont.c */
void get_vfont(int, PREAMBLE*, FONT_INFO*);

#ifdef BEZIERFONT
/* jgfont.c */
void get_bezfont(int, PREAMBLE*, FONT_INFO*);
#endif

int k_scale  = 952;
int k_base   = 0;
int f_get_pk = 0;
BOOL f_font_c;

#ifdef FDOWN
int f_assigned;
#endif

extern char *goth;
extern int  g_font;
extern  CHAR_INFO*  char_info_ptr;
extern char ptex_d;

PREAMBLE* get_font(int, FONT_INFO*);
long jtfm_width(FONT_INFO*, int);
static BUFFER* get_pk(int, FONT_INFO*, int*);
static int read_flag_byte(BUFFER**);
static PREAMBLE* set_preamble(BUFFER**, int, FONT_INFO*, int);
static void get_kfont(int, PREAMBLE*, FONT_INFO*);
static void get_knjpk(int, PREAMBLE*, FONT_INFO*);
static void exp_1001(BUFFER*, PREAMBLE*, FONT_INFO*);
static void exp_1002(BUFFER*, PREAMBLE*, FONT_INFO*);
static void exp_1003(BUFFER*, PREAMBLE*, FONT_INFO*);
static void font_dbl(PREAMBLE*);
static BUFFER *get_pkd(FONT_INFO*, long, int *);
static BUFFER *get_jxl(long *, FONT_INFO *, long, int *);
JFM_DATA* jfm_read(FONT_INFO* font, int char_code);


PREAMBLE* get_font(int char_code, FONT_INFO* font)
{
    PREAMBLE* preamble;
    BUFFER* pk;

    int flag_byte, flag_nf;
    CHAR_INFO* char_info;
    CHAR_INFO* new_info;
    CHAR_INFO* kc_info;
    FONT_INFO* fontk;

#ifdef FDOWN
    f_assigned =
#endif
    flag_nf = 0;

#ifdef FDOWN
    if ((font->f_goth & F_DWN)
      && ((font->d_flag[char_code>>3] & (1<<(char_code & 7)))))
        f_assigned = 1;
#endif

    if ((char_info = font->char_info[char_code & (CHAR_ROOT-1)]) != NULL) {
        if (char_info->code == char_code)  return(&(char_info->pre));
        for(;;) {
            if (char_info->code < char_code) {
                if (char_info->next_char != NULL) {
                    if ((char_info=char_info->next_char)->code == char_code) {
                        return(&(char_info->pre));
                    }
                }
                else{
                    flag_nf = 1;
                    break;
                }
            }
            else{
                if (char_info->former_char != NULL) {
                    if ((char_info=char_info->former_char)->code == char_code) {
                        return(&(char_info->pre));
                    }
                }
                else{
                    flag_nf = -1;
                    break;
                }
            }
        }
    }
    pk = get_pk(char_code, font, &flag_byte);
    preamble = set_preamble(&pk, flag_byte, font, char_code);

#ifdef LBP
    if (font->font_type == JLBP)  goto c_inf;
#endif
    if (font->font_type >= JIS_FONT) {
        for(fontk = font->k_top; fontk; fontk = fontk->k_next) {
            if (fontk == font
              || (kc_info = fontk->char_info[char_code & (CHAR_ROOT-1)]) 
                == NULL) continue;
            if (kc_info->code == char_code) {
k_found:        preamble->raster = kc_info->pre.raster;
                goto    c_inf;
            }
            for(; ;) {
                if (kc_info->code < char_code) {
                    if (kc_info->next_char == NULL) break;
                    if ((kc_info=kc_info->next_char)->code == char_code)
                        goto k_found;
                }else{
                    if (kc_info->former_char == NULL) break;
                    if ((kc_info=kc_info->former_char)->code == char_code) 
                        goto k_found;
                }
            }
        }
    }
    raster_buf_alloc(&preamble);
    switch(font->font_type) {
        case(PKD_FONT):
        case(PK_FONT):  decode_pk(pk, preamble, flag_byte, font);
                        goto c_inf;
        case(PXL1001):  exp_1001(pk, preamble, font);
                        goto c_inf;
        case(PXL1002):  exp_1002(pk, preamble, font);
                        goto c_inf;
        case(PXL1003):  exp_1003(pk, preamble, font);
                        goto c_inf;
        case(JXL):      decode_jxl(pk, preamble, flag_byte, font);
                        goto c_inf;
#ifndef NOVFONT
#ifdef BEZIERFONT
        case(JG_FONT):  get_bezfont(char_code, preamble, font);
                        break;
#endif
        case(ZS_FONT):  get_vfont(char_code, preamble, font);
                        break;
#endif
        case(JIS_FONT): get_kfont(char_code, preamble, font);
                        break;
        case(JIS_PK):   get_knjpk(char_code, preamble, font);
                        break;
    }
    if (font->f_goth & K_DBL)  font_dbl(preamble);
c_inf:
    new_info = expand_char_info_list();
    new_info->code = char_code;
    new_info->next_char = new_info->former_char = NULL;
    if (flag_nf == 0) font->char_info[char_code & (CHAR_ROOT-1)] = new_info;
    else if (flag_nf > 0) char_info->next_char = new_info;
    else                  char_info->former_char = new_info;
    return(preamble);
}

static  void font_dbl(PREAMBLE *preamble)
{
    uchar   *top;
    uint    ch1, ch2;
    int     lin, col, width;
    int     off1, off2, i;

    width = (preamble->byte_width+1)/2;
    top = preamble->raster;
    off1 = ((preamble->height)/2 - 1)*width;
    off2 = (preamble->height-2)*preamble->byte_width;
    for(lin = (preamble->height)/2-1; lin >= 0; lin--) {
        for(col = preamble->byte_width-1; col >= 0; col--) {
            ch1 = top[off1 + col/2];
            ch2 = 0;
            if (col & 1) {
                for(i = 1; i < 16; i += i) {
                    ch2 >>= 2;
                    if (ch1 & i) ch2 |= 0xc0;
                }
            }
            else {
                for(i = 16; i < 256; i += i) {
                    ch2 >>= 2;
                    if (ch1 & i) ch2 |= 0xc0;
                }
            }
            top[off2+col] = top[off2+preamble->byte_width+col] = ch2;
        }
        off1 -= width;
        off2 -= preamble->byte_width*2;
    }
}

static  void exp_1001(BUFFER* pk, PREAMBLE *preamble, FONT_INFO* font)
{
    int     i, j, k;
    BUFFER *dest;
    BUFFER  *pkt;

    k = ((preamble->byte_width+3)/4)*4;
    pk = font->pk + get_long(pk)*4;
    dest = preamble->raster;
    for(i = preamble->height; i > 0; i--) {
        pkt = pk;
        for(j = preamble->byte_width; j > 0; j--) *dest++ = *pkt++;
        pk += k;
    }
}

static  void exp_1002(BUFFER* pk, PREAMBLE *preamble, FONT_INFO* font)
{
    pk = font->pk + get_long(pk);
    memcpy(preamble->raster, pk, preamble->byte_width*preamble->height);
}

static  void exp_1003(BUFFER* pk, PREAMBLE *preamble, FONT_INFO* font)
{
    BUFFER* pkt;
    uchar   flag;

    flag = *(pkt = font->pk + get_long(pk));
    if (flag & 64) {
        memcpy(preamble->raster, pkt, preamble->byte_width*preamble->height);
    } else {
        flag = ((flag << 4) & 0xff) + ((flag & 128) >> 4);
        decode_pk(pkt+1, preamble, flag, font);
    }
}

static void get_kfont(int char_code, PREAMBLE* preamble, FONT_INFO* font)
{
    int     sc;
    KFONT   *kf;

    kf = font->ext.kdir;
    sc = (((uint)char_code >> 8) - 0x21) * 94;
    sc += ((char_code & 0xff) - 0x21);
    lseek(kf->fh, (long)sc*(long)(kf->size), SEEK_SET);
    file_read(kf->fh, preamble->raster, kf->size);
}


static void get_knjpk(int char_code, PREAMBLE* preamble, FONT_INFO* font)
{
    int     sc, i;
    long    seek_pt;
    uchar   temp[3];
    KFONT   *kf;

    kf = font->ext.kdir;
    sc = (((uint)char_code >> 8) - 0x21) * 94;
    sc += ((char_code & 0xff) - 0x21);
    lseek(kf->fh, (long)sc*3+32, SEEK_SET);
    read(kf->fh, temp, 3);
    seek_pt = (((long)(temp[0] & 0x3f))<<16);
    seek_pt += (((long)(temp[1]))<<8) + temp[2];
    lseek(kf->fh, seek_pt, SEEK_SET);
    if ((temp[0] & 0x80) == 0) {
        ask_pk_buf((long)(kf->size));
        file_read(kf->fh, pk_buf_pointer->current, kf->size);
        decode_pk(pk_buf_pointer->current, preamble, (12<<4) + ((temp[0] & 0x40)>>3), font);
    }else   file_read(kf->fh, preamble->raster, kf->size);
}

static BUFFER *get_pk(int char_code, FONT_INFO* font, int* flag_ptr)
{
    int flag_byte, tmp, width, total, i, pos;
    long base_ratio;
    JFM_DATA* jtfm;
    long packet_length, id;
    BUFFER* pk;
    BUFFER* pk0;
    KFONT*  kf;
    long *jxl_info;
    uchar f_head[35];

    f_get_pk = 1;
    if (! font->pk) {
        font->pk = load_pk(font->name);

        switch(tmp = (int)*((short *)font->pk)) {
            case(0):  (BUFFER*)pk = (BUFFER*)(pk_buf_pointer->current) - 6;
                      pos = (int)get_short(pk);
                      tmp = *(((int *)font->pk) + 1);
                      if (tmp == PXL1001_ID) {
                          font->font_type = PXL1001;
                          font->ext.dir = font->pk + pos*4;
                      }
                      else {
                          if (tmp == PXL1002_ID) {
                              font->font_type = PXL1002;
                              font->ext.dir = font->pk + pos;
                          }
                          else {
                              if (tmp == PXL1003_ID) {
                                  font->font_type = PXL1003;
                                  font->ext.dir = font->pk + pos;
                              }
                              else goto err_id;
                          }
                      }
                      (BUFFER*)pk = (BUFFER*)(pk_buf_pointer->current) - 20;
                      goto checksum;

            case(JXL4_0_ID):
                      if (*(((int *)font->pk) + 1) != JXL4_1_ID) goto err_id;
                      pk = font->pk + 4;
                      font->font_type = JXL;
                      goto checksum;

            case(JTFMP_ID):
            case(JTFM_ID):
                     if (font->font_type != UNKNOWN) break;
                   /*jtfm = jfm_read(font, char_code);*/
                     jtfm = jfm_read(font, 0x3021);
                     width = sptopixel(mult20((id = jtfm->width), font->size_para));
                     base_ratio = (jtfm->depth) / ((jtfm->depth + jtfm->height)>>10);
                     base_ratio = (base_ratio*1000L)>>10;
                     tmp = sptopixel(mult20(id, 
                     font->size_para/40*k_scale));
                     font->k_width = (tmp + 12)/ 25;
                     tmp = ((tmp % 25 <= 12)? 1 : 0);
                     if (k_scale > 1000) width = font->k_width;
                     kf = open_kfont(font, width, tmp);
                     if (font->font_type != ZS_FONT
#ifdef BEZIERFONT
                        && font->font_type != JG_FONT
#endif
#ifdef LBP
                        && font->font_type != JLBP
#endif
                     ) {
                         lseek(kf->fh, 0L, SEEK_SET);
                         read(kf->fh, f_head, 35);
                         if (*((ushort *)f_head) == 0
                           && *(((ushort *)f_head)+1) == 247 + ('K'<<8)
                           && f_head[4] == kf->width) {
                             font->font_type = JIS_PK;
                             font->last = ((f_head[33]<<8)+f_head[34]-32)/3;
                             kf->height = f_head[6];
                             kf->size = (kf->width+7)/8*kf->height;
                          }
                          else {
                             font->font_type = JIS_FONT;
                             font->last = filelength(kf->fh) / kf->size;
                          }
                          font->last = ((font->last/94)<<8) + (font->last%94)+0x2120;
                          if (font->last < 0) font->last = 0x7fff;
                      }
                      else  font->last = 0x7e7e;
#ifdef LBP
                      if (font->font_type != JLBP)
#endif
                      {
                          kf->base = (int)(((1000L - k_base - base_ratio) 
                                     * kf->height + 500) / 1000L);
                          font->ext.kdir = kf;
                          if (strstr(kf->name, goth) != (char*)NULL) goto checkj;
#ifndef NOVFONT
                          if (font->font_type == ZS_FONT
#ifdef BEZIERFONT
                             || font->font_type == JG_FONT
#endif
                          ) {
                              switch (chk_vfont_goth(font)) {
                                  case 1  : font->f_goth |= (g_font = 0);
                                            break;
                                  case 2  : font->f_goth |= (g_font = 1);
                                            break;
                                  default : font->f_goth |= (g_font = chk_goth(font));
                              }
                          }
                          else  font->f_goth |= (g_font = chk_goth(font));
#else
                          (font->f_goth) |= (g_font = chk_goth(font));
#endif
                       /* printf(">>font base = %d\n", kf->base);*/
                      }
checkj:               pk = font->pk + 28;
                      goto checksum;

            case(PKDW_ID):
                      font->font_type = PKD_FONT;
                      goto check0;

            case(PKW_ID):
                      font->font_type = PK_FONT;
check0:               pk = font->pk;
                      pk += *((uchar *)(pk+2)) + 7;
checksum:             if (f_font_c && (id = get_long(pk))!= 0
                         && font->c != 0 && font->c != id)
                          error(WARNING, "check sum doesn't match in Font%3d: "
                                "dvi(%08lX) %s(%08lX)", font->font_code, font->c,
                                (font->font_type>=ZS_FONT)?"jfm":"font",id);
                      break;

            default:
err_id:               error(UNPACK_ERROR, "Illegal font ID: %s(%04X)", get_font_name(font), tmp);
        }
    }
fcode:
    pk = font->pk;
    if (font->font_type >= ZS_FONT) {
        if (char_code < 0x2121 || char_code > font->last)
no_font:    error(FONT_MISMATCH, "%5d: %s %s(%XH)", 
                font->font_code, get_font_name(font), 
                (font->font_type >=JIS_FONT)?(font->ext.kdir)->name:"",
                char_code);
        move_ptr(pk, 2);
        total = (int)get_short(pk);      /* nt */
        move_ptr(pk, 2);
        tmp = (int)get_short(pk);        /* lh */
        move_ptr(pk, (tmp + 5)<<2);
        pk0 = pk;
        while(total-- > 0) {
            if (char_code == (int)get_short(pk)) return(pk);
            move_ptr(pk, 2);
        }
        return(pk0+2);
    }

    switch(font->font_type) {
      case(PK_FONT):
        move_ptr(pk, 2);
        tmp = get_ubyte(pk);
        move_ptr(pk, tmp + 16);
        for (;;) {
            if ((flag_byte = read_flag_byte(&pk)) == PK_POST) goto no_font;
            switch (flag_byte & 7) {
            case 7 :
                /* long-preamble */
                packet_length = get_long(pk);
                if (char_code == get_long(pk)) {
                    *flag_ptr = flag_byte;
                    return(pk);
                }
                break;
            case 4 : case 5 : case 6 :
                /* ex-short-preamble */
                packet_length = /* ((flag_byte & 3) << 16) +    */
                    (int)get_short(pk);
                if (char_code == get_ubyte(pk)) {
                    *flag_ptr = flag_byte;
                    return(pk);
                }
                break;
            default :
                /* short preamble */
                packet_length = ((flag_byte & 7) << 8) +
                    get_ubyte(pk);
                if (char_code == get_ubyte(pk)) {
                    *flag_ptr = flag_byte;
                    return(pk);
                }
                break;
            }
            move_ptr(pk, packet_length);
        }

    case(PXL1001):
    case(PXL1002):
    case(PXL1003):
        return(font->ext.dir + (char_code<<4));

    case(PKD_FONT):
        if (font->ext.fh <= 0)  open_pk(font);
        if ((tmp = char_code - *(font->pk+4)) < 0
          || char_code > *(font->pk+6))
            goto no_font;
        pk = font->pk + (font->k_width + 3*tmp);
        if ((packet_length = get_mid(pk)) == 0L) goto no_font;
        return(get_pkd(font, packet_length, flag_ptr));

    case(JXL):
        if (font->ext.fh <= 0) open_jxl(font);
        jxl_info = (long *)(font->pk);
        if (char_code < 0x3021) {
            if (jxl_info[i = 13] == jxl_info[14]) goto no_font;
            tmp = 0x21;
        }
        else if (char_code < 0x5021) {
            if (jxl_info[i = 14] == 0) goto no_font;
            tmp = 0x30;
        }
        else{
            if (jxl_info[i = 15] == 0) goto no_font;
            tmp = 0x50;
        }
        pos = ((char_code >> 8) - tmp)*94 + (char_code & 0x7f) - 0x21;
        return(get_jxl(jxl_info, font, jxl_info[i] + (pos<<3),
                flag_ptr));
    }
}


static BUFFER *get_jxl(long *jxl_info, FONT_INFO *font, long pos, int *flag_ptr)
{
    int     flag_s, fd, i;
    uchar   tmp[8], tmp0[2];
    long    pos_x, pos_y, pos_tfm, pos_raster, packet_length;
    BUFFER  *pk;

    lseek(fd = font->ext.fh, pos, SEEK_SET);
    read(fd, &(tmp[0]), 8);

    pos_tfm = jxl_info[6]  + ((uint)tmp[0]<<2);
    pos_x   = jxl_info[8]  + (((uint)tmp[4]<<8) + (uint)tmp[5])*6L;
    pos_y   = jxl_info[10] + (((uint)tmp[6]<<8) + (uint)tmp[7])*6L;
    pos_raster = (((long)(tmp[1]))<<16) + (((long)(tmp[2]))<<8) + tmp[3]; 

    lseek(fd, pos_raster, SEEK_SET);
    read(fd, &(tmp[0]), 1);
    flag_s = ((*flag_ptr = tmp[0]) & 0x30) >> 4;
    if ((*flag_ptr & 0x0f) == 14) {
        lseek(fd, pos_x+2, SEEK_SET);
        read(fd, &(tmp0[0]) , 2);
        packet_length = (tmp0[0]<<8) + tmp0[1];
        lseek(fd, pos_y+2, SEEK_SET);
        read(fd, &(tmp0[0]) , 2);
        packet_length *= (tmp0[0]<<8) + tmp0[1];
        packet_length = (packet_length + 7)>>3;
        lseek(fd, pos_raster+1, SEEK_SET);
    }
    else{
        packet_length = 0;
        while(flag_s-- >= 0) {
            read(fd, &(tmp[0]), 1);
            packet_length = (packet_length<<8) + tmp[0];
        }
    }
    if (ask_pk_buf(packet_length) == FAILURE)
            error(UNPACK_ERROR, 
                "Too big raster in %s", get_font_name(font));
    file_read(fd, (pk = pk_buf_pointer->current) + 12, (int)packet_length);
    *((int *)(&(pk[packet_length+12]))) = 0x0000;
    *((int *)(&(pk[packet_length+14]))) = 0x0f0f;
    lseek(fd, pos_x, SEEK_SET);
    file_read(fd, pk, 4);
    lseek(fd, pos_y, SEEK_SET);
    file_read(fd, pk+4, 4);
    lseek(fd, pos_tfm, SEEK_SET);
    file_read(fd, pk+8, 4);

    return(pk);
}


BUFFER* get_pkd(FONT_INFO* font, long pos, int *flag_ptr)
{
    long    packet_length;
    int     fd, k;
    uchar   tmpc;

    union   NUM {
            long    l;
            uint    i;
            uchar   ch[4];
    }num;

    lseek(fd = font->ext.fh, pos, SEEK_SET);

    read(fd, &tmpc, 1);
    *flag_ptr = tmpc;
    if ((tmpc &= 7) == 7) k = 3;
    else{
        if (tmpc > 3) k = 1;
        else           k = 0;
    }
    while(k >= 0) read(fd, &(num.ch[k--]), 1);

    switch (tmpc) {

        case 7: packet_length = num.l;
                read(fd, &num.l, 3);
                break;

        case 4: case 6: case 5:
                packet_length = ((ulong)(tmpc & 3) << 16) + num.i;
                break;

        default:
                packet_length = (tmpc << 8) + num.ch[0];
                break;
    }
    read(fd, &num.l, 1);

    if (ask_pk_buf(packet_length) == FAILURE)
            error(UNPACK_ERROR, "Too big raster in %s", font->pk+5);
    file_read(fd, pk_buf_pointer->current, (int)packet_length);
    return(pk_buf_pointer->current);
}

static int read_flag_byte(BUFFER** pk_ptr)
{
    int flag_byte;
    BUFFER* pk;
    long tmp;
    
    pk = *pk_ptr;
    flag_byte = get_ubyte(pk);
    while (flag_byte >= 240 && flag_byte != PK_POST) {
        /* Skip Specials
         */
        switch (flag_byte) {
        case 240 : case 241 : case 242 : case 243 :
            for (tmp = 0; flag_byte >= 240; flag_byte--)
                tmp = (tmp << 8) + get_ubyte(pk);
            
            move_ptr(pk, tmp);
            break;
        case 244 :
            move_ptr(pk, 4);
            break;
        case 245 : case 246 :
            break;
        default :
            error(UNPACK_ERROR, "Unexpected %d !", flag_byte);
        }
        flag_byte = get_ubyte(pk);
    };
    
    *pk_ptr = pk;
    return(flag_byte);
}

static PREAMBLE* set_preamble(BUFFER** pk_ptr, int flag_byte,
    FONT_INFO* font, int char_code)
{
    PREAMBLE* preamble;
    BUFFER* pk;
    int tmp;

    preamble = alloc_preamble();
    pk = *pk_ptr;
    
    switch(font->font_type) {
      case(JIS_FONT):
      case(JIS_PK):
      case(ZS_FONT):
      case(JLBP):
#ifdef BEZIERFONT
      case(JG_FONT):
#endif
     /* preamble->x_offset = 0; */
        preamble->y_offset = (font->ext.kdir)->base;
        preamble->width    = (font->ext.kdir)->width;
        preamble->height   = (font->ext.kdir)->height;
        if (font->f_goth & K_DBL) {
            preamble->width   *= 2;
            preamble->height  *= 2;
            preamble->y_offset*= 2;
        }
        preamble->tfm_width = 
                mult20(jtfm_width(font, (int)get_short(pk)), font->size_para);
        if ((tmp = preamble->width - sptopixel(preamble->tfm_width)) == 0) {
            preamble->x_offset = 0;
        }
        else {
            if (  char_code <= 0x2125
              || (char_code <= 0x212c && char_code >= 0x212b)
              || (char_code >= 0x216b && char_code <= 0x216d)) {
                tmp = 0;
            }
            else if (char_code <= 0x215b && char_code >= 0x214a) {
                if (char_code & 1) tmp = 0;
            }
            else
                tmp /= 2;
            preamble->x_offset = tmp;
        }
        break;

    case(PK_FONT):
    case(PKD_FONT):
      switch (flag_byte & 7) {

          case 7 :
            /* long format */
            preamble->tfm_width = mult20(get_long(pk), font->size_para);
            move_ptr(pk, 8);
                /* Skip hor_esc */
            preamble->width    = (int)get_long(pk);
            preamble->height   = (int)get_long(pk);
            preamble->x_offset = (int)get_long(pk);
            preamble->y_offset = (int)get_long(pk);
            break;

          case 4 : case 5 : case 6 :
            /* ex-short format */
            preamble->tfm_width = mult20(get_mid(pk), font->size_para);
            move_ptr(pk, 2);
                /* Skip hor_esc */
            preamble->width    = (int)get_short(pk);
            preamble->height   = (int)get_short(pk);
            preamble->x_offset = (int)get_short(pk);
            preamble->y_offset = (int)get_short(pk);
            break;

          default :
            /* short format */
            preamble->tfm_width = mult20(get_mid(pk), font->size_para);
            move_ptr(pk, 1);
                /* Skip hor_esc */
            preamble->width    = (int)get_ubyte(pk);
            preamble->height   = (int)get_ubyte(pk);
            preamble->x_offset = (int)get_byte(pk);
            preamble->y_offset = (int)get_byte(pk);
            break;
        }
        break;

    case(JXL):
        preamble->x_offset = (int)get_short(pk);
        preamble->width    = (int)get_short(pk);
        preamble->y_offset = (int)get_short(pk);
        preamble->height   = (int)get_short(pk);
        preamble->tfm_width = mult20(get_long(pk), font->size_para);
        break;

    case(PXL1001):
    case(PXL1002):
    case(PXL1003):
        preamble->width    = (int)get_short(pk);
        preamble->height   = (int)get_short(pk);
        preamble->x_offset = (int)get_short(pk);
        preamble->y_offset = (int)get_short(pk);
        get_long(pk);
        preamble->tfm_width = mult20(get_long(pk), font->size_para);
        pk -= 8;
        break;
    }
    preamble->byte_width = (preamble->width + 7) / 8;
    *pk_ptr = pk;
    preamble->shift_up_ptex = preamble->shift_right_ptex = 0;
    preamble->rotate_ptex = 0;
    return(preamble);
}

long jtfm_width(FONT_INFO* font, int c_type)
{
    BUFFER* pk;
    int tmp, tmp1;

    pk = font->pk + 2;  
    tmp = (int)get_short(pk);            /* nt */
    move_ptr(pk, 2);                     /* lf */
    tmp += (int)get_short(pk);           /* lh */
    move_ptr(pk, 2);                     /* bc = 0 */
    tmp1 = (int)get_short(pk);           /* ec */
    move_ptr(pk, (tmp + 4 + c_type)<<2);
    tmp += (int)get_byte(pk);     /* width_index */
    pk = font->pk;
    move_ptr(pk, (tmp+tmp1+8)<<2);
    return(get_long(pk));
}

JFM_DATA* jfm_read(FONT_INFO* font, int char_code)
/* read JFM files properly   21 July 1992, by Naochan! */
{
    static JFM_DATA jtfm;
    BUFFER* pk;
    int i, c_type;
    int tmp, tmp_w, w_to_h, h_to_d, d_to_i, tmp2, tmp3;
    int index_w, index_h, index_d, index_i, get_byte_tmp;
    int lh, nt, ec, nw, nh, nd;

    pk = font->pk;
    tmp = get_short(pk);    /* id */
    jtfm.id = tmp;
    nt = get_short(pk);     /* nt */
    move_ptr(pk, 2);        /* skip lf  */
    lh = get_short(pk);     /* lh */
    move_ptr(pk, 2);        /* skip bc(=0) */
    ec = get_short(pk);     /* ec */
    nw = get_short(pk);     /* nw */
    nh = get_short(pk);     /* nh */
    nd = get_short(pk);     /* nd */
    
    move_ptr(pk, (lh + 3)<<2);     /* set c_type */
    c_type = get_short(pk);
    for(i=1;i<nt;i++) {
        tmp2 = get_short(pk);
        tmp3 = get_short(pk);
        if (char_code == tmp2) {
            c_type = tmp3;
            move_ptr(pk, (nt - i - 1)<<2);
            break;
        }
    }
    jtfm.char_type = c_type;
    
    move_ptr(pk, c_type<<2);
    index_w = (int)get_byte(pk);                 /*  width_index */
    get_byte_tmp = (int)get_byte(pk);
    index_h = (get_byte_tmp & 0xf0L)>>4;         /* height_index */
    index_d = get_byte_tmp & 0xfL;                 /*  depth_index */
    index_i = ((int)get_byte(pk) & 0xfcL)>>2;  /* italic_index */

    tmp_w = (lh + nt + ec + index_w + 8)<<2;
    w_to_h = (nw - index_w + index_h - 1)<<2;
    h_to_d = (nh - index_h + index_d - 1)<<2;
    d_to_i = (nd - index_d + index_i - 1)<<2;

    pk = font->pk;                /* return to the home position */
    move_ptr(pk, tmp_w);
    jtfm.width  = get_long(pk);
    move_ptr(pk, w_to_h);
    jtfm.height = get_long(pk);
    move_ptr(pk, h_to_d);
    jtfm.depth  = get_long(pk);
    move_ptr(pk, d_to_i);
    jtfm.italic = get_long(pk);

    return(&jtfm);
}
  
/* end of file getfont.c */

