/*
*       3D-Bezier JGfont for TeX Device Driver
*
*                     jgfont.c
*
*
*             12 Nov 1992  by Naochan!
*/

#include "config.h"
#include "dd.h"
#ifdef msdos
#include <dir.h>
#endif
#include "err.h"
#include "vfont.h"
#include "vraster.h"
#include "jgfont.h"

#ifdef CC_JPN
#include <jctype.h>
#else
ushort jistojms(ushort);
int jisl1(ushort);
int jisl2(ushort);
#endif
int jislgaiji(ushort);

extern V_FONT *vfont[MAX_VFONT];
extern const int s_ratio[3];

int bezier, bezmax;
LOCATE w[97], keep, kv;    /* (2^5)*3 + 1 */

/* vfont.c */
void bbox(BUFFER *, int);
V_JFM *get_vjfm(int vjfm_no);
int draw_character(VPARA_TBL *);

/* vraster.c */
int draw_fine(VPARA_TBL *);

/* vdata.c */
int get_yofs(VPARA_TBL *);

#ifndef NOTATEGAKI
void TateKanji_type(unsigned int);
#endif

/* fontdef.c */
extern char tmp_buf[];

void get_bezfont(int, PREAMBLE *, FONT_INFO *);
int get_twelve(VPARA_TBL *, GET_TEN_SW);
BEZ_LOCATE bez_locate(VPARA_TBL *);
BEZ_LOCATE bez_emulate(VPARA_TBL *);

#ifdef BEZIERFONT
/*******************************************************
 *  get_bezfont                                        *
 *******************************************************/
void get_bezfont(int char_code, PREAMBLE *preamble, FONT_INFO *font)
{
    int font_no, i, size_bytes, jis_level;
    int max_width, max_height, max_shift;
    unsigned sjis_code, utmp;
    KFONT *kf;
    BUFFER *buffer;
    V_JFM *vjfm;
    VPARA_TBL v_table;

    kf = font->ext.kdir;
    vjfm = get_vjfm(kf->fh);
    font_no = vjfm->font_no;

    sjis_code = jistojms(char_code);
#ifndef NOTATEGAKI
    TateKanji_type(sjis_code);
#endif

    v_table.v_font_name = vfont[font_no]->v_font_name;
    v_table.font_type = vfont[font_no]->font_type;
    if (vfont[font_no]->yoffset == 0x400) {
        v_table.vfn_file = vfont[font_no]->v_fh[0];    /* vf1 file */
        v_table.char_code = STD_CODE;
        utmp = STD_CODE - 0x3000;
        v_table.v_code = (utmp >> 8) * 0x5e + (utmp & 0xff) - 0x21;
        v_table.long_wide = (int)'a';
        v_table.slant = (int)'a';
        vfont[font_no]->yoffset = get_yofs(&v_table);
    }
    v_table.char_code = char_code;
    if (jisl2(sjis_code) != 0 || jislgaiji(sjis_code) != 0) {
        jis_level = 1;
        utmp = char_code - 0x5000;
    }
    else if (jisl1(sjis_code) != 0) {
        jis_level = 0;
        utmp = char_code - 0x3000;
    }
    else {
        /* not KANJI */
        jis_level = 2;
        utmp = char_code - 0x2100;
    }
    v_table.v_code = (utmp >> 8) * 0x5e + (utmp & 0xff) - 0x21;
    v_table.vfn_file = vfont[font_no]->v_fh[jis_level];
    v_table.long_wide = vjfm->long_wide;
    v_table.slant = vjfm->slant;
    v_table.width = kf->width;
    v_table.height = kf->height;
    v_table.yoffset = vfont[font_no]->yoffset;
    v_table.thin = vjfm->thin;
    v_table.xfat = vjfm->_xfat;
    v_table.yfat = vjfm->_yfat;
    v_table.rotation = vjfm->rotation;

    max_width = 0x3ff;            /* 10bit-max */
    max_height = 0x3ff;
    max_shift = s_ratio[(v_table.slant - (int)'b') % 3];

    if ((int)'b' <= v_table.slant && v_table.slant <= (int)'g')
        max_width += max_shift;
    else if ((int)'h' <= v_table.slant && v_table.slant <= (int)'m')
        max_height += max_shift;

    v_table.max_width = max_width;
    v_table.max_height = max_height;

    size_bytes = (v_table.width + 7) / 8 * v_table.height;
    v_table.buffer = buffer = preamble->raster;
    if (v_table.vfn_file == -1) {
        sprintf(tmp_buf, "[JGfont] JIS level 2 is used. %s(%XH)",
                font->name, char_code);
        error(WARNING, tmp_buf);
        bbox(buffer, size_bytes);
        return;
    }

    for (i = 0; i < size_bytes; i++)
        buffer[i] = 0;

    if ((vjfm->draw_sw == (int)'n' || vjfm->draw_sw == (int)'f')
        && v_table.width < vjfm->threshold) {
        if (draw_fine(&v_table) == FAILURE)
            bbox(buffer, size_bytes);
    }
    else {
        if (vjfm->draw_sw == (int)'o')
            v_table.dtype = VTRACE;
        else if (vjfm->draw_sw == (int)'f')
            v_table.dtype = FILL;
        else
            v_table.dtype = BOTH;

        if (draw_character(&v_table) == FAILURE)
            bbox(buffer, size_bytes);
    }
}

/**********************************
 *  get_twelve                    *
 **********************************/
int get_twelve(VPARA_TBL *tbl_ptr, GET_TEN_SW get_ten_sw)
{
    static short bufx[VF_BUF_SIZE];
    static int remaining_bits;
    static int remaining_words;
    static int remaining_words_save;
    static int buf_pos;
    static union {
        short s[2];
        long l;
    } d;
    static int buffering = 0;

    if (get_ten_sw == GET_DATA) {
        if (remaining_bits < 12) {
            d.s[1] = d.s[0];
            if (remaining_words == 0) {
                remaining_words
                    = (read(tbl_ptr->vfn_file, bufx, VF_BUF_SIZE * 2)) / 2;
                remaining_words_save = remaining_words;
                buffering++;
                if (remaining_words == 0) {
                    sprintf(tmp_buf,
                            "[JGfont] 12 bits data error.(%08x) %s",
                            tbl_ptr->char_code, tbl_ptr->v_font_name);
                    error(WARNING, tmp_buf);
                    return (-1);
                }
                buf_pos = 0;
            }
            d.s[0] = bufx[buf_pos++];
            remaining_words--;
            remaining_bits += 16;
        }
        remaining_bits -= 12;
        return ((int)((d.l >> remaining_bits) & 0xfff));
    }
    else if (get_ten_sw == KEEP_BUF && buffering == 1) {
        remaining_words = remaining_words_save;
        buf_pos = 0;
        buffering = 0;
        return (-1);
    }
    else {
        remaining_bits = 0;
        remaining_words = 0;
        buffering = 0;
        return (-1);
    }
}

BEZ_LOCATE bez_locate(VPARA_TBL *tbl_ptr)
{
    int x, y;
    char type;
    BEZ_LOCATE bl;

    x = get_twelve(tbl_ptr, GET_DATA);
    y = get_twelve(tbl_ptr, GET_DATA);
    if (x == -1 || y == -1) {
        bl.type = BEZ_ERROR;
        return bl;
    }
    bl.x = 0x200 + ((x & 0x0400) ? -(x & 0x03ff) : (x & 0x03ff));
    bl.y = 0x200 - ((y & 0x0400) ? -(y & 0x03ff) : (y & 0x03ff));

    if (x & 0x800)
        bl.type = BEZ_BEZIER;
    else
        bl.type = BEZ_LINE;
    if (x == 0xfff)
        bl.type = BEZ_STOP;

    return bl;
}

BEZ_LOCATE bez_emulate(VPARA_TBL *tbl_ptr)
{
    int i, j, tmpi, tmpj, bez_emu_dimen, target_size;
    long l0, l1, s;
    LOCATE tmp1, tmp2, tmp3;
    BEZ_LOCATE bl, rv;

    target_size = tbl_ptr->width;
emu_rep:
    if (bezier > 0) {
        rv.x = w[(bezmax - bezier + 1) * 3].x;
        rv.y = w[(bezmax - bezier + 1) * 3].y;
        rv.type = BEZ_LINE;
        bezier--;
    }
    else {
        bl = bez_locate(tbl_ptr);
        switch (bl.type) {
          case BEZ_LINE:
              keep.x = rv.x = bl.x;
              keep.y = rv.y = bl.y;
              rv.type = BEZ_LINE;
              break;
          case BEZ_BEZIER:
              w[0].x = keep.x;
              w[0].y = keep.y;
              w[1].x = bl.x;
              w[1].y = bl.y;
              bl = bez_locate(tbl_ptr);
              w[2].x = bl.x;
              w[2].y = bl.y;
              bl = bez_locate(tbl_ptr);
              w[3].x = bl.x;
              w[3].y = bl.y;
              keep.x = w[3].x;
              keep.y = w[3].y;
              l0 = (long)(w[1].x - w[0].x) * (w[2].y - w[0].y)
                  - (long)(w[2].x - w[0].x) * (w[1].y - w[0].y);
              l1 = (long)(w[2].x - w[0].x) * (w[3].y - w[0].y)
                  - (long)(w[3].x - w[0].x) * (w[2].y - w[0].y);
              s = ((l0 < 0L) ? -l0 : l0) + ((l1 < 0L) ? -l1 : l1);
              s = (s * target_size) >> 10;
              s = (s * target_size) >> 11;
              if (s < 4L)         bez_emu_dimen = 0;
              else if (s < 16L)   bez_emu_dimen = 1;
              else if (s < 64L)   bez_emu_dimen = 2;
              else if (s < 256L)  bez_emu_dimen = 3;
              else if (s < 1024L) bez_emu_dimen = 4;
              else                bez_emu_dimen = 5;
              for (i = 0; i < bez_emu_dimen; i++) {
                  for (j = 1 << i; j > 0; j--) {
                      tmpi = (j - 1) * 3;
                      tmpj = tmpi << 1;
                      tmp1.x = w[tmpi].x + w[tmpi + 1].x;
                      tmp1.y = w[tmpi].y + w[tmpi + 1].y;
                      tmp2.x = w[tmpi + 1].x + w[tmpi + 2].x;
                      tmp2.y = w[tmpi + 1].y + w[tmpi + 2].y;
                      tmp3.x = w[tmpi + 2].x + w[tmpi + 3].x;
                      tmp3.y = w[tmpi + 2].y + w[tmpi + 3].y;
                      w[tmpj + 6].x = w[tmpi + 3].x;
                      w[tmpj + 6].y = w[tmpi + 3].y;
                      w[tmpj + 1].x = (tmp1.x + 1) >> 1;
                      w[tmpj + 1].y = (tmp1.y + 1) >> 1;
                      w[tmpj + 2].x = (tmp1.x + tmp2.x + 2) >> 2;
                      w[tmpj + 2].y = (tmp1.y + tmp2.y + 2) >> 2;
                      w[tmpj + 3].x = (tmp1.x + (tmp2.x << 1) + tmp3.x + 4) >> 3;
                      w[tmpj + 3].y = (tmp1.y + (tmp2.y << 1) + tmp3.y + 4) >> 3;
                      w[tmpj + 4].x = (tmp2.x + tmp3.x + 2) >> 2;
                      w[tmpj + 4].y = (tmp2.y + tmp3.y + 2) >> 2;
                      w[tmpj + 5].x = (tmp3.x + 1) >> 1;
                      w[tmpj + 5].y = (tmp3.y + 1) >> 1;
                  }
              }
              rv.x = w[3].x;
              rv.y = w[3].y;
              rv.type = BEZ_LINE;
              bezier = (bezmax = 1 << bez_emu_dimen) - 1;
              break;
          default:
              rv.type = BEZ_STOP;
              kv.x = kv.y = 0;
              return rv;
        }
    }
    /*  rv.x = (int)(((long)(rv.x) * target_size + 512L) >> 10);
    rv.y = (int)(((long)(rv.y) * target_size + 512L) >> 10);
    if (rv.x == kv.x && rv.y == kv.y) goto emu_rep;
    kv.x = rv.x; kv.y = rv.y; */
    return rv;
}

#endif
/* end of jgfont.c */
