/*
 *      knjtopk.exe    KANJI FONT file <-> PK Format
 *                      supported dviout.exe dviprt.exe v2.07-
 *              v1.0    June 1990, written by SHIMA
 *              v1.1    Dec. 1990, The last kaji (JIS 737E) was ignored
 *              v1.2    Dec. 1990, Support PK format -> KANJI FONT file
 *              v1.3    Dec. 1990, o & . + option
 *              v1.4    Feb. 1991, MSG 8424 -> 7424
 *              v1.5    Aug. 1991, enlarge *_BUF_SIZE
 *
 *      modified for Minix-386 1992 by youchan
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>


#define KNJ_PK_ID       (247*256 + 'K')

#define REPT            14
#define DYN_F           12

#define DYN_2           ((12 - DYN_F)*16 + 15 + DYN_F + 1)      /* 28 */
#define DYN_L           ((13 - DYN_F)*16 + DYN_F - 15)          /* 13 */

#define FONT_BUF_SIZE   5000
#define WORK_BUF_SIZE   11000

#define MAX_CODE        0x7424
#define MAX_CHAR    ((((unsigned int)MAX_CODE>>8)-0x21)*94 + (MAX_CODE & 0xff) - 0x20)
                    /* -0x21 -> -0x20  in v1.1 */

unsigned char   f_buf[FONT_BUF_SIZE];
unsigned int    w_buf[WORK_BUF_SIZE];
unsigned char   *f_dir;
unsigned char   f_head[32];
unsigned char   t_head[32];
unsigned char   row[240];
int             power[8];
int             gpower[9];

unsigned int    max_char = MAX_CHAR;

static  unsigned int    bit[8] = { 128, 64, 32, 16, 8, 4, 2, 1 };

void    set_num( unsigned int**, int );
int     rpt_cnt( int, int, int );
int     pxltopk( int, int );
void    usage( void );
int     get_bit( void );
void    decode_pk( int, int, int, int );
int     get_pknum( void );
void    printx( void );

void main( int argc, char** argv )
{
    FILE    *fp, *fpk;
    int     i, j, k, size, block;
    int     width, height, b_width, f_deb, f_rev, max_num;
    long    pos;

    if( argc < 4 ) usage();

    if( tolower(*argv[3]) == 'r' ){
        f_rev = 1;
    }else{
        f_rev = 0;
        width = atoi( argv[3] );
        if( argc >= 5 ) height = atoi( argv[4] );
        else height = width;
    }
    if( f_rev == 0 && (width == 0 || height == 0) ) usage();
    
    f_deb = 0;
    for( i = 5; i < argc; i++ ){
        if( tolower(argv[i][0]) == 'f' ){
            fprintf( stderr, "Do you really want 'f' option (it becomes very slow)? "
                    "Input (y/n):" );
            fflush( stderr );
            while( (i = tolower( getc(stdin) )) != 'y' && i != 'n' );
            if( i == 'n' ) {
                fprintf( stderr, "\n");
                exit(1);
            }
            f_deb = 1;
        }
        if( argv[i][0] == '+' ){
            max_char += atoi( argv[i] + 1 );
        }
    }

    if( (fp = fopen( argv[1], "r" )) == NULL ){
        fprintf( stderr, "Cannot open [%s]\n", argv[1] );
        exit( 1 );
    }

    if( (fpk = fopen( argv[2], "w" )) == NULL ){
        fprintf( stderr, "Cannot open [%s]\n", argv[2] );
        exit( 1 );
    }

    fread( t_head, 32, 1, fp );
    if( t_head[0] == 0
        && t_head[1] == 0
        && t_head[2] == ((KNJ_PK_ID >> 8) & 0xff)
        && t_head[3] == (KNJ_PK_ID & 0xff) ){
            if( f_rev == 0 ){
                fprintf( stderr, "%s is a Kanji PK Font file.\n", argv[1] );
                exit( 1 );
            }
        width   = t_head[4];
        height  = t_head[6];
        b_width = t_head[8];
        size    = b_width*height;

        fread( row, 3, 1, fp );

        power[ 0 ] = 1;
        for( i = 1; i <= 7; i++ )
            power[ i ] = power[ i - 1 ] * 2;
        gpower[ 0 ] = 0;
        for( i = 1; i <= 8; i++ )
            gpower[ i ] = gpower[ i - 1 ] + power[ i - 1 ];

        pos  = (((long)(row[0] & 0x3f))<<16);
        pos += (((long)(row[1]))<<8) + row[2];
        max_num = pos - 32L;
        if( max_num == (max_char-5)*3 ){
            fprintf( stderr, "This file is made by Ver 1.0.\n" );
        }
        else if( max_num == (max_char-4)*3 ){
            fprintf( stderr, "This file is made by Ver 1.1 or 1.2.\n" );
        }
        if( (f_dir = malloc( max_char*3 )) == NULL ){
            fprintf( stderr, "Not enough memory!\n" );
            exit( 1 );
        }
        fseek( fp, 32L, SEEK_SET );
        fread( f_dir, max_num, 1, fp );
        printx();
        for( i = 0; i < max_num; i+=3 ){
            pos  = (((long)(f_dir[i] & 0x3f))<<16);
            pos += (((long)(f_dir[i+1]))<<8) + f_dir[i+2];
            fseek( fp, pos, SEEK_SET );

            if( (f_dir[i] & 0x80) == 0 ){
                fread( f_buf, size, 1, fp );
                decode_pk( width, height, b_width, 
                    (12<<4) + ((f_dir[i] & 0x40)>>3) );
            }else{
                fread( w_buf, size, 1, fp );
            }
write:      if( fwrite( w_buf, size, 1, fpk ) != 1 ) goto w_err;
            if( i%(100*3) == 0 )  {
                fprintf( stderr,"o" );
                fflush( stderr );
            }
        }
        if( max_num < max_char*3 ){
            max_num += 3;
            for( j = 0; j < size; j +=2 ) w_buf[j] = 0;
            goto write;
        }
	fprintf( stderr,"\nConvert is complete.\n" );
        exit( 0 );
    }
    if( f_rev == 1 ){
        fprintf( stderr, "%s is not a Kanji PK Font file.\n", argv[1] );
        exit( 1 );
    }

    if( (f_dir = malloc( max_char*3 )) == NULL ){
        fprintf( stderr, "Not enough memory!\n" );
        exit( 1 );
    }
    b_width = (width+7)/8;

    if( (block = b_width*height) > FONT_BUF_SIZE ){
        fprintf( stderr, "Too big FONT size\n" );
        exit( 1 );
    }

    f_head[2] = (KNJ_PK_ID >> 8);
    f_head[3] =  KNJ_PK_ID & 0xff;
    f_head[4] = width;
    f_head[6] = height;
    f_head[8] = b_width;

    pos = 32L+(max_char*3);

    if( fwrite( f_head, 32, 1, fpk ) != 1 
     || fwrite( f_dir, max_char, 3, fpk ) != 3 ){
w_err:  fprintf( stderr, "Write error!\n" );
        exit( 1 );
    }
    fseek( fp, 0L, 0 );
    if( f_deb == 0 ) printx();
    for( i = 0; i < max_char; i++ ){
        fread( f_buf, b_width, height, fp );
        k = i*3;
        f_dir[k++] = (pos >> 16);
        f_dir[k++] = ((pos >> 8 ) & 0xff);
        f_dir[k] = (pos & 0xff);
        size = to_pk( width, height );

        if( f_deb != 0 ){
            k = width*height;
            for( j = 0; j < k; j++ ){
                if( j%width == 0 ) fprintf( stderr,"\n" );
                if( (f_buf[j>>3] & bit[j & 7]) != 0 ) fprintf( stderr,"*" );
                else fprintf( stderr," " );
            }
            fprintf( stderr,"\n" );
            for( k = 0; k < size; k++ ){
                fprintf( stderr,"%d ", w_buf[k] );
            }
            fprintf( stderr,"[%d]\n", size );
        }

        j = (size+1)/2;
        if ( j < block ){
            if ( (f_buf[0] & 0x80) != 0 ) f_dir[3*i] |= 0x40;
            for ( k = 0; k < j; k++ ){
                f_buf[k] = (w_buf[k+k]<<4) + w_buf[k+k+1];
            }
            if ( fwrite( f_buf, j, 1, fpk ) != 1 ) goto w_err;
            pos += j;
        } else {
            f_dir[3*i] |= 0x80;
            if ( fwrite( f_buf, b_width, height, fpk ) != height ) goto w_err;
            pos += block;
        }
        if ( f_deb == 0 && i%100 == 0 ) {
            fprintf( stderr,"o" );
            fflush( stderr );
        }
    }
    fseek( fpk, 32L, 0 );
    if ( fwrite( f_dir, max_char, 3, fpk ) != 3 ) goto w_err;
    if ( pos >= 0x400000L ){
        fprintf( stderr, "Too big FONT file\n" );
        exit( 1 );
    }
    fprintf( stderr,"\nConvert is complete.\n");
}

void    printx()
{
    int i;

    for( i = 0; i <= max_char/100; i++ ) fprintf( stderr,"." );
    fprintf( stderr,"\r" );
    fflush( stderr );
}

void    set_num( unsigned int **ptr, int num )
{
    int i, count, temp[4];

    if( num <= DYN_F ){
        *((*ptr)++) = num;
    }else{
        if( num <= DYN_2 ){
            num -= (DYN_F+1);
            *((*ptr)++) = num/16 + (DYN_F+1);
            *((*ptr)++) = (num & 15);
        }
        else{
            for( num -= DYN_L, count = 0 ; num != 0; num /= 16 ){
                temp[count++] = (num & 15);
            }
            for( i = count; i > 1; i-- ) *((*ptr)++) = 0;
            while( count-- > 0 ) *((*ptr)++) = temp[count];
        }
    }
}


int     rpt_cnt( int c_len, int len, int width )
{
    unsigned char *f_buf0, *f_buf1;
    int     i, len_pt;

    f_buf0 = f_buf + width*c_len;

    for( len_pt = c_len + 1; len_pt < len; len_pt++ ){
        f_buf1 = f_buf + width*len_pt;
        for( i = 0; i < width; i++ ){
            if( f_buf0[i] != f_buf1[i] ) goto quit;
        }
    }
quit:   return( len_pt - c_len - 1 ) ;
}

int     to_pk( int width, int len )
{
    unsigned char *ptr;
    unsigned int  *w_ptr;
    int data, total, shift, b_wid, w_b_ptr, c_len, c_wid, b_cnt, r_cnt, f_b;

    ptr = f_buf;
    w_ptr = w_buf;
    b_cnt = 0;
    b_wid = (width+7)/8;
    total = b_wid*len;
    f_b = ((*ptr & 0x80) == 0 )? 0:1;
    for( c_len = 0; c_len < len; c_len++ ){
        for( r_cnt = -1, c_wid = 0; c_wid < width; c_wid++ ){
            if( (shift = (c_wid & 7)) == 0 ){
                data = *ptr++;
            }
            if( (data & bit[shift]) != 0 ){
                if( f_b != 0 ) b_cnt++;
                else{
                    f_b = 1;
                    goto rep_end;
                }
            }else{
                if( f_b == 0 ) b_cnt++;
                else{
                    f_b = 0;
rep_end:            set_num( &w_ptr, b_cnt );
                    b_cnt = 1;
                    if(   r_cnt < 0
                      && (r_cnt = rpt_cnt( c_len, len, b_wid )) != 0 ){
                                            /* repeated same lines */
                        if( r_cnt == 1 ) *w_ptr++ = 15;
                        else{
                            *w_ptr++ = REPT;
                            set_num( &w_ptr, r_cnt );
                        }
                        ptr += r_cnt*b_wid;
                        c_len += r_cnt;
                    }
                }
            }
        }       /* end a line */
        if( (w_ptr - w_buf) >= total*2 ) return( w_ptr - w_buf );
    }
    set_num( &w_ptr, b_cnt );
    w_ptr[1] = 0;
    return( w_ptr - w_buf );
}

void    usage()
{
    printf(
    "<<< knjtopk: kanji font <-> PK file, supported by dviprt, dviout v2.07- >>>\n"
    "\t\t\t\tver 1.5, 1990-91, written by SHIMA\n\n"
    "Usage: knjtopk original_file pk_file width [height [+number] [f]] (Make PK file)\n"
    "       knjtopk pk_file original_file r\t\t\t (Recover original file)\n\n"
    "original_file: JIS CODE 2121-7424 (default)\n"
    "width        : width by dots\n"
    "height       : height by dots\n"
    "number       : End of JIS Code = 7424 + number\n"
    "\t\t-4 (JIS 78), 0 (JIS 83), 2 (JIS 90)\n"
    "f            : flag for debug\n"
    );
    exit( 1 );
}

int     bit_weight, input_byte, buf_pt, repeat_count;


#define get_ubyte() f_buf[buf_pt++]

#define get_nyb() ((bit_weight^=1)?((input_byte=get_ubyte( ))>>4):(input_byte & 0x0f))

int     dyn_f;
typedef enum { FALSE = 0, TRUE = 1 } BOOL;

void decode_pk( int c_width, int c_height, int byte_width, int flag_byte )
{
    BOOL    turn_on;        /* black or white */
    unsigned char   *buf;
    int     h, v;
    int r_p,
        h_bit,
        rows_left,
        byte,
        byte_weight,
        count;

    bit_weight = buf_pt = 0;
    buf = (unsigned char*)w_buf;
    dyn_f = flag_byte >> 4;
    turn_on = ( ( flag_byte & 8 ) != 0 );

    if ( dyn_f == 14 ){
        for ( v = 1; v <= c_height; v++ ) {
            byte = 0;
            byte_weight = 7;
            for ( h = 1; h <= c_width; h++ ) {
                if ( get_bit() )    byte |= power[byte_weight];
                byte_weight--;
                if ( byte_weight == -1 ) {
                    *buf++ = byte;
                    byte = 0;
                    byte_weight = 7;
                }
            }
            if ( byte_weight < 7 )  *buf++ = byte;
        }
    }
    else{
        rows_left = c_height;
        h_bit = c_width;
        repeat_count = 0;
        byte_weight = 8;
        byte = 0;
        r_p = 1;
    
        while ( rows_left > 0 ) {
            count = get_pknum();
            while ( count > 0 ) {
                if ( ( count < byte_weight ) && ( count < h_bit ) ) {
                    if ( turn_on )
                        byte |= ( gpower[ byte_weight ] -
                            gpower[ byte_weight-count ] );
                
                    h_bit -= count;
                    byte_weight -= count;
                    break;
                }
                else if ( ( count >= h_bit ) &&
                    ( h_bit <= byte_weight ) ) {
                    if ( turn_on )
                        byte |= ( gpower[ byte_weight] -
                            gpower[ byte_weight-h_bit ] );
                
                    row[ r_p ] = byte;
                    
    
                    for ( v = 0; v <= repeat_count; v++ )
                        for ( h = 1; h <= byte_width; h++ )
                            *buf++ =row[ h ];

                    rows_left = rows_left - repeat_count - 1;
                    repeat_count = 0;
                    r_p = 1;
                    byte = 0;
                    byte_weight = 8;
                    count -= h_bit;
                    h_bit = c_width;
                }
                else {
                    if ( turn_on )
                        byte |= gpower[ byte_weight ];
                
                    row[ r_p ] = byte;
                    r_p++;
                    byte = 0;
                    count -= byte_weight;
                    h_bit -= byte_weight;
                    byte_weight = 8;
                }
            }
            turn_on = ( ! turn_on );
        }
        if ( ( rows_left != 0 ) || ( h_bit != c_width ) ){
            printf( "More bits than required!" );
            exit( 1 );
        }
    }
}


int get_bit( void )
{
    if ( bit_weight == 0 ) {
        input_byte = get_ubyte();
        bit_weight = 8;
    }
    return( ( input_byte & power[ --bit_weight ] ) != 0 );
}

int get_pknum( void )
{
    int i;
    int j;
    
    i = get_nyb();
    if ( i == 0 ) {
        do {
            j = get_nyb();
            i++;
        } while ( j == 0 );
        while ( i > 0 ) {
            j = ( j << 4 ) + get_nyb();
            i--;
        }
        return( j - 15 + ( ( 13 - dyn_f ) << 4 ) + dyn_f );
    }
    else if ( i <= dyn_f )
        return( i );
    else if ( i < 14 )
        return( ( ( i - dyn_f - 1 ) << 4 ) + get_nyb() + dyn_f + 1 );
    else {
        if ( i == 14 )
            repeat_count = get_pknum();
        else
            repeat_count = 1;
        
        return( get_pknum() );
    }
}
