/*
 * Routine to use scalable/downloaded font of LIPS3 (dviprt)
 *                       30 June    1992, written by SHIMA
 * Modifyed by OkI for LIPS3 & ESC/Page 14 Nov. 1992.
 */

#include "config.h"

#ifdef LIPS3
#include "dd.h"
#include "prtout.h"
#include "err.h"
#include "prtctl.h"
#include "lips3.h"

/* lips3.c local */
static void pr_attr(int flag);
static void pr_normal(int flag);
static uint change_code(uint code);

/* p_out.c function */
static void lips_init(void);
static void lips_page_format(int f_rotate, int f_copy);
static void lips_end(void);
static void set_lips3_font(FONT_INFO *, KFONT *, int, int, int);
static void move_cap(PIXEL, PIXEL);
static void move_homepoint(PIXEL, PIXEL);
static void pre_put_char(FONT_INFO *, uint *);
static void put_jis_font(uint);
static void put_dl_font(uint);
static long get_dlhsize(int num);
static void pre_download(FONT_INFO *, DL_INFO *, int, int);
static void download_rastor(FONT_INFO *, DL_INFO *, uint, int);
static void post_download(FONT_INFO *, DL_INFO *);
static void lips_drawbox(PIXEL, PIXEL, PIXEL, PIXEL, int);
static void nop();

#ifdef FDOWN
static int e_lbp_font = -1;    /* current font in G0 */
#endif
static int j_lbp_font = -1;    /* current font in G1 */

extern char ptex_d;            /* ptex direction: 0 = yoko, 1 = tate */
extern int last_dir;
extern long ptex_dpi_adj;

/* externals from prtinit.c */
extern char *youshi;

#ifndef NOTATEGAKI
extern int tategaki;
#endif

#ifdef FDOWN
extern int maxdwidth;     /* download font width limit */
extern int maxdheight;    /* download font height limit */
extern int maxdfont;      /* num of kinds of download limit*/
#endif

extern int weight_org;
extern int k_base_l;
extern int k_base_d;
extern int k_scale_l;
extern int lbp_orig_dpi;
extern char str_exp[];

/* externals from pret.c */
extern int g_font;

PREAMBLE *get_font(uint, FONT_INFO *);
static uint change_code(uint code);

/* p_out.c routine call */
void pr_putnc(char *, int);
void pr_puts(char *);
void pr_putsn(char *, int);
void pr_putc(char);
void pr_int2(uint num);

static struct LIPS_LFONT lips_data[] =
{
    {"Mincho-Medium.J83", 80},
    {"Gothic-Medium.J83", 81},
    {"User", 80}
};

LBP_FUNC lips3_func =
{
    lips_init, lips_page_format, lips_end, set_lips3_font,
    move_cap,
/*  move_cap_right,  */
/*  move_cap_below,  */
/*  move_cap_home,   */
    move_homepoint,
    pre_put_char, put_jis_font, put_dl_font, get_dlhsize,
    pre_download, download_rastor, post_download,
    lips_drawbox, lips_data
};

static void nop()
{
  /* NOP */
}

static void pr_attr(int flag)
{
    if (flag & F_GOTH)    pr_puts(BOLD);
    if (flag & F_SHADOW)  pr_puts(SHADOW);
    if (flag & F_OUTLINE) pr_puts(OUTLINE);
    if (flag & F_FILL)    pr_puts(PFILL);
    if (flag & F_GGOTH)   pr_puts(BBOLD);
}

static void pr_normal(int flag)
{
    if (flag & (F_GOTH | F_GGOTH)) pr_puts(NO_BOLD);
    if (flag & F_SHADOW)           pr_puts(NO_SHADOW);
    if (flag & F_OUTLINE)          pr_puts(NO_OUTLINE);
    if (flag & F_FILL)             pr_puts(NO_PFILL);
}

void lips_end(void)
{
}

void lips_init(void)
{
    char *s;
    int paper, yoko, tate, houkou, kyuushi;

    if (k_base_l == 10000)   k_base_l = K_BASE_L;
    if (k_scale_l == 10000)  k_scale_l = K_SCALE_L;
    if (lbp_orig_dpi == 0)   lbp_orig_dpi = LBP_ORIG_DPI;
    k_base_d = K_BASE_D;
    weight_org = WEIGHT;

#ifdef FDOWN
    maxdwidth = MAXDWIDTH;      /* maximal width of download font cell */
    maxdheight = MAXDHEIGHT;    /* maximal height of download font cell */
    maxdfont = MAXDFONT;        /* maximal number of download font sets */
#endif

    pr_puts("\x1b" "%@");        /*  3 bytes: Text mode of LIPS */
    pr_putsn("\x90" "31;%u;1J\x9c", lbp_orig_dpi);
    /* 11 bytes: Start job: LIPS3, JIS code*/
    pr_puts(
        "\x1b" "<"        /*  2 bytes: Software reset */
        "\x9b" "11h"      /*  4 bytes: Unit is size */
        "\x9b" "7 I"      /*  4 bytes: Size parameter is dot */
        "\220" "2y"       /*  3 bytes: Display on LCD */
        "DVI printing"    /* 12 bytes: Message */
        "\234"            /*  1 byte : */
    );
    s = youshi;
    /* Default : use installed cassette */
    yoko = tate = houkou = kyuushi = paper = 0;
    while (*s) {
        switch (*(s++)) {
          case 'A':
          case 'a':
              if (*(s++) == '5') paper = 16;    /* A5 portrait */
              else               paper = 14;    /* A4 portrait */
              while (isdigit(*s)) s++;
              break;
          case 'B':
          case 'b':
              if (*(s++) == '4') paper = 24;    /* B4 portrait */
              else               paper = 26;    /* B5 portrait */
              while (isdigit(*s)) s++;
              break;
          case 'H':            /* Postcard */
          case 'h':
              paper = 18;
              while (isdigit(*s)) s++;
              break;
          case 'T':            /* Set user number */
          case 't':
              paper = 0;
              while (isdigit(*s)) {
                  paper = paper * 10 + (*s - '0');
                  s++;
              }
              break;
          case 'F':            /* Free size */
          case 'f':
              paper = 80;
              yoko = tate = 0;
              while (isdigit(*s)) {
                  yoko = yoko * 10 + (*s - '0');
                  s++;
              }
              if (*s == ':') {
                  s++;
                  while (isdigit(*s)) {
                      tate = tate * 10 + (*s - '0');
                      s++;
                  }
              }
              else tate = yoko;
              break;
          case 'L':            /* Landscape */
          case 'l':
              houkou = 1;
              kyuushi = 0;
              while (isdigit(*s)) {
                  kyuushi = kyuushi * 10 + (*s - '0');
                  s++;
              }
              break;
          case 'P':            /* portait */
          case 'p':
              houkou = 0;
              kyuushi = 0;
              while (isdigit(*s)) {
                  kyuushi = kyuushi * 10 + (*s - '0');
                  s++;
              }
              break;
        }
    }
    if (houkou) paper++;                /* portrait --> landscape */
    pr_puts("\x9b" "?2;3;23;25h");      /* 12 bytes: No auto-FF, No move CAP,
                                                 Poor man's bold */
    pr_puts("\x9b" "?1;4;5;6;27l");     /* 13 bytes: No auto-LF, up margine,
                                                 LF without CR, FF without CR
                                                 No original attribute */

    sprintf(str_exp, "\x9b" "%u;%u;%up", paper, yoko, tate);
    pr_puts(str_exp);
    pr_putsn("\x09b" "%uq", kyuushi);
    pr_puts("\x09b" "8;0 G");           /* 6 bytes: VMI = 8, HMI = 0 */
#ifndef NOTATEGAKI
    if (tategaki) pr_puts("\x9b" "?29h");
#endif
    ptex_dpi_adj = (long)lbp_orig_dpi *100L;
}

static int graytab[] = {9, 12, 13, 3};
static void lips_drawbox(PIXEL vs, PIXEL hs, PIXEL height, PIXEL width, int tone)
{
    int graylevel;

    graylevel = graytab[tone];
    move_cap(vs, hs);
    pr_puts("\x9b{");
    move_cap(vs - height, hs + width);
    pr_putsn("\x9b;;;%d}", graylevel);
}

static void set_lips3_font(FONT_INFO *fontc, KFONT *kf, int attr, int weight, int face)
{
    /* select font */
    g_font = (fontc->f_goth & (F_ATTR | F_LBP));
    pr_putc(SHIFTOUT);                  /* call G1 */
    pr_puts(G1GRPH2);                   /* select Graphic Set */
    pr_putc(JIS83 - 0x80);
    pr_putsn(CHR_PITCH, kf->height);    /* select pitch */
    pr_putsn(CHR_SIZE, kf->size);       /* select point */
    pr_putsn(FWEIGHT, weight);          /* select weight, style=23 */
    pr_putsn(FACE, face);               /* select face */
    pr_putsn(ASSIGN_G1, fontc->font_code);    /* assign font_number */
    pr_putc(SHIFTIN);                   /* recover G0 */
    j_lbp_font = -1;
}

static PIXEL v_offset = 0, h_offset = 0; /* Refered by move_cap & move_homepoint */
static PIXEL last_V = 0, last_H = 0;     /* Last CAP position */

/* debugged by M.Izumi, modified by OkI */
static void move_cap(PIXEL v, PIXEL h)
{
    if (v == -1) {                /* reset last CAP position */
        last_V = last_H = -1;
        return;
    }
    v += v_offset;
    h += h_offset;
    if (v == last_V && h > last_H)
        sprintf(str_exp, MOVE_RIGHT, h - last_H);
    else if (h == last_H && v > last_V)
        sprintf(str_exp, MOVE_BELOW, v - last_V);
    else {
        sprintf(str_exp, MOVE_CAP, v, h);
        /* When v==0 (h==0), CAP moves the posion of v==1(h==1) not 0.
       Then relative move of -1 dot is required after then.         */
        if (v == 0)  strcat(str_exp, MOVE_ABOVE1);
        if (h == 0)  strcat(str_exp, MOVE_LEFT1);
    }
    pr_puts(str_exp);
    last_V = v;
    last_H = h;
}

static void move_homepoint(PIXEL v, PIXEL h)
{
    v_offset = 0;
    h_offset = 0;
    move_cap(v, h);
    pr_puts(SET_POS);
    pr_puts("\x9b" "0t");
    v_offset = v;
    h_offset = h;
    last_H = last_V = -1;
}

static int f = -1;
static int last_attr = -1;

static void pre_put_char(FONT_INFO *font, uint *code)
{
    if (f != font->font_code) {
        f = font->font_code;
        pr_normal(last_attr);
        last_attr = 0;
    }
    if (ptex_d == 1 && (*code & 0xff00)) {
        *code = change_code(*code);
    }
#ifdef FDOWN
#endif
}
static void put_jis_font(uint code)
{
    if (last_dir != 2) {
        pr_putsn(ROTATE_CHAR, 0);
        last_dir = 2;
        goto change;
    }
    if (f != j_lbp_font) {
change:
        pr_putsn(CHANGE_G1, j_lbp_font = f);    /* change G1 font */
#ifndef NOTATEGAKI
        if (tategaki) pr_puts("\x9b" "15 S");
#endif
    }
    if ((g_font & F_ATTR) != last_attr)
        pr_attr(last_attr = (g_font & F_ATTR));
    pr_int2(code | 0x8080);        /* output JIS code */
}

#define SHIFT_ASSIGN  0x200

static void put_dl_font(uint code)
{
    if (code & 0x80) {
        f += SHIFT_ASSIGN;
        code &= 0x7f;
    }
#ifndef NOTATEGAKI
    if (tategaki && last_dir != 0)  pr_puts("\x9b" " S");
#endif

    if (ptex_d == 1 && last_dir != 1) {
        pr_putsn(ROTATE_CHAR, 270);
        last_dir = 1;
    }
    else if (ptex_d == 0 && last_dir != 0) {
        pr_putsn(ROTATE_CHAR, 0);
        last_dir = 0;
    }
    if (f != e_lbp_font)        /* change G0 font */
        pr_putsn(CHANGE_G0, e_lbp_font = f);
    if (code <= 0x20 || code == 0x7f)
        pr_puts(SPECIALC);
    pr_putc(code);                /* output ASC code */
}

#ifdef FDOWN

static long get_dlhsize(int num)
{
    return((long)num * D_HEAD);
}

static void pre_download(FONT_INFO *dfont, DL_INFO *dl, int num_dfont, int flag)
{
    if (flag) dl->assign_num += SHIFT_ASSIGN;
    pr_putsn("\220%d;1x\234", dl->assign_num);    /* assign a number to FONT */
    sprintf(str_exp, "\233%ld;%d;0;0;%d;", dl->total_send, dl->numoffonts,
        3840 + num_dfont);
    pr_puts(str_exp);            /* total bytes to send */
    /* total number of characters */
    /* default mark = always 0 */
    /* font orientation 0: portrait */
    /* graphic set number */

    pr_putsn("8;20;100;0;0;200;%d;", dl->maxwidth);
    /* spacing flag 8: fixed pitch, real size */
    /* pitch (characters/inch)*100 */
    /* character size */
    /* style */
    /* weight */
    /* shotai number */
    /* cell width */

    pr_putsn("%d;1;2;0;0;0;1.p", dl->maxheight + 1);
    /* cell height */
    /* baseline */
    /* under line */
    /* flag of length of character code */
    /* level to keep 0: temporary */
    /* signature of baseline */
    /* flag of control characters */
}

static void download_rastor(FONT_INFO *dfont, DL_INFO *dl, uint code, int flag)
{
    PREAMBLE *preamble;
    int bytesize;
    BUFFER *pat;

    preamble = get_font(flag + code, dfont);
    pr_putc(code);                 /* character code */
    pr_int2(preamble->width);      /* real width */
    pr_int2(0);                    /* x-offset */
    pr_int2(preamble->height);     /* real height */
    pr_int2(0);                    /* y-offset */
    bytesize = preamble->byte_width * preamble->height;
    pat = preamble->raster;
    while (bytesize--)  pr_putc(*pat++);
}

static void post_download(FONT_INFO *dfont, DL_INFO *dl)
{
  /* dummy */
}
#endif

static uint change_code(uint code)
{
    if (code < 0x2122 || (code > 0x2166 && code < 0x2421) || code > 0x2576)
        return code;
    if (code >= 0x2141 && code <= 0x2145)
        code += 0x70c;
    else if (code >= 0x2146 && code <= 0x2149)
        code += 0xce2;
    else if (code >= 0x214a && code <= 0x215b)
        code += 0x708;
    else if (code >= 0x2162 && code <= 0x2166)
        code += 0xcce;
    else {
        switch (code) {
          case 0x2122:
          case 0x2123:
              code += 0x724;
              break;
          case 0x2124:
          case 0x2125:
              code += 0xcfd;
              break;
          case 0x2127:
          case 0x2128:
              code += 0xd06;
              break;
          case 0x2131:
          case 0x2132:
              code += 0x717;
              break;
          case 0x213c:
          case 0x213d:
          case 0x213e:
              code += 0x70e;
              break;
          case 0x215d:
              code = 0x2e2f;
              break;
          case 0x2161:
              code = 0x2846;
              break;
          case 0x2162:
              code = 0x2e30;
              break;
          case 0x2421:
              code = 0x2865;
              break;
          case 0x2423:
              code = 0x2866;
              break;
          case 0x2425:
              code = 0x2867;
              break;
          case 0x2427:
              code = 0x2868;
              break;
          case 0x2429:
              code = 0x2869;
              break;
          case 0x2443:
              code = 0x286e;
              break;
          case 0x2463:
              code = 0x286a;
              break;
          case 0x2465:
              code = 0x286b;
              break;
          case 0x2467:
              code = 0x286c;
              break;
          case 0x246e:
              code = 0x286e;
              break;
          case 0x2521:
              code = 0x286f;
              break;
          case 0x2523:
              code = 0x2870;
              break;
          case 0x2525:
              code = 0x2871;
              break;
          case 0x2527:
              code = 0x2872;
              break;
          case 0x2529:
              code = 0x2873;
              break;
          case 0x2543:
              code = 0x2878;
              break;
          case 0x2563:
              code = 0x2874;
              break;
          case 0x2565:
              code = 0x2875;
              break;
          case 0x2567:
              code = 0x2876;
              break;
          case 0x256e:
              code = 0x2877;
              break;
          case 0x2575:
              code = 0x2879;
              break;
          case 0x2576:
              code = 0x287a;
              break;
        }
    }
    return(code);
}

static void lips_page_format(int f_rotate, int f_copy)
{
    if (f_rotate) {
        pr_puts(ROTATE);
        move_homepoint(0, 0);
    }
    if (f_copy) pr_putsn(COPYS, f_copy);
}

#endif

/* end of lips3.c */
