/*
 *  TEX Device Driver  ver 2.02-2.29
 *  copyright(c) 1988, 1989 by TSG, 1990 by SHIMA
 *
 *  loadpk.c :
 *            April 15, 1989 : first edition
 *                             programed by T.IWAI
 *
 *  modified for Minix-386 1992 by obuk, youchan
 */

#include "config.h"
#include "dd.h"
#include <fcntl.h>
#include "dviread.h"
#include "err.h"

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

/* buffer.c */
void pk_buf_flush(void);

int file_read(int, BUFFER *, int);
int ask_pk_buf(long size);
int openf(char *, int);
void flush_pk(void);

int f_flush_pk;

#define READ_SIZE (1024*8)

static void buf_load(int fd, BUFFER *buf, long size)
{
    int i, n;
    
    for (i = 0; i < size; i += n) {
        n = min(READ_SIZE, size-i);
        if (file_read(fd, buf+i, n) == FAILURE) {
            if (i == 0) break;
            error(FILE_FAULT, "");
        }
    }
}


int openf(char* name, int mode)
{
    int fd;

    if ((fd = open(name, mode)) > 0) return(fd);
    flush_pk();
    return(open(name, mode));
}


void flush_pk()
{
    static  FONT_INFO *font;
    FONT_INFO   *fontc;

    if (font == NULL) font = first_font_info;
    fontc = font;
    do{
        if ((fontc->font_type == PKD_FONT || fontc->font_type == JXL)
          && fontc->ext.fh > 0) {
            close(fontc->ext.fh);
            f_flush_pk++;
            fontc->ext.fh = 0;
            font = font->next_font;
            return;
        }
        fontc = fontc->next_font;
        if (fontc == NULL) fontc = first_font_info;
    }while(font != fontc);
}


int ask_pk_buf(long size)
{
    if ((BUFFER*)pk_buf_pointer->end - 
            (BUFFER*)pk_buf_pointer->current <= size + 32) {
        if (pk_buf_pointer->size < size || size > 0xffe0L)
                return(FAILURE);
        pk_buf_flush();
    }
    return(NOTHING);
}

static BUFFER *load_pk_data(char *name)
{
    int fd, i, j, k;
    long file_pos, file_size;
    long *jxl_info;
    long tmp;
    FILE *fp;
    BUFFER *pk;

    if (*name == '*') {
        file_pos = *((long *)(name+1));
        file_size = *((long *)(name+5));
#ifdef msdos
        if ((fd = openf(name+9, O_RDONLY|O_BINARY)) == FAILURE
#else
        if ((fd = openf(name+9, O_RDONLY)) == FAILURE
#endif
          || lseek(fd, file_pos, SEEK_SET) == -1L)
            error(NO_FILE, name);
    }else{
#ifdef msdos
        if ((fd = openf(name, O_RDONLY|O_BINARY)) == FAILURE)
#else
        if ((fd = openf(name, O_RDONLY)) == FAILURE)
#endif
            error(NO_FILE, name);
        file_size = filelength(fd);
    }
    read(fd, &tmp, 4);
    if (tmp == 0x258B0100L) {
        if (*name == '*') 
            error(FILE_FAULT, "%s contains JXL4 file", name+9);
        file_size = 64L;
        fp = fdopen(fd, "rb");
        fseek(fp, -64L, SEEK_END);
    }
    if (ask_pk_buf(file_size) == FAILURE)
                error(NO_MEMORY, "font file over: %s", name);
    if ((*((long *)pk_buf_pointer->current) = tmp) == 0x258B0100L) {
        jxl_info = (long *)(pk_buf_pointer->current);
        for(i=15; i > 0; i--) *(++jxl_info) = read_long(fp);
        if (read_long(fp) != 0x18B25L)
            error(FILE_FAULT, "JXL4 file %s", name);
        fclose(fp);
    }
    else{
        buf_load(fd, pk_buf_pointer->current + 4, file_size - 4);
        close(fd);
    }
    pk = pk_buf_pointer->current;
    (BUFFER*)pk_buf_pointer->current += file_size;
    return(pk);
}


BUFFER *load_pk(char *name)
{
    return(load_pk_data(name));
}


int file_read(int fd, BUFFER* buf, int size)
{
    return(read(fd, buf, size) == size ? 0 : FAILURE);
}


/* end of file : loadpk.c */
