/*
 *  TEX Device Driver  ver 2.02, 2.05-2.29
 *  (c) copyright 1988, 1989 by TSG, 1990 by SHIMA
 *
 *  main.c : MAIN MODULE
 *      Apr. 22, 1989 : 1st edition
 *      Jun.  1, 1989 : 2nd edition
 *
 *  Modified for IBM-PC with DOS/V by SOLITON 12/10/1991
 *  This module becomes device-dependent which can be
 *  comment out by #undef IBMPC
 *
 *  modified for LIPS3 by SHIMA 30 June 1992
 *  modified for Minix-386 and Linux by obuk, youchan  1992
 *  modified for LIPS3 and ESC/Page by OkI and H.Tomiie 30 Sep 1992
 *  sligtly modified for JGfont(3D-Bezier) by Naochan! 12 Nov 1992
 *  modified for GO32(dviprt) by mkiguchi, ogochan  1992
 */

#include "config.h"
#include "dd.h"
#ifdef msdos
#include <dos.h>
#include <dir.h>
#endif
#include "err.h"
#include "vfont.h"
#include <signal.h>

#define OPTION_FLG    '-'
#define EXTENSION_FLG '.'

char* const default_dvifile_ext = "dvi";
static DVIFILE_INFO dvifile;

static void information_out(DIMENSION *);
static void dvifile_open(DVIFILE_INFO *, int, char **);
static void dvifile_close(DVIFILE_INFO *);
static char* name_link(char*, char*);
static void catchint(int);
extern void ttclose(void);

extern int test_flg;
    /* -f is exist or not? (definition in init.c) */
extern DIMENSION dviout_dimension;
    /* dimension for dviout (definition in init.c) */

extern BOOL device_open_flag;
extern char* const progname;
    /* program name */
extern char* const config_name;
    /* config file name */
extern char* const pk_env_name;
    /* name of the env-variable for pk-font path */
extern char* const title_name;
extern char* const title_comment;
extern char* const title_copyright;

extern int r_b_f_count;
extern int p_b_f_count;
extern int f_flush_pk;
extern uint num_char;
int    num_view_buf;
int    num_view = -1 ;

extern int tt_flag;

/* init.c */
extern void get_config(DIMENSION *, int, char **);
extern void initialize(DVIFILE_INFO *, DIMENSION *);
void help_out(void);

/* putdvi.c */
void put_dvi(DVIFILE_INFO *, DIMENSION *, int, char **);

/* device.c or p_out.c */
void device_end(void);

/* buffer.c */
char* marea(unsigned int);
void save_fc(void);

/* fontdef.c */
char *get_font_name(FONT_INFO *);

#ifndef NOVFONT
/* vfont.c */
extern V_JFM *get_vjfm(int);
extern V_FONT *vfont[];
#endif

void main(int argc, char** argv)
{
    signal(SIGINT, catchint);
    signal(SIGQUIT, SIG_IGN);

    fprintf(stderr,"%s%s%s\n\n", title_name, title_comment, title_copyright);
    get_config(&dviout_dimension, argc, argv);
 
    dvifile_open(&dvifile, argc, argv);
    initialize(&dvifile, &dviout_dimension);

    if (test_flg) {
        put_dvi(&dvifile, &dviout_dimension, argc, argv);
        save_fc();
        dvifile_close(&dvifile);
        exit(0);
    }
    information_out(&dviout_dimension);
    dvifile_close(&dvifile);
    exit(0);
}


int no_extension(char *name)
{
    int     len, len0;

    len0 = ((len = strlen(name)) > 4)? len-4:0;
    while (--len >= len0) {
        if (name[len] == EXTENSION_FLG) return(0);
        else if (name[len] == PATHCHAR) break;
    }
    return(1);
}


static void dvifile_open(DVIFILE_INFO* dvi, int argc, char** argv)
{
    while (--argc) {
        if (**++argv != OPTION_FLG) {
            dvi->file_name = *argv;
            if (no_extension(*argv))
                dvi->file_name = name_link(*argv, default_dvifile_ext);
#ifdef msdos
            while ((dvi->file_ptr = fopen(dvi->file_name, "rb")) == NULL) {
#else
            while ((dvi->file_ptr = fopen(dvi->file_name, "r")) == NULL) {
#endif
                error(NO_FILE, "'%s' is not found.", dvi->file_name);
            }
            return;
        }
    }
    help_out();
    error(ILLEGAL_ARGS, "No dvifile");
}

static void dvifile_close(DVIFILE_INFO* dvi)
{
    FONT_INFO* font;

    fclose(dvi->file_ptr);

    for (font = first_font_info; font; font = font->next_font) {
        if (font->font_type >= JIS_FONT) {
            if (font == font->k_top)   close((font->ext.kdir)->fh);
        }
        else if (font->font_type >= PKD_FONT  && font->ext.fh > 0)
            close(font->ext.fh);
    }
}

static char* name_link(char* base, char* ext)
{
    char *name;
    int base_len;
    
    base_len = strlen(base);
    name = marea(base_len + 5);
    strcpy(name, base);
    name[ base_len ] = EXTENSION_FLG;
    strcpy(name + base_len + 1, ext);
    return(name);
}

static void information_out(DIMENSION* dim)
{
    FONT_INFO *font;
    int   font_num = 0;

    for (font = first_font_info; font != NULL; font = font->next_font)  font_num++;
    printf("font : total %d (%d dpi)\n", font_num, dim->dpi);
    printf("mag  : %d\n", dim->mag);
    printf("page : %d\n", dim->total_page);
    printf("buffer size : %d x %d (%ld bytes)\n", dim->buf_width, dim->buf_height, bitmap_buf_pointer->size);
    printf("text   size : %d x %d\n\n", dim->text_width, dim->text_height);
}

static void catchint(int signo)
{
    if (tt_flag != 0)  ttclose();
    if (device_open_flag == TRUE)  device_end();
    dvifile_close(&dvifile);
    fprintf(stderr, "TEXDVI: Interrupt -- abort\n");
    exit(0);
}
