/*
 * TeX device driver dviout/dviprt Ver 2.36
 *
 * support routine for Minix-386 1992 by obuk
 * changed by youchan
 */

#include "config.h"
#include "dd.h"
#include "err.h"
#include <unistd.h>
#include <sys/stat.h>

SCALED_PT mult20(SCALED_PT num_1, SCALED_PT num_2)
{
/* return( (num_1 * num_2) >> 20 ); */
  return (SCALED_PT)(((double) num_1 * (double) num_2 / (1 << 20)));
}

#ifdef minix
char *strdup(char *p)
{
  char *q = malloc(strlen(p)+1);
  if (q == NULL) error(NO_MEMORY, "virtual memory exhausted");
  return strcpy(q, p);
}
#endif

#ifndef linux
char *strlwr(char *p)
{
  char *original = p;
  while (*p != '\0') {
  	*p = tolower(*p);
  	++p;
  }
  return original;
}

char *strupr(char *p)
{
  char *original = p;
  while (*p != '\0') {
  	*p = toupper(*p);
  	++p;
  }
  return original;
}
#endif

long filelength(int fd)
{
  long pos = lseek(fd, 0, SEEK_CUR);
  long end = lseek(fd, 0, SEEK_END);
  lseek(fd, pos, SEEK_SET);
  return end;
}

int tt_flag = 0;

#if defined minix
/*
 * Terminal I/O control for MINIX (V7 compatible)
 */
#include <sgtty.h>

struct sgttyb ostate;
struct sgttyb nstate;
struct tchars otc;
struct tchars ntc;

void ttopen()
{
    gtty(0, &ostate);
    nstate = ostate;
    nstate.sg_flags |= (CBREAK);
    nstate.sg_flags &= ~(RAW|ECHO|CRMOD);
    stty(0, &nstate);
    ioctl(0, TIOCGETC, (struct sgttyb*)&otc);
    ntc = otc;
    ntc.t_quitc = ntc.t_startc = ntc.t_stopc = ntc.t_eofc = ntc.t_brkc = 0xff;
    ioctl(0, TIOCSETC, (struct sgttyb*)&ntc);
    tt_flag = 1;
}

void ttclose()
{
    stty(0, &ostate);
    ioctl(0, TIOCSETC, (struct sgttyb*)&otc);
    tt_flag = 0;
}

#elif defined linux
/*
 *  Terminal I/O control for Linux (SVR3 subset)
 */
#include <termios.h>

struct termio tio, tin;

void ttopen()
{
    ioctl(0, TCGETA, &tio);
    tin = tio;
    tin.c_iflag &= ~(BRKINT | PARMRK | INPCK | IUCLC | IXON | IXOFF);
    tin.c_iflag |= (IGNBRK | IGNPAR);
    tin.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ECHONL);
    tin.c_lflag |= ISIG;
    tin.c_cc[VMIN] = 1;
    tin.c_cc[VTIME] = 0;
    ioctl(0, TCSETA, &tin);
    tt_flag = 1;
}

void ttclose()
{
    ioctl(0, TCSETA, &tio);
    tt_flag = 0;
}
#elif defined msdos
/*
 * Terminal I/O control is dummy at GO32
 */

void ttopen()
{
    tt_flag = 1;
}

void ttclose()
{
    tt_flag = 0;
}
#endif

#ifndef msdos
int getch()
{
  int c;

  ttopen();
  c = fgetc(stdin) & 0377;
  ttclose();
  return c;
}
#endif
