/*
 *  TEX Device Driver  ver 2.02-2.33
 *  copyright(c) 1988, 1989 by TSG, 1990 by SHIMA
 *
 *  option.c : standard option-parse module
 *
 *  modified for non PC-9801 machines by sempa 1992
 *  modified for `Hidden Options' by Naochan! 6 August 1992
 *  modified for `COMMENT LINE' by sempa 2 Sept 1992
 *
 *  modified for Minix-386 1992 by obuk, youchan
 */


#include "config.h"
#include "dd.h"
#include "option.h"
#include "err.h"

#define OPTION_FLG '-'
#define OFF_USG 2
#define MAX_LINE 0x1000
#define COMMENT_CHAR '#'

#define is_hex(c)     ((c>='0'&&c<='9')||(c>='a'&&c<='f')||(c>='A'&&c<='F'))
#define is_oct(c)     (c>='0'&&c<='7')
#define is_dig(c)     (c>='0'&&c<='9')
#define hex_to_dig(c) ((c>='0'&&c<='9')?(c-'0'):(toupper(c)-('A'-10)))
#define to_dig(c)     (c-'0')

/* buffer.c */
char *dup_string(char *);

/* init.c */
extern const char* const pk_env_name;
extern char* pk_search_path;
extern char* pkd_search_path;
extern char* knj_search_path;

int     sum_option;
static  int f_file;     /* nesting the function reading options from a file */

static char* strchk(char* line, char* token)
    /* If the head token of 'line' matchs 'token', return 'line'.
     * Otherwise, returning NULL
     */
{
    while (*token)
        if (*token++ != *line++)
            return(NULL);
    
    return(line);
}

static long stol(char* str)
    /* set string-value to long-integer. Oct- or Hex-values in thought.
     */
{
    long num = 0;
    int sign = 1;
    
    if (*str == '-') {
        sign = -1;
        str++;
    }
    if (*str == '0') {
        if (*++str == 'x' || *str == 'X') {
            str++;
            while (is_hex(*str)) {
                num *= 16;
                num += hex_to_dig(*str);
                str++;
            }
        }
        else {
            while (is_oct(*str)) {
                num *= 8;
                num += to_dig(*str);
                str++;
            }
        }
    }
    else {
        while (is_dig(*str)) {
            num *= 10;
            num += to_dig(*str);
            str++;
        }
    }
    return(num * sign);
}

static void var_set(ARG_TABLE* option, char* arg)
    /* set args on variables directed in argment-table */
{
    switch(option->type) {
    case BOOLEAN :
        if (*arg == '+') *(BOOL *)option->var = TRUE;
        else {
            if (*arg == '-') *(BOOL *)option->var = FALSE;
            else *(BOOL *)option->var = (*(BOOL *)option->var == TRUE)? FALSE:TRUE;
        }
        break;
    case INTEGER :
        *(int *)option->var = (int)stol(arg);
        break;
    case LONGINT :
        *(long *)option->var = stol(arg);
        break;
    case PROCEDURE :
        (*(void_func_ptr)option->var)(arg);
        break;
    case STRING :
        *(char **)option->var = dup_string(arg);
        break;
    default :
        error(ILLEGAL_ARGS, option->message);
        break;
    }
}

static BOOL analyse(ARG_TABLE* tbl, char* line)
    /* option-analyse */
{
    char* arg;

    if (*line++ != OPTION_FLG) {
        if ((arg = strchr(--line, '=')) != NULL) {
            *arg++ = '\0';
            if (!strcmp(line, pk_env_name)) {
                pk_search_path = (char*)strdup(arg);
                return(TRUE);
            }
            else if (!strcmp(line, "TEXPKD")) {
                pkd_search_path = (char*)strdup(arg);
                return(TRUE);
            }
            else if (!strcmp(line, "TEXKNJ")) {
                knj_search_path = (char*)strdup(arg);
                return(TRUE);
            }
        }
        return(FALSE);
    }

    for(arg = line; *arg ;) sum_option += *arg++;
    for (; tbl->option; tbl++)
        if ((arg = strchk(line, tbl->option)) != NULL) {
            var_set(tbl, arg);
            return(TRUE);
        }
    
    error(ILLEGAL_ARGS, "no match %s", line);
}

void set_option(ARG_TABLE *tbl, int argc, char **argv)
    /* analyse option-arguments of command-line */
{
    while(--argc && analyse(tbl, *++argv))  ;
}

static BOOL is_white(int c)
{
    static BOOL f_in_comment = FALSE;

    if (c==EOF) return FALSE;
    if (!f_in_comment) {
        if (c==COMMENT_CHAR) {
            f_in_comment = TRUE;
            return TRUE;
        }
        else return (c==' '||c=='\t'||c=='\n');
    }
    else {
        if (c=='\n') f_in_comment = FALSE;
            return TRUE;
    }
}

void set_config(ARG_TABLE *tbl, char *config_file)
    /* analyse option-arguments of config-file */
{
    char line[ MAX_LINE ], *tmp;
    int code;
    FILE *fptr;

    if (f_file >= 2) {
        error(ILLEGAL_ARGS, "Nesting of configuration file %s", config_file);
    }
#ifdef msdos
    if ((fptr = fopen( config_file, "rt")) == NULL) return;
#else
    if ((fptr = fopen( config_file, "r")) == NULL) return;
#endif
    f_file++;
    do {
        while(code = getc(fptr), is_white(code))
            ;
        for (tmp = line; code != EOF &&
            ! is_white(code) && tmp - line < MAX_LINE-1;) {
            *tmp++ = code;
            code = getc(fptr);
        }
        
        *tmp = '\0';
        
        if (code != EOF)
            ungetc(code, fptr);
    } while(analyse(tbl, line) && code != EOF);
    
    fclose(fptr);
    f_file--;
}

void option_usage(ARG_TABLE *tbl)
    /* print out usage */
{
    int     i;

    for (i=OFF_USG; tbl->option; i++, tbl++)
            printf("  -%-5s: %s\n", tbl->option, tbl->message);
}


/* end of file : option.c */

