/*
 *  TEX Device Driver  ver 2.00-2.31
 *  copyright(c) 1988,89 by TSG, 1990 by SHIMA
 *
 *  PRINTER DRIVER ver 1.50
 *  device.c : print out module
 *             3rd edition
 *
 *  1990, Modified by SHIMA
 *  1992, Slightly extended for Kyocera L-880 & HP LaserJet+ by Oh-Yeah?
 *
 *  1992, modified by sempa
 *
 *  modified for non PC-9801 machines by sempa 1992
 *  modified for LIPS3 by SHIMA 30 June 1992
 *  modified for ESC/Page by H.Tomiie, 31 July 1992
 *  modified for Minix-386 and Linux by obuk, youchan  1992
 *  modified for LIPS3 and ESC/Page by H.Tomiie and OkI, 30 Sep 1992
 */

#define PR_BUF_SIZE     256
#define STR_EXP_SIZE    32

#define HWMIN       1
#define HWMAX       (1440/8)

#include "config.h"
#include "dd.h"
#ifdef msdos
#include <pc.h>
#endif
#include "prtout.h"
#include "err.h"
#include "prtctl.h"

#define ESC         0x1b
#define CR          0x0d
#define LF          0x0a

int     pr_buf_len;
BOOL    f_pause;
int     ff_flg = -1;
BOOL    f_top_margin;
BOOL    f_r_format;
char    pr_buf[PR_BUF_SIZE];
char    str_exp[STR_EXP_SIZE];
char    *prt_initialize;
#ifdef msdos
#if defined(PC98) || defined(FMR)
char    *f_out;
FILE    *fp_out;
#endif
#endif
static  int vert_byte_unit = V_BDIV;
static  int vert_bit_unit  = V_BDIV<<3;

static  uint    hwmin = HWMIN;
static  uint    hwmax = HWMAX;
static  uint    mult_const;
static  int     num_cr;
static  long    trans_byte;

extern  DIMENSION   dviout_dimension;
extern  PIXEL       vert_divide;

#ifdef LBP
static int last_H;
static int last_V;
int f_rotate;
int f_vh_ratio;
int f_copy = 0;
char *youshi;
extern int f_use_lfont;
LBP_FUNC *lbpf;

#ifdef LIPS3
extern LBP_FUNC lips3_func;
#endif
#ifdef ESCPAGE
extern LBP_FUNC escpage_func;
#endif

int k_base_l = 10000;  /* This is relative according to the size of Kanji in LBP.
                          1000 is corresponds the size of Kanji. A larger number
                          means that Kanji fonts move to lower position */
int k_base_d = 1;      /* Arrange of dot height of scalable Kanji fonts. A larger
                          number means that Kanji fonts move to lower position. */
int k_scale_l = 10000; /* This is factor for Kanji width adjustment of LBP scalable
                          font. It was after multiplied by 'k_scale' factor. */
int lbp_orig_dpi = 0;
int k_width_adj = 0;
int last_dir = -1;
#endif
  
/* init.c prtinit.c */
extern char* cfg_path;
extern BOOL f_lbp_prt;

/*  buffer.c    */
char* marea(unsigned int);

/* main.c */
int no_extension(char *);

void pr_putnc(char *, int);
void pr_puts(char *);
void pr_putsn(char *, int);
void pr_putc(char);
void pr_crlf(void);

void p_sub(ulong, BUFFER*, int);             /* p_sub.asm    */
void p_subnec(ulong, BUFFER*, int);          /* p_subnec.asm */

void conv_bit(ulong, BUFFER*);               /* convert bit pattern */

char get_column_byte(BUFFER *ptr, int bitno, int hmax);

void pr_new_page(void);
void pr_flush(void);

void device_pause(void)
{
}

void device_cont(void)
{
}

static char *line_feed = "\x02\x00\x00\x00\x0d\x0a";
static char *form_feed = "\x02\x00\x00\x00\x0d\x0c";
static char *after_bit_image = "\x00\x00\x00\x00";
static char *bit_row_header = "\x00\x00\x00\x00";

static char prt_ctl[MAX_PRINTER] ={
    /* ctrl of printer 
     * lower 4 bits represents the vertical length by  bytes
     */

        VERT_DIVIDE/8,                              /* ESC/P printer */

       (VERT_DIVIDE/8)|TOP_IS_LOW,                  /* PC-PR printer */

       (VERT_DIVIDE/8)|TOP_IS_LOW                   /*  NM   printer */
#ifdef LIPS3
     , (8/8)|LEFT_IS_HIGH|NON_MOVING                /* lips3 */
#endif
#ifdef ESCPAGE
     , (8/8)|LEFT_IS_HIGH                           /* ESC/Page */
#endif
};

static char *dot_cr[MAX_PRINTER] = {
        /* change to a line feed for bit image */

    "\x07\x00\x00\x00"          /* ESC/P printer */
    "\x1b""3\x18"               /* 3 bytes: 24/180 inch line feed */
    "\x1b""?Z""\x27"            /* 4 bytes: ESC Z ... -> ESC * m \x27 ... */
,

    "\x06\x00\x00\x00"          /* PC-PR printer */
    "\x1b""T18"                 /* 4 bytes: 18/120 = 24/160 inch line feed */
    "\x1b"">"                   /* 2 bytes: print in one direction */
,

    "\x0a\x00\x00\x00"          /* NM printer */
    "\x1a""C"                   /* 2 bytes: 180 dpi for bits image */
    "\x1a""F"                   /* 2 bytes:  1/120 inch pitch */
    "\x1b""T16"                 /* 4 bytes: 24/180 inch line feed */
    "\x1b"">"                   /* 2 bytes: print one direction */
#ifdef LIPS3
,
    "\x00\x00\x00\x00"          /* Use call lips3_init() */
#endif
#ifdef ESCPAGE
,
    "\x00\x00\x00\x00"          /* Use call escpage_init() */
#endif
};

static char *nrm_cr[MAX_PRINTER] = {
        /* change to a normal line feed */

    "\x04\x00\x00\x00"          /* ESC/P printer */
    "\x0d\x0c"
    "\x1b""2"                   /* 2 bytes: 1/6 inch line feed */
,

    "\x06\x00\x00\x00"          /* PC-PR printer */
    "\x0d\x0c"
    "\x1b""A"                   /* 2 bytes: 1/6 inch line feed */
    "\x1b""]"                   /* 2 bytes: print in both direction */
,

    "\x06\x00\x00\x00"          /* NM printer */
    "\x0d\x0c"
    "\x1b""A"                   /* 2 bytes: 1/6 inch line feed */
    "\x1b""]"                   /* 2 bytes: print in both direction */
#ifdef LIPS3
,
    "\x08\x00\x00\x00"          /* LIPS3 */
    "\x0d\x0c"                  /* 2 bytes: CR & FF */
    "\x1b" "<"                  /* 2 bytes: Software reset */
    "\x90" "0J" "\x9c"          /* 4 bytes: End Job */
#endif
#ifdef ESCPAGE
,
    "\x06\x00\x00\x00"          /* ESC/Page */
    "\x0d\x0c"
    "\x1d" "rhE"
#endif
};

static char *dotg[MAX_PRINTER] = {
        /* printing bit image */

    "\x04\x03\x02\x01"          /* ESC/P printer */
    "\x1b""Z\x00\x00"
,

    "\x06\x03\x02\x04"          /* PC-PR printer */
    "\x1b""J0000"
,

    "\x06\x03\x02\x04"          /* NM printer */
    "\x1b""J0000"
#ifdef LIPS3
,
    "\x05\x83\x00\x86"          /* LIPS3 */
    "\x9b" "%u;" "\x00"
    "\x05\x00\x00\x36"
    "%u.r"
#endif
#ifdef ESCPAGE
,
    "\x05\x83\x00\x06"          /* ESC/Page */
    "\x1d" "%u;" "\x00"
    "\x09\x03\x00\x06"
    "%u;8bi{I" "\x00"
#endif
};


static char *dots[MAX_PRINTER] = {
        /* dots spacing */

    "\x04\x03\x02\x01"          /* ESC/P printer */
    "\x1b\x5c\x00\x00"
,

    "\x09\x03\x02\x04"          /* PC-PR printer */
    "\x1b""U0000\x00\x00\x00"
,

    "\x09\x03\x02\x04"          /* NM printer */
    "\x1b""U0000\x00\x00\x00"
#ifdef LIPS3
,
    "\x05\x03\x00\x06"          /* LIPS3 */
    "\x9b" "%ua"
#endif
#ifdef ESCPAGE
,
    "\x05\x03\x00\x06"          /* ESC/Page */
    "\x1d" "%uH" "\x00"
#endif
};

static ushort _get_ushort(FILE *fptr)
{

    ushort    dl, dh;

    dl = (uchar)getc(fptr);
    dh = (uchar)getc(fptr);
    return(dl + (dh << 8));
}

static  uchar* get_len(uchar* buf)
{
    int     flag;

    do{
        flag = buf[1];
        buf[1] |= 3 - ((buf[3] & DIVIDE_ALL) >> 4);
        buf += buf[0]+4;
    }while(flag & 0x80);
    return(buf);
}

int get_prt_ctl(uchar* filename)
/* Read the configuration file for Printer Device
 */
{
    uint    pos, size;
    int     i, ch;
    char    file[MAXPATH+4], fname[MAXPATH], *srcpath;
    uchar   *prtctl_buf;
    FILE    *fptr;

    if (*filename == 0)
        strcpy(fname, prt_ctr_file);
    else {
        strncpy(fname, filename, MAXPATH);
        if (no_extension(fname))
            strcat(fname, ".cfg");
    }
    if (strpbrk(fname, PATH_PREFIX) != NULL) {   /* check path prefixed */
        strncpy(file, fname, MAXPATH);
    } else {
        strncpy(file, cfg_path, MAXPATH);
        strcat(file, fname);
    }
#ifndef msdos
    if ((fptr = fopen(file, "r")) == NULL) {
#else
    if ((fptr = fopen(file, "rb")) == NULL) {
#endif
        fprintf(stderr, "Cannot open printer configulation file %s.\n"
                        "See -po option !\n", file);
        return(FAILURE);
    }
    for(i = 16; i > 0; --i) {
        fprintf(stderr, "%c", ch = (uchar)getc(fptr));
        if (ch == 0) while(--i > 0) getc(fptr);
    }
    fprintf(stderr, "\n");

        /* skip 16 bytes */

    if ((ch = getc(fptr)) == 'Q')
        f_r_format = FALSE;
    else if (ch == 'R')
        f_r_format = TRUE;
    else
        goto ptrcfg_err;
    prt_ctl[PR_OTHER] = (uchar)getc(fptr);
    if ((vert_byte_unit = (prt_ctl[PR_OTHER] & 0x0f)) == 0)
        goto ptrcfg_err;
    vert_bit_unit = vert_byte_unit << 3;

    if (getc(fptr) != 0xff)
        goto ptrcfg_err;
    ch = getc(fptr);
    size = (int)filelength(fileno(fptr)) - ch;

    mult_const = (uchar)getc(fptr);
    getc(fptr);
    getc(fptr);
    hwmin = (uint)_get_ushort(fptr);
    hwmax = (uint)_get_ushort(fptr);
    if (f_r_format) {
        ch = (int)_get_ushort(fptr);
        if (dviout_dimension.dpi==0) dviout_dimension.dpi = ch;
    }

    if (hwmin == 0 || hwmin == 0) {
ptrcfg_err:     fclose(fptr);
                fprintf(stderr, "Error in %s. Maybe old format.\n", fname);
                return(FAILURE);
    }
    prtctl_buf = (uchar *)marea(size);
    for(i = 0; i < size; i++) prtctl_buf[i] = (uchar)getc(fptr);

    dot_cr[PR_OTHER] = (uchar*)prtctl_buf;
    nrm_cr[PR_OTHER] = get_len((uchar*)prtctl_buf);
    dotg[PR_OTHER] = get_len(nrm_cr[PR_OTHER]);
    dots[PR_OTHER] = get_len(dotg[PR_OTHER]);
    vert_divide = (PIXEL)vert_byte_unit*8;
    line_feed = get_len(dots[PR_OTHER]);
    form_feed = get_len(line_feed);
    after_bit_image = get_len(form_feed);
    bit_row_header = get_len(after_bit_image);
    get_len(bit_row_header);
    fclose(fptr);
    return(NOTHING);
}

void pr_putc(char x)
{
    pr_buf[pr_buf_len] = x;
    if (++pr_buf_len >= PR_BUF_SIZE) pr_flush();
}

#if defined(PC98) || defined(FMR)      /* Output print data to printer port directly */
void pr_flush(void)
{
    char *p;
    int  i;
    static unsigned int num, h_num;

    if (pr_buf_len > 0) {
        trans_byte += pr_buf_len;
        if (f_out == NULL) {
#ifdef FMR
            for (i = 0, p = pr_buf; i < pr_buf_len; i++, p++) {
                while(((inportb(0x0800) & 0x80)) == 0x80);
                while(((inportb(0x0800) & 0x01)) == 0x00);
                outportb(0x0800, *p);
	    }
#endif
#ifdef PC98
            {
                unsigned char hah;
                unsigned short xcx;
                do{
                    asm("pushw %es");
                    asm("pushl %ebx");
                    asm("pushl %ecx");
                    asm("movw %0, %%ax"::"g"(rom_pr->selecter):"ax");
                    asm("movw %ax, %es");
                    asm("movw $0, %bx");
                    asm("movw %0, %%cx"::"g"(pr_buf_len):"cx");
                    asm("movb $0x30, %ah");
                    asm("int $0x1a");
                    asm("movb %%ah, %0":"=g"(hah)::"ax");
                    asm("movw %%cx, %0":"=g"(xcx)::"cx");
                    asm("popl %ecx");
                    asm("popl %ebx");
                    asm("popw %es");
                } while(((hah & 2) == 0) && (xcx != 0));
            }
#endif
        } else {
            if (*f_out == 0) {
                for (i = 0; i < pr_buf_len;) {
                    if ((num & 0xf) == 0)  printf("%04X: ", h_num++);
                    printf((++num & 0xf) == 0 ? "%02X\n":"%02X ",
                        (uchar)(pr_buf[i++]));
                }
            } else  fwrite(pr_buf, pr_buf_len, 1, fp_out);
        }
        pr_buf_len = 0;
    }
}
#else     /* Output print data to 'stdout' */
void pr_flush(void)
{
    if (pr_buf_len > 0) {
        trans_byte += pr_buf_len;
        fwrite(pr_buf, pr_buf_len, 1, stdout);
        fflush(stdout);
        pr_buf_len = 0;
    }
}
#endif

void pr_putnc(char *str, int num)
{
    while(num--) {
        pr_buf[pr_buf_len] = *str++;
        if (++pr_buf_len >= PR_BUF_SIZE) pr_flush();
    }
}

void pr_puts(char *str)
{
    while(*str != 0) {
        pr_buf[pr_buf_len] = *str++;
        if (++pr_buf_len >= PR_BUF_SIZE) pr_flush();
    }
}

#ifdef LBP
void pr_putsn(char *str, int num)
{
    sprintf(str_exp, str, num);
    pr_puts(str_exp);
}
#endif
  
static void pr_numb(char* str, uint num0)
{
    int     f_mode, mode, len, pos, num, flag;

    do{
        num = num0;
        len = (uchar)(*str++);
        flag = (uchar)(*str++);
        pos = (uchar)(*str++);
        f_mode = (uchar)(*str++);

        if (f_mode) {
            num <<= (flag & 3);
            if (f_mode & TOTAL_BYTE) num *= vert_byte_unit;
            if (f_mode & MULT_CONST) num *= mult_const;

            switch(mode = f_mode & 0x7) {

                case(BINARY_LTOH):
                        str[pos]    = num & 0xff;
                        str[pos+1]  = num >> 8;
                        break;

                case(BINARY_HTOL):
                        str[pos]    = num >> 8;
                        str[pos+1]  = num & 0xff;
                        break;

                case(DECIMAL_3):
                case(DECIMAL_4):
                case(DECIMAL_5):
                        sprintf(str_exp, "%05u", num);
                        mode = DECIMAL_5 - mode;
                        while(mode < 5) str[pos++] = str_exp[mode++];
                        if (f_mode & ISO_NUMBER)   str_exp[pos-1] += 0x10;
                        break;

                case(DECIMAL_V):
                        sprintf(str_exp, str, num);
                        mode = strlen(str_exp);
                        if (f_mode & ISO_NUMBER)   str_exp[mode-1] += 0x10;
                        pr_puts(str_exp);
                        goto cont;

                default:
                        break;
            }
        }
        pr_putnc(str, len);
cont:   str += len;
    }while(flag & 0x80);
}

void get_sheet(void)
{
    int i;

    pr_numb(dotg[dviout_dimension.prt_type], 1);
    for(i = vert_bit_unit; i > 0; i--) {
        if (*bit_row_header)
            pr_numb(bit_row_header, 1);
        pr_putc(0);
    }
    pr_numb(after_bit_image, 1);
    pr_putc(0x0d);
}

#define hex_to_dig(c)  ((c>='0' && c<='9')?(c-'0'):(toupper(c)-('A'-10)))

void device_init(DIMENSION* dim)
{
    char file[MAXPATH+4], *srcpath;
    int  ch, i;
    FILE *fptr;

#ifdef msdos
#if defined(PC98) || defined(FMR)
    if (f_out != NULL && *f_out != 0 && (fp_out = fopen(f_out, "wb")) == NULL) {
        error(FILE_FAULT, "Can't open %s", f_out);
    }
#endif
#endif
    pr_numb(dot_cr[dviout_dimension.prt_type], 0);
#ifdef LIPS3
    if (dviout_dimension.prt_type == LBP_LIPS3) {
        line_feed = "\x01\x00\x00\x00\x85";
        lbpf = &lips3_func;
        vert_byte_unit = 1;
        vert_bit_unit = vert_divide = 8;
        hwmax = 500;
    }
#endif
#ifdef ESCPAGE
    if (dviout_dimension.prt_type == LBP_ESCPAGE) {
        lbpf = &escpage_func;
        vert_byte_unit = 1;
        vert_bit_unit = vert_divide = 8;
        hwmax = 500;
    }
#endif
#ifdef LBP
    if (f_lbp_prt)  (*(lbpf->init))();
#endif
    if (f_top_margin) get_sheet();

    if (prt_initialize == (char*)NULL)  goto skp;
    if (strpbrk(prt_initialize, PATH_PREFIX) != NULL) { /* check path prefixed */
        strncpy(file, prt_initialize, MAXPATH);
    }else {
        strncpy(file, cfg_path, MAXPATH);
        strcat(file, prt_initialize);
    }
#ifndef msdos
    if ((fptr = fopen(file, "r")) == NULL)
#else
    if ((fptr = fopen(file, "rb")) == NULL)
#endif
        error(FILE_FAULT, "Cannot open %s", file);
    while((ch = getc(fptr)) != EOF) {
        if (ch != '\\') {
            if (ch > ' ')  pr_putc(ch);
        }else {
            if ((i = getc(fptr)) == '\\') {
                while((ch = getc(fptr)) != 0x0a
                       && ch != 0x0d && ch != EOF);
                continue;
            }
            i = hex_to_dig(i);
            ch = getc(fptr);
            pr_putc(i*16 + hex_to_dig(ch));
        }
    }
    fclose(fptr);
skp: 
#ifdef LBP
    if (f_lbp_prt) (*(lbpf->init2))(f_rotate, f_copy);
#endif
}


void pr_crlf(void)
{
    while(num_cr) {
        num_cr--;
        pr_numb(line_feed, 0);
    }
}

  
void pr_new_page(void)
{
    static int Page=0;

    if (ff_flg != -1) {
#ifdef LBP
        if (f_lbp_prt) {
            Page++;
            if (Page & 1) {
                (*(lbpf->move_homepoint))(0, ff_flg);
                num_cr = 0;
	    }
            else {
                pr_numb(form_feed, 0);
                (*(lbpf->move_homepoint))(0, 0);
                num_cr = 0;
	    }
	}
        else
#endif
            pr_crlf();
    }
    else {
        pr_numb(form_feed, 0);
        num_cr = 0;
    }
    pr_flush();
#ifdef LBP
    if (f_lbp_prt)  (*(lbpf->move_cap))(-1, -1);
#endif
    if (f_pause) {
        fprintf(stderr, "\nPush any key to continue (or ESC -> abort).\n");
        if (getch() == ESC) {
            error(PROGRAM_STOP, "User abort.");
        }
        fprintf(stderr, "\x1b[1A\x1b[0K");
    }
    if (f_top_margin) get_sheet();
}

void device_end(void)
{
    pr_numb(nrm_cr[dviout_dimension.prt_type], 0);
    pr_flush();
#ifdef LBP
    if (f_lbp_prt)  (*(lbpf->end))();
#endif
    fprintf(stderr, "\nSent %lu bytes.\n", trans_byte);
}

static void p_out(int h_max, int v_max, BUFFER * top)
{

    BUFFER *ptr;
    BUFFER *ptr_z;
    BUFFER *ptr_f;
    uint    x, y, hmin, hmax, htmax, vmax, hsend;
    uint    skip, exist;
    int     i, j, f_nul, type, f_non_move;

    vmax = v_max;
    if ((type = (prt_ctl[dviout_dimension.prt_type] & TYPE_BIT))
       == LEFT_IS_LOW) {
        conv_bit(((ulong)(vmax))*h_max*8, top);
    }
    f_non_move = prt_ctl[dviout_dimension.prt_type] & NON_MOVING;
    for(y = 0 ; y < vmax; y += vert_byte_unit) {
        ptr_z = top + (ulong)h_max*(y*8);

        for(hmin = skip = exist = 0; hmin < h_max;) {
            if (h_max > hmin + hwmin)  hmax = hmin + hwmin;
            else                        hmax = h_max;

            if (*dots[dviout_dimension.prt_type] == 0) goto f_exist;
                                        /* support dots spaces */
            for(f_nul = 0, x = hmin; x < hmax;  ++x) {
                ptr= ptr_z + x;
                for(i=vert_bit_unit; i > 0; --i) {
                    if (*ptr == 0) ptr += h_max;
                    else {
                        f_nul = 1;
                        goto f_space;
                    }
                }
            }

f_space:    if (f_nul == 0) {               /* spaces */
                skip+=hwmin;                    /* store spaces */
                if (exist != 0) goto prt_out;
                goto next_block;
            }
                                            /* not spaces */
            pr_crlf();
            while(skip != 0) {
                hsend = (skip > hwmax)? hwmax : skip;
                pr_numb(dots[dviout_dimension.prt_type], hsend);
                skip -= hsend;
            }

f_exist:    exist += (hmax - hmin);
            if (hmax != h_max) goto next_block;    /* not EOL */
            else hmin = hmax;

prt_out:    pr_crlf();
            hmin -= exist;
            if (f_non_move) skip += exist;
            while(exist != 0) {
                hsend = (uint)((exist > hwmax)? hwmax : exist);
                htmax = hmin + hsend;
                pr_numb(dotg[dviout_dimension.prt_type], hsend);
                if (type & FLAG_LBP) {
                    ptr = ptr_z + hmin;
                    for( i = vert_bit_unit; i > 0; ptr += h_max, i--) {
                        if (*bit_row_header)
                            pr_numb(bit_row_header, hsend);
                        for(ptr_f = (BUFFER*)ptr, j = hsend; j > 0; j--) {
                            pr_buf[pr_buf_len] = *ptr_f++;
                            if (++pr_buf_len >= PR_BUF_SIZE) pr_flush();
                        }
                    }
                }
                else {
                    for(x = hmin; x < htmax; ++x) {
                        ptr = ptr_z + x;
                        if ((type & TOP_IS_LOW) == 0)
                                p_sub(h_max, ptr , vert_byte_unit);
                        else    p_subnec(h_max, ptr, vert_byte_unit);
                    }
                }

                if (*after_bit_image)
                    pr_numb(after_bit_image, hsend);
                exist -= hsend;
                hmin = htmax;
            }
next_block: hmin = hmax;
        }
        num_cr++;
    }
}

static void pl_out(int h_max, int v_max, BUFFER * top)
{

    BUFFER  *ptr;
    BUFFER  *ptr_z;
    uint    x, y, ymax, vmin, vtmax, vmax, hsend;
    uint    skip, exist;
    int     i, j, f_nul, type, f_non_move;

    if ((type = (prt_ctl[dviout_dimension.prt_type] & TYPE_BIT))
       == TOP_IS_LOW)
        conv_bit(((ulong)(h_max))*v_max*8, top);
    f_non_move = prt_ctl[dviout_dimension.prt_type] & NON_MOVING;

    for(x = 0 ; x < h_max; x += vert_byte_unit) {
        ptr_z = top + (ulong)(v_max * 8 - 1) * h_max + (ulong)x;

        for(vmin = skip = exist = 0; vmin < v_max;) {
            if (v_max > vmin + hwmin)  vmax = vmin + hwmin;
            else                        vmax = v_max;

            if (*dots[dviout_dimension.prt_type] == 0) goto f_exist;
            /* support dots spaces */
            for (f_nul = 0, y = vmin << 3, ymax = vmax << 3; y < ymax; y++) {
                ptr = ptr_z - (ulong)y * h_max;
                for(i = vert_byte_unit; i > 0; --i) {
                    if (*ptr == 0) ptr++;
                    else {
                        f_nul = 1;
                        goto f_space;
                    }
                }
            }

f_space:    if (f_nul == 0) {                   /* spaces */
                skip += hwmin;                  /* store spaces */
                if (exist != 0) goto prt_out;
                goto next_block;
            }
            /* not spaces */
            pr_crlf();
            while(skip != 0) {
                hsend = (skip > hwmax)? hwmax : skip;
                pr_numb(dots[dviout_dimension.prt_type], hsend);
                skip -= hsend;
            }

f_exist:    exist += (vmax - vmin);
            if (vmax != v_max) goto next_block;    /* not EOL */
            else vmin = vmax;

prt_out:    pr_crlf();
            vmin -= exist;
            if (f_non_move) skip += exist;
            while(exist != 0) {
                hsend = (uint)((exist > hwmax)? hwmax : exist);
                vtmax = vmin + hsend;
                pr_numb(dotg[dviout_dimension.prt_type], hsend);
                ptr = ptr_z - (ulong)vmin * 8 * h_max;
                if (type & FLAG_LBP) {
                    for( i = vert_byte_unit; i > 0; ptr++, i--) {
                        for (j = 7; j >= 0; j--) {
                            if (*bit_row_header)
                                pr_numb(bit_row_header, hsend);
                            for (y = 0; y < hsend; y++)
                                pr_putc(get_column_byte((BUFFER *)(ptr
                                    - (ulong)y * 8 * h_max), j, -h_max));
                        }
                    }
                }
                else {
                /* debugged by H.Sasaki */
                    for (y = 0, ymax = hsend << 3; y < ymax; y++, ptr -= h_max)
                        for (i = 0; i < vert_byte_unit; i++)
                            pr_putc((i<(h_max-x)) ? *(ptr + (ulong)i) : 0);
                }
                if (*after_bit_image)
                    pr_numb(after_bit_image, hsend);
                exist -= hsend;
                vmin = vtmax;
            }
next_block: vmin = vmax;
        }
        num_cr++;
    }
}

void device_clear(OUTPUT_INFO *out)
{
#ifdef LBP
    if (f_lbp_prt) {
        (*(lbpf->move_cap))((out->split > 1) ?
            (out->split - 1)*dviout_dimension.buf_height : 0, 0);
        (*(lbpf->move_cap))(-1, -1);
    }
#endif
}

NEXT_ACTION device_out(OUTPUT_INFO *out)
{
#ifdef LBP
    last_dir = 3;
#endif
#ifdef msdos
#if defined(PC98) || defined(FMR)
    if (f_out != NULL && *f_out == 0) printf("\n");
#endif
#endif
    if (out->print_direction == VERTICAL)
        pl_out(out->byte_width, out->byte_height, out->bitmap_ptr);
    else
        p_out(out->byte_width, out->byte_height, out->bitmap_ptr);
    pr_flush();
#ifdef LBP
    if (f_lbp_prt)  (*(lbpf->move_cap))(-1, -1);
#endif
    return(NextPart);
}

/* end of file : p_out.c */

/* additional function by Oh-Yeah? */

char get_column_byte(BUFFER *ptr, int bitno, int hmax)
    /* get vertical 8 bits starting at bitno of ptr (increment=hmax bytes) */
{
    int i = 8;
    char byte = 0;
    unsigned char bit = 0x80, mask = 0x01 << bitno;

    while (i--) {
        if (*ptr & mask)
            byte |= bit;
        bit >>= 1;
        ptr += hmax;
    }
    return(byte);
}

#ifdef LBP
/*
 *  Common routine for scalable/downloaded font of LIPS3 / ESCPage
 *                     30 Jun  1992, written by SHIMA
 *                      1 Oct  1992, modified by OkI
 */

#ifdef FDOWN
/* LBP dependent factor. Change them in init() if necessary.*/
long max_dsize = 0x80000L;    /* download capacity (default 0.5M byte) */
int maxdwidth = 128;          /* download font width limit */
int maxdheight = 128;         /* download font height limit */
int maxdfont = 32;            /* num of kinds of download limit*/

/* program control valiables */
int d_mincount = 2;           /*??? minimal count of usage to download */
int f_download;               /*??? flag to download font */
long cur_dsize;               /* current downloaded size */
static int num_dfont;         /* current number of download fonts */
char *font_def;
#endif /* FDOWN */

int weight_org;               /* kanji font weight default */

extern char ptex_d;           /* ptex direction: 0 = yoko, 1 = tate */
extern char ptex_mode;
extern BOOL draw_boxpic;
extern int k_scale;

PIXEL sptopixel(SCALED_PT);
long jtfm_width(FONT_INFO *, int);
int chk_goth(FONT_INFO *);
SCALED_PT mult20(SCALED_PT, SCALED_PT);
PREAMBLE *get_font(uint, FONT_INFO *);
char *get_fontname(FONT_INFO *);
void p_out_char(PIXEL, PIXEL, FONT_INFO *, uint, PREAMBLE *);

KFONT *get_lbp_font(FONT_INFO *fontc)
{
    int font_no, len, vh_ratio, fg, weight;
    long tmp;
    char *name, *pt;
    KFONT *kf;

    fg = F_LBP;
    vh_ratio = f_vh_ratio;
    weight = weight_org;
    len = strlen(name = get_fontname(fontc));
    tmp = (int)((long)k_scale * (long)k_scale_l / 1000L);
    tmp = sptopixel(mult20(jtfm_width(fontc, 0), fontc->size_para / 40 * tmp));
    fontc->k_width = tmp / 25;    /* width by dots */
    /*
        f<name>=<shotai>{G}{H}{F}{S}{O};<vh_ratio>;<weight>;<dots>;<dots>;...
        can be omitted after any ;
     */
    if ((pt = font_def) == NULL)  goto org;
    for (font_no = 2;;) {
rep:     while (*pt != 'f') {
            while (*pt != 0 && *pt != ';' && *pt != ':') pt++;
            if (*pt == 0) goto org;
            pt++;
        }
        if (strncmp(name, ++pt, len) == 0 && pt[len] == '=') {
            pt += len + 1;
            lbpf->data[font_no].no = atoi(pt);
            while (*pt >= '0' && *pt <= '9') pt++;
            for (;;) {
                switch (*pt) {
                  case 'G':
                      fg |= F_GOTH;
                      break;
                  case 'F':
                      fg |= F_FILL;
                      break;
                  case 'S':
                      fg |= F_SHADOW;
                      break;
                  case 'O':
                      fg |= F_OUTLINE;
                      break;
                  case 'H':
                      fg |= F_GGOTH;
                      break;
                  default:
                      goto next;
                }
                pt++;
            }
next:       while (*pt == ';' || *pt == ':') pt++;
            if (*pt >= '0' && *pt <= '9')  vh_ratio = atoi(pt);
            else  goto getf;
            while (*pt != 0 && *pt != ';' && *pt != ':') pt++;
            if (*pt != 0) pt++;
            if (*pt >= '0' && *pt <= '9') {
                weight = atoi(pt);
            }
            else  goto getf;
            for (; *pt;) {
                while (*pt != 0 && *pt != ';' && *pt != ':') pt++;
                if (*pt != 0) pt++;
                if (*pt < '0' || *pt > '9') break;
                if (fontc->k_width == atoi(pt)) {
                    if (vh_ratio == 1) return(NULL);
                    else if (vh_ratio == 0)
                        goto getf;
                    goto rep;
                }
            }
            if (vh_ratio == 0) return(NULL);
            goto getf;
        }
    }
org:
    font_no = (chk_goth(fontc)) ? 1 : 0;
getf:
    if (vh_ratio <= 1)  vh_ratio = f_vh_ratio;
    fontc->font_type = JLBP;
    fontc->f_goth |= fg;

    /* select font */
    kf = fontc->ext.kdir = (KFONT *)marea(KFONT_SIZE);
    kf->height = (2500L * lbp_orig_dpi + (tmp / 2)) / tmp;
    kf->width = fontc->k_width;
    if (fg & F_OUTLINE) fontc->k_width += k_width_adj;
    kf->name = lbpf->data[kf->fh = font_no].font_name;

    kf->size = ((unsigned long)kf->width * vh_ratio + 500L) / 1000L;

    kf->base = (int)(((long)(-k_base_l) * kf->size + 500) / 1000L) - k_base_d;
    /*  ESC/Page    (int)((((long)(152-k_base_l)*kf->size + 500) / 1000L) */

    (*lbpf->set_font) (fontc, kf, fg, weight, lbpf->data[font_no].no);
    return(kf);
}

long ptex_dpi_adj;
static int f = -1;

void p_out_char(PIXEL h, PIXEL v, FONT_INFO *font, uint code, PREAMBLE *pxl)
{
    /* we need the following for the use of FONT CACHE */
    if (font->font_type == UNKNOWN)
        get_lbp_font(font);

    if (ptex_d == 1) {            /* TATEGAKI */
        PIXEL charheight;

        if (code & 0xff00) {
            h -= pxl->width / 2;
            v += (PIXEL)(ptex_dpi_adj / pxl->height) - pxl->x_offset;
        }
        else {
            h -= pxl->height - pxl->y_offset;
            v -= pxl->x_offset;
        }
    }
    else {
        h -= pxl->x_offset;
        v -= pxl->y_offset;
    }

    (*(lbpf->move_cap)) (v, h);
    (*(lbpf->pre_put_char)) (font, &code);

    if (code & 0xff00) {
        if (draw_boxpic) {
            (*(lbpf->drawbox)) (v, h, (PIXEL)(ptex_dpi_adj / pxl->height),
                                pxl->width, 1);
        }
        else {
            (*(lbpf->put_jis_font)) (code);
        }
    }
#ifdef    FDOWN
    else if (draw_boxpic) {
        if (ptex_d == 1) {
            (*(lbpf->drawbox)) (v, h, -pxl->width, pxl->height, 2);
        }
        else if (ptex_d == 0) {
            (*(lbpf->drawbox)) (v, h, pxl->height, pxl->width, 2);
        }
    }
    else {
        (*(lbpf->put_dl_font)) (code);
    }
#endif
}

#ifdef FDOWN

int d_download(FONT_INFO *dfont, int flag)
{
    uchar b;
    int i, j;
    DL_INFO dl;
    BUFFER *pat;
    uchar *dp;
    PREAMBLE *preamble;

    dp = dfont->d_flag;
    if (flag)
        dp += (128 >> 3);
    dfont->f_goth &= ~F_DWN;    /* for the character with code >= 128 */
    if (num_dfont >= maxdfont) {
        for (i = 0; i < 16; i++)
            dp[i] = 0;
        return(0);
    }
    dl.total_send = 0L;
    dl.maxheight = dl.maxwidth = dl.numoffonts = 0;
    dl.mincharcode = -1;
    for (i = j = 0; i < 16; i++) {
        for (b = 1; b != 0; b <<= 1) {
            if (dp[i] & b) {
                preamble = get_font(flag + j, dfont);
                if (preamble->width > maxdwidth
                    || preamble->height > maxdheight) {
                    dp[i] &= ~b;
                }
                else {
                    dl.numoffonts++;
                    dl.total_send += (long)preamble->byte_width * preamble->height;
                    if (preamble->width > dl.maxwidth)
                        dl.maxwidth = preamble->width;
                    if (preamble->height > dl.maxheight)
                        dl.maxheight = preamble->height;
                    if (dl.mincharcode == -1)
                        dl.mincharcode = j;
                    dl.maxcharcode = j;
                }
            }
            j++;
        }
    }

    dl.total_send += (*(lbpf->get_dlhsize)) (dl.numoffonts);

    /* check if there is a room in LBP to download */
    if ((cur_dsize += dl.total_send) > max_dsize) return(1);
    if (dl.numoffonts == 0) return(0);

    num_dfont++;
    fprintf(stderr,"<%d:%d>", dl.assign_num = dfont->font_code, dl.numoffonts);

    (*(lbpf->pre_download)) (dfont, &dl, num_dfont, flag);

    for (i = -1, j = b = 0; j < 128; b <<= 1, j++) {
        if (b == 0) {
            b = 1;
            i++;
        }
        if (dp[i] & b)
            (*(lbpf->download_raster)) (dfont, &dl, j, flag);
    }
    (*(lbpf->post_download)) (dfont, &dl);
    dfont->f_goth |= F_DWN;
    return(0);
}

#endif

void pr_int2(ushort num)
{
    pr_putc(num >> 8);
    pr_putc(num & 0xff);
}

void pr_int4(long num)
{
    pr_putc(*((uchar*)&num + 3));
    pr_putc(*((uchar*)&num + 2));
    pr_putc(*((uchar*)&num + 1));
    pr_putc(*((uchar*)&num + 0));
}

void pr_putsc(char *str, int num)
{
    while (num-- > 0 && *str != '\0')
        pr_putc(*str++);
    while (num-- >= 0)
        pr_putc(0);
}

#endif /* LBP */

/* end of p_out.c */
