/*
 *  TEX Device Driver  ver 2.02-2.22
 *  (c) copyright 1988, 1989 by TSG, 1990 by SHIMA
 *
 *  pret.c : interpret module
 *     Jul.  1, 1989 : 2nd edition
 *     17 April 1992 : xxx1() modified for tpic specials by Oh-Yeah?
 *     14 June  1992 : modified for ptex by Oh-Yeah? (not finished yet ^^;)
 *                     Based on the specifications in pTeX 2.99 j1.7 1.0.9F
 *     30 June  1992 : modified for LIPS3 by SHIMA
 *     22 July  1992 : modified for ptex by Naochan (almost finished...)
 *                     Based on the specifications in pTeX 2.99 j1.7 1.0.9F
 *     8 August 1992 : modified for pTeX+LIPS3 by OkI
 *
 *  modified for Minix-386 1992 by obuk, youchan
 */

#include "config.h"
#include "dd.h"
#include "err.h"
#include "dviread.h"
#ifdef LBP
#include "prtout.h"
#endif

static SCALED_PT h, v, w, x, y, z;
char ptex_d; /* ptex direction: 0 = yoko, 1 = tate */
char ptex_mode;
int use_lateral_kanji = 0;
extern int prt_flag;

static PIXEL H0, V0;
#ifdef LBP
static PIXEL H1, V1;
static PIXEL H2, V2;
#endif
static FILE* dvi;
static long c[ 10 ];
    /* count0-9 (count0 is already read in init.c for making page-index) */
static FONT_INFO* font;
int    g_font;
    /* current font */

/* bitmap.c */
void write_font(PIXEL, PIXEL, PREAMBLE *);
void write_rule(PIXEL, PIXEL, PIXEL, PIXEL);

/* buffer.c */
char *marea(unsigned int size);

/* getfont.c */
PREAMBLE* get_font(int, FONT_INFO *);

/* size.c, prtsize.c */
PIXEL sptopixel(SCALED_PT);
PIXEL vtopixel(SCALED_PT);

extern  FONT_INFO* font_info_root;
extern  char*      goth;
extern  BOOL       draw_baseline;
extern  BOOL       draw_boxpic;

/* stack.c */
struct STACK {
    /* stack */
    SCALED_PT  h, v, w, x, y, z;
    char ptex_d; /* ptex direction: 0 = yoko, 1 = tate */
};

extern struct STACK* stack_buf;
extern int depth, STACK_MAX;
extern double sp_per_h, v_per_h;

#ifndef NOTPIC
/* tpic.c */
extern BOOL (*tpic_special)(char *special);
/* ptr to func. interpret or ignore special. Ret TRUE if interpreted */
#endif

#ifdef FDOWN
#define D_BASE    0    /* arrange of dot height of downloaded fonts.
                          larger number means upper position. */

#define MIN_COUNT 4    /* minimal number of occurency / fonts to download */

extern int f_use_lfont;
extern int f_download;
extern int f_assigned;
extern int d_mincount;
extern long cur_dsize;

int  d_download(FONT_INFO *, int);
void d_work(void);
void d_init(void);
void d_resume(void);
static void d_put(ushort);
#endif

void interpret(OUTPUT_INFO *);
static void work(void);
static SCALED_PT spwidth(PREAMBLE*);
static SCALED_PT cput(int);
static SCALED_PT rput(void);
static void change_font(long);

#ifndef NOTPIC
char *read_special(FILE *fp, unsigned int len);
PIXEL get_current_x(void);
PIXEL get_current_y(void);
#endif

#ifdef LBP
extern DIMENSION dviout_dimension;
void  p_out_char(PIXEL, PIXEL, FONT_INFO *, uint, PREAMBLE *);
#endif

static void rewrite_preamble_ptex(PREAMBLE* , uint);

void interpret(OUTPUT_INFO* output)
{
    int i;
    
    font = NULL;
    h = v = w = x = y = z = 0;
    ptex_d = 0;
    dvi = output->dvifile_ptr;
    H0 = output->h_0;
    V0 = output->v_0;
#ifdef LBP
    H1 = output->width;
    V1 = output->height;
    if (output->print_direction == HORIZONTAL) {
        H2 = 0;
        V2 = (output->split - 1)*dviout_dimension.buf_height;
    }else{
        H2 = (output->split - 1)*dviout_dimension.buf_width;
        V2 = 0;
    }
#endif
    
    if ((uchar)read_byte(dvi) != BOP)
        error(COMMAND_ERROR, "No BOP");
    
    for (i = 0; i < 10; ++i)
        c[ i ] = read_long(dvi);
    
    fseek(dvi,  4L, SEEK_CUR);
        /* skip file-offset of next page */

#ifdef FDOWN
    if (f_download == 1)  d_work();
    else
#endif    
    work();
}


static SCALED_PT cput(int num)
    /* put character num on (h,v) (not change both h and v) */
    /* return sp-width of the character */
{
    PIXEL H, V, feed;
    int   g_loop;
    PREAMBLE* preamble;
    
    preamble = get_font((uint)num, font);
    if (ptex_mode
#ifdef LBP
        && (font->font_type != JLBP || draw_boxpic)
#endif
      )  rewrite_preamble_ptex(preamble, num);
    if (num == 0x2121) goto end;
    H = sptopixel(h) - H0;
    V = vtopixel(v) - V0;
    if (draw_baseline != 0) {
        if (ptex_d == 1) {
            feed = vtopixel(preamble->tfm_width);
            write_rule(H, V + feed, feed, 1);
        }else{
            write_rule(H, V, 1, sptopixel(preamble->tfm_width));
        }
        if (draw_boxpic) {
            write_font(H, V, preamble);
            goto end;
	}
    }
    if (prt_flag != 0) {
#ifdef FDOWN
        if (f_assigned) {           /* downloaded charactor */
            if (ptex_d==1)  H += D_BASE;
            else  V += preamble->height+D_BASE;
            goto usedw;
        }
#endif
#ifdef LBP
        if (g_font & F_LBP) {       /* scalable font */
usedw:      if (H >= 0 && V >= 0 && H < H1 && V < V1) {
                p_out_char(H + H2, V + V2, font, num ,preamble);
            }
            goto end;
        }
#endif
    }
    g_loop = (g_font & F_GOTH);
rep:
#if 0
    srp_hold = preamble->shift_right_ptex;
    sup_hold = preamble->shift_up_ptex;
    preamble->shift_right_ptex = preamble->shift_up_ptex = 0;
    write_font(H, V, preamble);
    preamble->shift_right_ptex = srp_hold;
    preamble->shift_up_ptex = sup_hold;
#endif
    write_font(H, V, preamble);
    if (g_loop) {
        g_loop = 0;
        if (ptex_d == 1) V++;
        else H++;
        goto rep;
    }
end:
    return(preamble->tfm_width);
}

static SCALED_PT rput(void)
    /* put rule on (h,v) (not change both h and v) */
    /* return sp-width of the rule */
{
    PIXEL A, B;
    SCALED_PT a, b;
    
    a = read_long(dvi);
    b = read_long(dvi);
    if (ptex_d == 1) {
        if (b > 0 && a != 0) {
            A = sptopixel(a + (SCALED_PT)((a>0)?(sp_per_h):(-sp_per_h)));
            B = vtopixel(b + (SCALED_PT)v_per_h);
            write_rule(sptopixel(h) - H0, vtopixel(v + v_per_h + b) - V0,
                        B, A);
        }
    }
    else {
        if (b > 0 && a != 0) {
            A = vtopixel(a + (SCALED_PT)((a>0)?(v_per_h):(-v_per_h)));
            B = sptopixel(b + (SCALED_PT)sp_per_h);
            write_rule(sptopixel(h) - H0, vtopixel(v) - V0, A, B);
        }
    }
    return(b);
}

static void font_change(int num)
    /* change current font into font 'num' */
{
    for (font = font_info_root; font != NULL ;) {
        if (font->font_code == num) {
            g_font = (font->f_goth & (F_LBP|F_DWN|F_ATTR));
            return;
        }
        font = (font->font_code < num)?
            font->after_font:font->before_font;
    }
    
    error(FONT_MISMATCH, "%d", num);
}

static void set1(int code)
{
    SCALED_PT  tmp;

    tmp = cput((uint)read_n(dvi, code - SET1 + 1));
    if (ptex_d == 1)  v += tmp;
    else              h += tmp;
}

static void set_rule(void)
{
    SCALED_PT  tmp;

    tmp = rput();
    if (ptex_d == 1)  v += tmp;
    else              h += tmp;
}

static void put1(int code)
{
    (void)cput((uint)read_n(dvi, code - PUT1 + 1));
}

static void put_rule(void)
{
    (void)rput();
}

static void nop(void)
{
}

static void push()
{
    struct  STACK* stack;

    (stack = &(stack_buf[depth]))->h = h;
    stack->v = v;
    stack->w = w;
    stack->x = x;
    stack->y = y;
    stack->z = z;
    stack->ptex_d = ptex_d;
    if (++depth >= STACK_MAX)  error(PROGRAM_STOP, "Stack Over flow");
}

static void pop()
{
    struct  STACK*  stack;
    
    if (--depth < 0)  error(PROGRAM_STOP, "Stack Under flow");
    h = (stack = &(stack_buf[depth]))->h;
    v = stack->v;
    w = stack->w;
    x = stack->x;
    y = stack->y;
    z = stack->z;
    ptex_d = stack->ptex_d;
}

static void right1(int code)
{
    SCALED_PT  tmp;

    tmp = signed_read_n(dvi, code - RIGHT1 + 1);
    if (ptex_d == 1)  v += tmp;
    else              h += tmp;
}

static void w0()
{
    if (ptex_d == 1)  v += w;
    else              h += w;
}

static void w1(int code)
{
    w = signed_read_n(dvi, code - W1 + 1);
    if (ptex_d == 1)  v += w;
    else              h += w;
}

static void x0()
{
    if (ptex_d == 1)  v += x;
    else              h += x;
}

static void x1(int code)
{
    x = signed_read_n(dvi, code - X1 + 1);
    if (ptex_d == 1)  v += x;
    else              h += x;
}

static void down1(int code)
{
    SCALED_PT  tmp;

    tmp = signed_read_n(dvi, code - DOWN1 + 1);
    if (ptex_d == 1)  h -= tmp;
    else              v += tmp;
}

static void y0()
{
    if (ptex_d == 1)  h -= y;
    else              v += y;
}

static void y1(int code)
{
    y = signed_read_n(dvi, code - Y1 + 1);
    if (ptex_d == 1)  h -= y;
    else              v += y;
}

static void z0()
{
    if (ptex_d == 1)  h -= z;
    else              v += z;
}

static void z1(int code)
{
    z = signed_read_n(dvi, code - Z1 + 1);
    if (ptex_d == 1)  h -= z;
    else              v += z;
}

static void fnt1(int code)
{
    font_change((int)read_n(dvi, code - FNT1 + 1));
}

static void xxx1(int code)
    /* specials are handed to external interpreters */
{
#ifndef NOTPIC
    char *special;

    special = read_special(dvi, (uint)read_n(dvi, code - XXX1 + 1));
    if (tpic_special(special)) return;  /* found */

      /* Here may come a call to non-tpic interpreters ... */
#else
    fseek(dvi, (long)read_n(dvi, code - XXX1 + 1), SEEK_CUR);
      /* just ignore and skip it */
#endif
}

static void fnt_def_1(int code)
{
    int skip;

    fseek(dvi,  (long)(code - FNT_DEF_1 + 13), SEEK_CUR);
    skip = (uchar)read_byte(dvi) + (uchar)read_byte(dvi);
    fseek(dvi,  (long)skip, SEEK_CUR);
}

static void illegal(int code)
    /* unknown commands */
{
    if (ptex_mode != 0 && code == 255)
        ptex_d = read_byte(dvi);
    else  error(COMMAND_ERROR, "Illegal command %d", code);
}


/*      set_char_0,          0 SET_CHAR_0   */

static void_func_ptr cmd128[]={
        set1,           /* 128 SET1         */
        set1,           /* 129              */
        set1,           /* 130              */
        set1,           /* 131              */
        set_rule,       /* 132 SET_RULE     */
        put1,           /* 133 PUT1         */
        put1,           /* 134              */
        put1,           /* 135              */
        put1,           /* 136              */
        put_rule,       /* 137 PUT_RULE     */
        nop,            /* 138 NOP          */
        illegal,        /* 139 BOP          */
        illegal,        /* 140              */
        push,           /* 141 PUSH         */
        pop,            /* 142 POP          */
        right1,         /* 143 RIGHT1       */
        right1,         /* 144              */
        right1,         /* 145              */
        right1,         /* 146              */
        w0,             /* 147 W0           */
        w1,             /* 148 W1           */
        w1,             /* 149              */
        w1,             /* 150              */
        w1,             /* 151              */
        x0,             /* 152 X0           */
        x1,             /* 153 X1           */
        x1,             /* 154              */
        x1,             /* 155              */
        x1,             /* 156              */
        down1,          /* 157 DOWN1        */
        down1,          /* 158              */
        down1,          /* 159              */
        down1,          /* 160              */
        y0,             /* 161 Y0           */
        y1,             /* 162 Y1           */
        y1,             /* 163              */
        y1,             /* 164              */
        y1,             /* 165              */
        z0,             /* 166 Z0           */
        z1,             /* 167 Z1           */
        z1,             /* 168              */
        z1,             /* 169              */
        z1              /* 170              */
};

    /*      fnt_num_0,     171 FNT_NUM_0    */

static void_func_ptr cmd235[]={
        fnt1,           /* 235 FNT1         */
        fnt1,           /* 236              */
        fnt1,           /* 237              */
        fnt1,           /* 238              */
        xxx1,           /* 239 XXX1         */
        xxx1,           /* 240              */
        xxx1,           /* 241              */
        xxx1,           /* 242              */
        fnt_def_1,      /* 243 FNT_DEF_1    */
        fnt_def_1,      /* 244              */
        fnt_def_1,      /* 245              */
        fnt_def_1,      /* 246              */
        illegal,        /* 247 PRE          */
        illegal,        /* 248              */
        illegal,        /* 249              */
        illegal,        /* 250              */
        illegal,        /* 251              */
        illegal,        /* 252              */
        illegal,        /* 253              */
        illegal,        /* 254              */
        illegal         /* 255 EOFNC(pTeX dir) */
};

static void_func_ptr *cmd128_ptr = cmd128;
static void_func_ptr *cmd235_ptr = cmd235;
  /* ptr to command tables. used for override work() */

static void work(void)
{
    int code;

    while((code = (uchar)read_byte(dvi)) != EOP) {
        if (code < 128) {
            if (ptex_d == 1)  v += cput((uint)code);
            else              h += cput(code);
	}
        else if (code < FNT_NUM_0) (cmd128_ptr[code - 128])(code);
        else if (code < FNT1)      font_change((int)(code - FNT_NUM_0));
        else                       (cmd235_ptr[code-FNT1])(code);
    }
}

#ifdef FDOWN
void d_put(ushort code)
{
    int i;

    font->ext.count++;
    if (code & 0xff00)  return;        /* return if 2 bytes code */
    if (font->pk == NULL) {
        font->ext.count = 0;
        font->pk = bitmap_buf_pointer->current;
        (BUFFER*)bitmap_buf_pointer->current += 256;
        if ((BUFFER*)bitmap_buf_pointer->current >
          (BUFFER*)bitmap_buf_pointer->end)
            error(NO_MEMORY, "Not enough bitmap buffer");
    }
    if ((font->pk)[code] != 0xff)  (font->pk)[code]++;
}

void d_resume(void)
{
    BUFFER    *pt;
    FONT_INFO *dfont;
    uchar  *dw;
    long   maxcount;
    int    i, j, k, c_max;
    uchar  b;

    i = 0;
    for(font = first_font_info; font; font = font->next_font) {
        font->last = 0;
        if ((pt = font->pk) == NULL)  continue;
        for(c_max = 0xff; c_max >= 0x80 && pt[c_max] == 0; c_max--);
        i += font->last = (c_max >= 0x80)?(256>>3):(128>>3);
    }

    if (i > 0)  dw = marea(i);

    for(font = first_font_info; font; font = font->next_font) {
        if ((pt = font->pk) == NULL) {
            font->ext.count = 0;
            continue;
        }
        font->pk = NULL;
        c_max = (font->last << 3);
        font->d_flag = dw;
        for(i = -1, j = k = b = 0; j < c_max; b <<= 1, j++) {
            if (b == 0) {
                b = 1;
                dw[++i] = 0;
            }
            if (pt[j] >= d_mincount && (j & 0x7f) != 0x20) {
                dw[i] |= b;
                k = j;
            }
        }
        if (k < 128)  font->last = (128 >> 3);
        dw += font->last;
    }

    fprintf(stderr,"\nDownLoading:");
    for(; ;) {
        for(maxcount = MIN_COUNT-1, dfont = NULL, font = first_font_info;
          font; font = font->next_font) {
            if ((font->f_goth & F_DWN) == 0 && font->last != 0
              && font->ext.count > maxcount) {
                  maxcount = font->ext.count;
                  dfont = font;
            }
        }
        if (dfont == NULL)  break;        /* no more font to download */
        if (d_download(dfont, 0) != 0)  goto end;
        if (dfont->last > (128>>3) && d_download(dfont, 0x80) != 0) goto end;
        dfont->last = 0;
    }
end:
    bitmap_buf_pointer->current = bitmap_buf_pointer->start;
    fprintf(stderr, " %dK byte", (int)((cur_dsize + 512) >> 10));
}

void d_work(void)
{
/*      set_char_0,          0 SET_CHAR_0   */
static uchar cmd128[]={
        0x31,           /* 128 SET1         */
        0x32,           /* 129              */
        0x33,           /* 130              */
        0x34,           /* 131              */
        8,              /* 132 SET_RULE     */
        1,              /* 133 PUT1         */
        2,              /* 134              */
        3,              /* 135              */
        4,              /* 136              */
        8,              /* 137 PUT_RULE     */
        0,              /* 138 NOP          */
        0x10,           /* 139 BOP          */
        0x10,           /* 140              */
        0,              /* 141 PUSH         */
        0,              /* 142 POP          */
        1,              /* 143 RIGHT1       */
        2,              /* 144              */
        3,              /* 145              */
        4,              /* 146              */
        0,              /* 147 W0           */
        1,              /* 148 W1           */
        2,              /* 149              */
        3,              /* 150              */
        4,              /* 151              */
        0,              /* 152 X0           */
        1,              /* 153 X1           */
        2,              /* 154              */
        3,              /* 155              */
        4,              /* 156              */
        1,              /* 157 DOWN1        */
        2,              /* 158              */
        3,              /* 159              */
        4,              /* 160              */
        0,              /* 161 Y0           */
        1,              /* 162 Y1           */
        2,              /* 163              */
        3,              /* 164              */
        4,              /* 165              */
        0,              /* 166 Z0           */
        1,              /* 167 Z1           */
        2,              /* 168              */
        3,              /* 169              */
        4               /* 170              */
};

    /*      fnt_num_0,     171 FNT_NUM_0    */

static uchar cmd235[]={
        0x41,           /* 235 FNT1         */
        0x42,           /* 236              */
        0x43,           /* 237              */
        0x44,           /* 238              */
        0x51,           /* 239 XXX1         */
        0x52,           /* 240              */
        0x53,           /* 241              */
        0x54,           /* 242              */
        0x21,           /* 243 FNT_DEF_1    */
        0x22,           /* 244              */
        0x23,           /* 245              */
        0x24,           /* 246              */
        0x10,           /* 247 PRE          */
        0x10,           /* 248              */
        0x10,           /* 249              */
        0x10,           /* 250              */
        0x10,           /* 251              */
        0x10,           /* 252              */
        0x10,           /* 253              */
        0x10,           /* 254              */
        0x10            /* 255 EOFNC        */
};
    int  code, mode;
    long tmp;

    while((code = (uchar)read_byte(dvi)) != EOP) {
        if (code < 128)  d_put((ushort)code);
        else if (code < FNT_NUM_0) {
            mode = cmd128[code - 128];
            goto work;
        }
        else if (code < FNT1)  font_change((int)(code - FNT_NUM_0));
        else{
            mode = cmd235[code - FNT1];
work:
            if (mode >= 0x30)  tmp = (long)read_n(dvi, mode & 0xf);
            switch (mode >> 4) {
                case 0:  while(mode--) read_byte(dvi);
                         break;
                case 1:  illegal(code);
                         break;
                case 2:  fseek(dvi, (long)((mode & 0xf) + 12), SEEK_CUR);
                         tmp = (uchar)read_byte(dvi) + (uchar)read_byte(dvi);
skip:                    while(tmp--) read_byte(dvi);
                      /* fseek(dvi, tmp, SEEK_CUR); the above is quicker*/
                         break;
                case 3:  d_put((ushort)tmp);
                         break;
                case 4:  font_change((int)tmp);
                         break;
                case 5:  goto skip;
            }
        }
    }
}
#endif
 
/* end of file : pret.c */

/* additional functions by Oh-Yeah? */

#ifndef NOTPIC

char *read_special(FILE *fp, uint len)
  /* read str. caution: buffer overwritten */
{
  static uint buf_len = 0;
  static char *buf_ptr = NULL;
  char *pt;

  if (buf_len < len) {    /* not enough, re-allocate */
      if (buf_ptr != NULL)   free(buf_ptr);
      buf_ptr = (char *)marea(len + 1);
      buf_len = (buf_ptr == NULL) ? 0 : len;
  }
  if (buf_ptr != NULL) {
      pt = buf_ptr;
      while(len-- > 0) *pt++ = getc(fp);
      *pt = '\0';
  }
  return(buf_ptr);
}

PIXEL get_current_x(void)
{
    return((PIXEL)sptopixel(h) - H0);
}

PIXEL get_current_y(void)
{
    return((PIXEL)vtopixel(v) - V0);
}

#endif /* end of ifndef NOTPIC */

static void rewrite_preamble_ptex(PREAMBLE* preamble, uint char_code)
{
    PIXEL   p_up, p_right, tmp;
    char    p_rotate, p_mirror;

    p_up = p_right = p_rotate = p_mirror = 0;
    if (ptex_d == 1) {
        tmp = sptopixel(preamble->tfm_width);
        if (0 < char_code && char_code < 0x0100)
            p_rotate = 3;
        else if ((0x2145 < char_code) && (char_code < 0x215C))
            p_rotate = 1;
        else 
            switch(char_code) {
                case 0x2122:
                case 0x2123:
                case 0x2124:
                case 0x2125: p_up = tmp*5/4;
                             p_right = tmp*11/8;
                             break;
                case 0x2421:
                case 0x2423:
                case 0x2425:
                case 0x2427:
                case 0x2429:
                case 0x2443:
                case 0x2463:
                case 0x2465:
                case 0x2467:
                case 0x246E:
                case 0x2521:
                case 0x2523:
                case 0x2525:
                case 0x2527:
                case 0x2529:
                case 0x2543:
                case 0x2563:
                case 0x2565:
                case 0x2567:
                case 0x256E:
                case 0x2575:
                case 0x2576: p_up = tmp/4;
                             p_right = tmp/6;
                             break;
                case 0x213D:
                case 0x2144:
                case 0x2145:
                case 0x2126:
                case 0x213E:
                case 0x2142:
                case 0x2143:
                case 0x2161:
                case 0x2162:
                case 0x222A:
                case 0x222B:
                case 0x222C:
                case 0x222D:
                case 0x222E: p_rotate = 1;
                             break;
                case 0x213C:
                case 0x2141: p_rotate = p_mirror = 1;
                default:     break;
            }
        if (!use_lateral_kanji) { 
            p_up = p_right = 0;
            p_rotate >>= 1;  /* p_rotate:  3 -> 1, 1 -> 0 */
        } else {
            p_rotate &= 1;
            if (p_mirror == 1) p_rotate |= 2;
        }
    }
    preamble->shift_up_ptex = p_up;
    preamble->shift_right_ptex = p_right;
    preamble->rotate_ptex = p_rotate;
}
  
/* end of all */
