/*
 *  TEX Device Driver  ver 2.02-2.32
 *  copyright(c) 1988, 1989 by TSG, 1990 by SHIMA
 *  1991 changed by T.Minagawa
 *
 *  init.c : initialize variables
 *      Apr. 29, 1989 : 1st edition
 *      Jun.  1, 1989 : 2nd edition
 *
 *  This module is a device-dependent one.
 *  modified for non PC-9801 by sempa 1992
 *           14 June 1992 : modified for tpic by Oh-Yeah?
 *  modifed for LIPS3 by SHIMA 30 June 1992
 *  modified for pTeX `tate' mode,  6 August 1992  by Naochan!.
 *  slightly modified by OkI 8 August 1992 (pTeX & Hires & LIPS3)
 *  modified for "-wait" option by sempa, 1 Sept 1992
 *  modified for Minix-386 and Linux by obuk, youchan 1992
 *  modified for LIPS3 and ESC/Page by OkI and H.Tomiie 30 Sep 1992
 */

#include "config.h"
#include "dd.h"
#include "err.h"
#include "dviread.h"
#include "prtout.h"
#include "option.h"
#include "version.h"
#include "vfont.h"

void get_config(DIMENSION*, int, char **);
void initialize(DVIFILE_INFO*, DIMENSION*);
static void set_prt_type(DIMENSION *dim);
static void read_post(DVIFILE_INFO*);
static void pixel_init(DIMENSION*);
static void set_dimension(DVIFILE_INFO*, DIMENSION*);
static void make_page_index(DVIFILE_INFO*, DIMENSION*);
static void opt_cfg(char*);

/* size.c, prtsize.c */
void set_convert(SIZE_PARA, int, int);
PIXEL sptopixel(SCALED_PT);
PIXEL vtopixel(SCALED_PT);
int set_size(PIXEL, PIXEL, PIXEL, PIXEL, PIXEL *, PIXEL *);

/* fontdef.c */
void make_font_list(DVIFILE_INFO *, DIMENSION *);

/* buffer.c */
void buffer_init(void);

/* p_out.c */
int get_prt_ctl(char *);

/* stack.c */
void stack_init(int);

#ifndef NOTPIC
/* tpic.c */
void tpic_init(PIXEL buf_height); /* alloc y-bucket[vmax] and activate */
#endif

/* vfont.c */
int get_vfont_name(char *);

/* vfont.c escpage.c Tomiie*/
#ifndef NOTATEGAKI
int    tategaki=0;
#endif

BOOL f_lbp_prt;
int test_flg = -1;
int prt_flag = 1;
static BOOL mag_half = FALSE;
static BOOL opt_mag = -1;
static BOOL f_center = FALSE;
BOOL dvifile_page = FALSE;

static  int add_width;
static  int add_height;
static  int vert_dpi;
static  int enlarge;

#define CFG_PATH    "TEXCFG"
char    *cfg_path;

const char* const config_name = "dviprt.cfg";
    /* config file name */
const char* const pk_env_name = "TEXPK";
    /* name of the env-variable for pk-font path */
const char* const title_name =
    "TeX PRINTER DRIVER-32 "
    MACHINE VERSION;
const char* const title_comment =
    "";
const char* const title_copyright =
    "\nCopyright(C) 1988-89 by TSG, 90-91 by SHIMA,"
    "\n             1991-92 changed by T.Minagawa, Akiii, Oh-Yeah?, SHIMA, sempa"
    "\n             1992    changed by Naochan!, hero.h, OkI, Tomiie"
    "\n             1992    changed by obuk, youchan (32bit version)"
    "\n             1992    changed by mkiguchi, ogochan (msdos version)";

const char* const vfont_config = "dviprt.vfn";
    /* vector font config file name */

#ifndef NOTPIC
BOOL f_tpic_turn_on = TRUE; /* default is on */
#endif
#ifdef LBP
extern int   f_use_lfont;
extern int   f_vh_ratio;
extern int   f_rotate;
extern int   f_copy;
extern int   d_mincount;
extern char  *youshi;
extern int   k_base_l;
extern int   k_scale_l;
extern int   lbp_orig_dpi;
extern LBP_FUNC *lbpf;
#ifdef LIPS3
extern LBP_FUNC lips3_func;
#endif
#ifdef ESCPAGE
extern LBP_FUNC escpage_func;
#endif
#endif

#ifdef FDOWN
#define MAX_FONTDEF  6
extern int   f_download;
extern long  max_dsize;
extern char* font_def;
#endif

extern char  prt_type;
extern int   ff_flg;
extern BOOL  f_long_name;
extern BOOL  f_font_c;
extern BOOL  f_pause;
extern BOOL  f_top_margin;
extern int   k_scale;
extern int   k_base;
extern int   f_search;
extern char  *goth;
extern char  *sub_font;
extern char  *fc_name;
extern char  *k_dbl;
extern char  *f_page_scroll;
#ifdef msdos
extern char  *f_out;
#endif
extern char  *prt_initialize;

extern BOOL  use_lateral_kanji;
    /* using lateral Kanji Font for vertical writing (pTeX)
     */
extern BOOL  draw_baseline;
    /* [for hidden option-1 by Naochan!]
     * draw the baseline.
     */
extern BOOL  draw_boxpic;
    /* [for hidden option-2 by Naochan!]
     * typesetting with blank boxes (instead of normal characters.)
     */
extern char  ptex_mode;
    /* when `extended dvi file' encountered, this value is not zero.(pTeX)
     */
extern int   f_use_vfont;

extern char tmp_buf[];

#define BUFFER_NUM 5

static BUF_INFO buffers[ BUFFER_NUM ] = {
    {NULL,NULL,NULL, DEF_RASTER_SIZE, f_HUGE},
    {NULL,NULL,NULL, DEF_PK_SIZE, f_HUGE},
    {NULL,NULL,NULL, DEF_BITMAP_SIZE, f_HUGE}
};

FONT_INFO* first_font_info = NULL;
FONT_INFO* font_info_root = NULL;

BUF_INFO* raster_buf_pointer    = &buffers[0];
BUF_INFO* pk_buf_pointer        = &buffers[1];
BUF_INFO* bitmap_buf_pointer    = &buffers[2];

DIMENSION dviout_dimension = {
    PRINTER_DPI,                /* dpi */
    0, 0,                     /* device width, height */
    0, 0,                       /* buffer width, height */
    DEF_MAG,                    /* mag */
    0, 0,                       /* text width, height */
    0,                          /* split */
    DEF_PRINT_DIR,              /* print direction */
    0.0,                        /* convert parameter */
    0, 0,                       /* x-offset, y-offset */
    0,                          /* offset of the page */
    0,                          /* total page */
    0, 0,                       /* start page, end page */
    PRT_TYPE,                   /* device type */
};

static char *pt_prt_type;

static ARG_TABLE option[] = {
    {"mag",  INTEGER,    (void *)&opt_mag,
        "mag step (0-5)"},
    {"half", BOOLEAN,    (void *)&mag_half,
        "mag step (half)"},
    {"f",    INTEGER,    (void *)&test_flg,
        "test-flag"},
    {"dpi",  INTEGER,    (void *)&(dviout_dimension.dpi),
        "horizontal dpi"},
    {"DPI",  INTEGER,    (void *)&(vert_dpi),
        "vertical dpi"},
    {"e",    INTEGER,    (void *)&(enlarge),
        "enlarge*1000"},
    {"page", BOOLEAN,    (void *)&dvifile_page,
        "pages are dvifile-pages"},
    {"nf",   INTEGER,    (void *)&ff_flg,
        " No use of form feed code"},
    {"o",    STRING,     (void *)&f_page_scroll,
        "output (r:reverse, e:even pages, o:odd pages)"},
    {"V",    BOOLEAN,    (void *)&(dviout_dimension.print_direction),
        "vertical-output"},
    {"X",    INTEGER,    (void *)&(dviout_dimension.x_offset),
        "x-offset(pixel)"},
    {"Y",    INTEGER,    (void *)&(dviout_dimension.y_offset),
        "y-offset(pixel)"},
    {"w",    INTEGER,    (void *)&add_width,
        "add width(pixel)"},
    {"h",    INTEGER,    (void *)&add_height,
        "add length(pixel)"},
    {"W",    INTEGER,    (void *)&(dviout_dimension.max_width), 
        "device width(pixel)"},
    {"H",    INTEGER,    (void *)&(dviout_dimension.max_height), 
        "device height(pixel)"},
    {"T",    BOOLEAN,    (void *)&f_top_margin,
        "correct Top margin for cut sheets"},
    {"C",    BOOLEAN,    (void *)&f_center,
        "Centering"},
#ifndef NOTPIC /* this should be before "t" */
    {"tpic", BOOLEAN,    (void *)&f_tpic_turn_on,
        "Support tpic special [default +]"},
#endif
    {"r",    STRING,     (void *)&fc_name,
        "Save Font"},
    {"L",    BOOLEAN,    (void *)&f_long_name,
        "cut center of long font names"},
    {"c",    BOOLEAN,    (void *)&f_font_c,
        "Check font matching"},
    {"F",    STRING,     (void *)&sub_font,
        "Font substiute"},
    {"A",    INTEGER,    (void *)&f_search,
        "Font search"},
    {"K",    STRING,     (void *)&k_dbl,
        "Double size Kanji"},
    {"G",    STRING,     (void *)&goth,
        "Gothic"},
    {"g",    BOOLEAN,    (void *)&use_lateral_kanji,
        "Use Lateral Kanji Font in pTeX `tate' mode"},
    {"base", BOOLEAN,    (void *)&draw_baseline,
        "draw the baseline"},
    {"box",  BOOLEAN,    (void *)&draw_boxpic,
        "typeset by boxes instead of characters"},
    {"vfn",  BOOLEAN,    (void *)&f_use_vfont,
        "Use Syotai Club's Font"},
    {"S", INTEGER,       (void *)&k_scale,
        "Kanji scale*1000 (default 952)"},
    {"J",    INTEGER,    (void *)&k_base,
        "Kanji base line*1000 (default 0)"},
#ifdef LBP
    {"y",    STRING,     (void *)&youshi,
        "Paper select [A4|B4|A5|B5|H|F[width]<:[height]>|T[number]]"
#ifdef LIPS3
        "\n\tLIPS3:     [p(port.)|l(land.)<0(auto)|1(manual)|2(upper)|3(lower)>]"
#endif
#ifdef ESCPAGE
        "\n\tESC/Page:  [p(port.)|l(land.)<1(normal)|2(option)<:bin>>]"
#endif
        },
    {"lc",   INTEGER,    (void *)&f_copy,
        "Number of copy by LBP"},
#endif
#ifndef NOTATEGAKI
    {"R",    BOOLEAN,    (void *)&tategaki,
        "Font rotational printing with "
#ifdef LBP
        "LIPS3,ESC/Page and "
#endif
        "Zeit vector font."},
#endif
#ifdef msdos
    {"O",    STRING,     (void *)&f_out,
        "File output"},
#endif
    {"Z",    BOOLEAN,    (void *)&f_pause,
        "Pause for every page"},
    {"I",    STRING,     (void *)&prt_initialize,
        "File with code initializing printer"},
    {"p",    STRING,     (void *)&pt_prt_type, 
        "type of printer e:ESC/P, p:PC-PR, n:NM, o:other (default prtctl.cfg)"
#ifdef LBP
        "\n\t"
#ifdef LIPS3
        "l:LIPS3 "
#endif
#ifdef ESCPAGE
        "m:ESC/Page "
#endif
        "(v:font_aspect d:download_size f:font_def D:dpi"
        "\n\t\tj:baseline(79/1000) s:size_adjust(1047/1000) o:other"},
#endif
    {"=",    PROCEDURE,  (void *)&opt_cfg,
        "configuration file"},
    {NULL}
};


void get_config(DIMENSION *dim, int argc, char **argv)
{
    int     i, l;
    char    real_name[MAXPATH], v_real_name[MAXPATH];
    char    *cfgpath;

    if ((cfgpath = getenv(CFG_PATH)) == NULL)
        cfgpath = CONFDIR;
    strncpy(real_name, cfgpath, MAXPATH-2);
    i = strlen(real_name);
    if (i > 0 && isalnum(real_name[i-1])) {
        real_name[i] = PATHCHAR;
        real_name[i+1] = '\0';
    }
    cfg_path = (char*)strdup(real_name);
    strcpy(v_real_name, real_name);
    strncat(real_name, (char *)config_name, MAXPATH-1);
    strncat(v_real_name, (char *)vfont_config, MAXPATH-1);

    set_config(option, real_name);
    set_option(option, argc, argv);
#ifndef   NOVFONT
    if (f_use_vfont && (0 != (l = get_vfont_name(v_real_name)))) {
        if (l > 0) {
            sprintf(tmp_buf,"[vfont] dviout.vfn/dviprt.vfn error.  line:%d\n", l);
            error(PROGRAM_STOP, tmp_buf);
        }
        else  error(PROGRAM_STOP,"[vfont] Can't open dviout.vfn/dviprt.vfn file.");
    }
#endif
}

void initialize(DVIFILE_INFO *dvi, DIMENSION *dim)
{
    read_post(dvi);
    set_prt_type(dim);
    set_dimension(dvi, dim);
    make_font_list(dvi, dim);
    make_page_index(dvi, dim);
    pixel_init(dim);
    buffer_init();
    stack_init(dvi->stack_depth);
#ifndef NOTPIC
    if (f_tpic_turn_on) tpic_init(dim->buf_height);
       /* alloc y-bucket[vmax] and activate tpic interpretation */
#endif
}

static void set_prt_type(DIMENSION *dim)
{
    int ch;
    char *pt;
#ifdef LBP
    int num;

    f_lbp_prt = FALSE;
#endif
    if (pt_prt_type != NULL) {
        switch(*pt_prt_type | 0x20) {
escp:       case (T_ESC_P):   
                dim->prt_type = ESC_P;
                if (dim->dpi==0) dim->dpi = 180;
                break;
            case (T_PC_PR):
                dim->prt_type = PC_PR;
                if (dim->dpi==0) dim->dpi = 160;
                break;
            case (T_PC_NM):
                dim->prt_type = PC_NM;
                if (dim->dpi==0) dim->dpi = 180;
                break;
#ifdef LIPS3
            case (T_LBP_LIPS3):
                dim->prt_type = LBP_LIPS3;
                lbpf = &lips3_func;
                f_lbp_prt = TRUE;
                break;  
#endif
#ifdef ESCPAGE
            case (T_LBP_ESCPAGE):
                dim->prt_type = LBP_ESCPAGE;
                lbpf = &escpage_func;
                f_lbp_prt = TRUE;
                break;  
#endif
            case (T_PR_OTHER):
                dim->prt_type = PR_OTHER;
                if (get_prt_ctl(pt_prt_type+1) == FAILURE)
                    error(ILLEGAL_ARGS, "Illegal printer control.");
                if (dim->dpi == 0) {
                    fprintf(stderr, "Warning : Assume Printer DPI = 180\n\n");
                        dim->dpi = 180;
                    }
                break;
  
            default:
                error(ILLEGAL_ARGS, "-p%s", pt_prt_type);
                break;
        }
    }
    else goto escp;

#ifdef LBP
    if (f_lbp_prt) {
        char *p;
        int page_orient = 0;

        f_use_lfont = TRUE;
        f_download = 1;
        pt = pt_prt_type + 1;
        while((ch=*pt++) != 0) {
            num = atoi(pt);
            switch(ch) {
              case 'v':  f_vh_ratio = num;
                         if (f_vh_ratio == 0) f_use_lfont = FALSE;
                         break;
              case 'd':  max_dsize = 400L * num;
                         if (max_dsize <= 0)  f_download = 0;
                         break;
              case 'j':  k_base_l = num;
                         break;
              case 's':  k_scale_l = num;
                         break;
              case 'D':  lbp_orig_dpi = num;
                         break;
              case 'o':  if (get_prt_ctl(pt) == FAILURE)
                             error(ILLEGAL_ARGS, "Illegal printer control.");
                         break;
              case 'm':  d_mincount = num;
                         break;
              case 'f':  if (font_def == NULL)  font_def = pt-1;
                         break;
            }
            while(*pt++ != ';' && *pt) pt++;
            if (*pt) pt++;
        }
        if (f_vh_ratio < 200 || f_vh_ratio > 5000)
            f_vh_ratio = 1000;
        if (dim->dpi==0) dim->dpi = 300;
        if (*youshi) {
            dim->print_direction = HORIZONTAL;
            for (p = youshi; (*p) && (tolower(*p) != 'l'); p++) ;
            if (*p == 'l') page_orient = 1;
        }
        if (dim->print_direction == VERTICAL || page_orient == 1) {
            if (dviout_dimension.max_width == 0)  dviout_dimension.max_width =  3387;
            if (dviout_dimension.max_height == 0) dviout_dimension.max_height = 2360;
            dim->print_direction = HORIZONTAL;
            f_rotate = TRUE;
        }
        else {
            if (dviout_dimension.max_width == 0)  dviout_dimension.max_width =  2360;
            if (dviout_dimension.max_height == 0) dviout_dimension.max_height = 3387;
        }
    }
#endif
    if (vert_dpi == 0) vert_dpi = dim->dpi;
    if (dviout_dimension.max_width == 0)
        dviout_dimension.max_width =  8*dim->dpi;
    if (dviout_dimension.max_height == 0)
        dviout_dimension.max_height = 11*vert_dpi;
#ifdef FDOWN
    if (f_download)  fc_name = NULL;
#endif
}

/* optional configuration file */
static void opt_cfg(char *opt)
{
    set_config(option, opt);
}

void help_out(void)
{
    printf("usage : dviprt [option(s)] dvifile [page(s) eg. -4 7 9 12-14 20-]\n\n");
    printf("options are as follows :\n");
    option_usage(option);
    printf("\noptions can be directed in %s\n", config_name);
    exit(0);
}

#define CM_INCH 2.54    /* cm per inch */

static  int pres;
static void set_dimension(DVIFILE_INFO* dvi, DIMENSION* dim)
{
    static  int true_mag[] = { 1000, 1200, 1440, 1728, 2074, 2488 };
    ulong   num1, num2;
    long    num3;
    int     q;

    dim->mag = (int)dvi->mag;
    
    if (mag_half)
            dim->mag = 1096;  /* magstep half */
    else {
        if (opt_mag != -1)
            if (opt_mag >= 0 && opt_mag <= 5)
                dim->mag = true_mag[ opt_mag ];
            else
                error(ILLEGAL_ARGS, "mag step over");
    }
    
    dim->total_page = dvi->total_page;

    if (dim->dpi <= 0) error(ILLEGAL_ARGS, "dpi");

    dim->size_para = ((double)dvi->num *(double)dim->dpi*(double)dim->mag)/
     ((double)dvi->den*CM_INCH*100000000.0);

    set_convert(dim->size_para, dim->dpi, vert_dpi);
        /* set pixel<==>scaled-pt parameter */

    dim->text_width = sptopixel(dvi->u);
    dim->text_height = vtopixel(dvi->l + HEADER_DEPTH + FOOTER_DEPTH);

    if (f_center == TRUE) {
        if (dim->print_direction == HORIZONTAL) {
            if (dim->max_width > 0) {
                dim->x_offset += (dim->max_width - dim->text_width)/2;
            }
            if (dim->max_height > 0) {
                dim->y_offset += (dim->max_height - dim->text_height)/2;
            }
        }else{
            if (dim->max_width > 0) {
                dim->x_offset += (dim->max_width - dim->text_height)/2;
            }
            if (dim->max_height > 0) {
                dim->y_offset += (dim->max_height - dim->text_width)/2;
            }
        }
    }
}


#define END_DVI 223
#define ID 2
#define ID_PTEX 3
    /* id number for searching for postamble
     */

static void read_post(DVIFILE_INFO* dvi)
{
    long endofs;
    int code;
    
    if ((uchar)read_byte(dvi->file_ptr) != PRE ||
      (uchar)read_byte(dvi->file_ptr) != ID) {
        error(FILE_FAULT, "Not DVI file");
    }

    for(endofs = -3L; fseek(dvi->file_ptr, endofs, SEEK_END),
        (code = (uchar)read_byte(dvi->file_ptr)) != ID; endofs--)
            /* Search id number
             */
        if (code == ID_PTEX) {
            fprintf(stderr,"Extended DVI file encountered. Turn on pTeX mode.\n");
            ptex_mode ++;
            break;
        } 
        else if (code == EOF || code != END_DVI)
            error(FILE_FAULT, "No ID.");
    
    fseek(dvi->file_ptr, endofs - 4L, SEEK_END);
    if ((dvi->post = read_long(dvi->file_ptr)) <= 0)
        error(COMMAND_ERROR, "Negative Pointer (POST) ");
        /* Read the position of POSTAMBLE */
    
    fseek(dvi->file_ptr, dvi->post, SEEK_SET);
        /* Set file-ptr at POSTAMBLE */
    
    if ((uchar)read_byte(dvi->file_ptr) != POST)
        error(COMMAND_ERROR, "No Postamble ");
    
    if ((dvi->last_bop = read_long(dvi->file_ptr)) <= 0)
        error(COMMAND_ERROR, "Negative Pointer (Last BOP) ");
    
    dvi->num = read_long(dvi->file_ptr);
    dvi->den = read_long(dvi->file_ptr);
    dvi->mag = read_long(dvi->file_ptr);
    
    dvi->l = (SCALED_PT) read_long(dvi->file_ptr);
    dvi->u = (SCALED_PT) read_long(dvi->file_ptr);
    
    dvi->stack_depth = (int)read_short(dvi->file_ptr);
    dvi->total_page = (int)read_short(dvi->file_ptr);
    
    if (dvi->num <= 0 || dvi->den <= 0 || dvi->mag <= 0)
        error(COMMAND_ERROR, "Reading Illegal Long ");
    
    if (dvi->stack_depth < 0 || dvi->total_page <= 0)
        error(COMMAND_ERROR, "Reading Illegal Integer ");
}


static void pixel_init(DIMENSION* dim)
{
    long    free_size;
    int     i;

    if (buffers[2].size == 0L)
        buffers[2].size = ((dim->max_width + 7) / 8) * (dim->max_height);
    if (buffers[0].size <= 0x10000) buffers[0].size = 0x10000;
    if (dim->print_direction == HORIZONTAL) {
        dim->text_width += dim->x_offset + add_width;
        dim->text_height += dim->y_offset + add_height;
        dim->split = set_size(dim->text_width, dim->text_height,
            dim->max_width, dim->max_height,
            &dim->buf_width, &dim->buf_height);
   }
    else {
        dim->text_width += dim->y_offset + add_height;
        dim->text_height += dim->x_offset + add_width;
        dim->split = set_size(dim->text_height, dim->text_width,
            dim->max_width, dim->max_height,        /* converted by SHIMA */
            &dim->buf_height, &dim->buf_width);
    }
}

static void make_page_index(DVIFILE_INFO* dvi, DIMENSION* dim)
{
    int i;
    long offset;
    
    if (dim->total_page <= 0)
        error(COMMAND_ERROR, "No pages in dvi-file");
    
    dim->page_index =
        (PAGE_INDEX *)marea(sizeof(PAGE_INDEX) * (dvi->total_page + 1));

    for (offset = dvi->last_bop, i = dvi->total_page; i > 0; i--) {
        fseek(dvi->file_ptr, offset, SEEK_SET);
        if ((uchar)read_byte(dvi->file_ptr) != BOP)
            error(COMMAND_ERROR, "No BOP command in page %d", i);
        
        dim->page_index[i].offset = offset;
        dim->page_index[i].number = read_long(dvi->file_ptr);
            /* Read count[0] */
        fseek(dvi->file_ptr, 36L, SEEK_CUR);
            /* Skip other 'count' values */
        offset = read_long(dvi->file_ptr);
    }
}

/* end of file : init.c */
